#==============================================================================
#
#  $Id$
#
"""
   The yed library.
   
   At present, this is just the application-layer stuff, but even so it 
   doesn't belong here and needs to be refactored out into other modules.
"""
#
#   Copyright (C) 2007 Michael A. Muller
#   Copyright (C) 2009 Google Inc.
#
#   This file is part of SpugEdit.
#
#   SpugEdit is free software: you can redistribute it and/or modify it under
#   the terms of the GNU Lesser General Public License as published by the
#   Free Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   SpugEdit is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with SpugEdit.  If not, see <http://www.gnu.org/licenses/>.
#
#==============================================================================

from core import *

# imports required by commands
import popen2
import pprint
import string
import time
import os
from StringIO import StringIO

# holds the text in the copy buffer
copyBuffer = ''

# last search string
lastSearch = ''

_app = None
def setApp(app):
   """
      Set the application object.
      parms:
         app::
            [@yed.app.App]
   """
   global _app
   _app = app

def getApp():
   """
      Returns the application object (a @yed.app.App instance).
   """
   return _app

def createBufferFromFile(top, filename, realname = None, bufferType = None,
                         initialContents = 'FILE NOT FOUND'
                         ):
   """
      Create a new buffer and load it with the contents of the named file.

      parms:
         top::
            [@Top] toplevel window that will own the new buffer.
         filename::
            [string] the fully qualified name of the file to load
         realname::
            [string or None] if specified, this is the fully qualified name of 
            the file that "filename"
         bufferType: [@BufferType or None] if specified, this is the buffer 
            type for the buffer if it is created or loaded.  If it is not 
            specified, the buffer type is determined from the file contents 
            and name.
         initialContents: [str or callable<@Buffer>] The intial contents of 
            the buffer if it is an empty file.
   """

   # create the buffer using the original filename
   buffer = top.createBuffer(filename)
   
   # and read it
   return readBufferFromFile(top, buffer, filename, realname, bufferType,
                             initialContents
                             )

def selectBufferFromFilename(top, filename, statInfo = None, callback = None,
                             bufferType = None,
                             initialContents = 'FILE NOT FOUND'
                             ):
   """
      Select a buffer containing the specified file.  If there is already a 
      buffer containing the file, just bring it to the foreground, reloading 
      the file if it is not up to date.  Otherwise create a new buffer and 
      load the file into it.
      
      parms:
         top: [@Top] the toplevel window for the buffer.
         filename: [str] the name of the file.
         statInfo: [tuple or None] file status tuple.  Retrieved if not 
            provided.
         callback: [LoadFileCmd.RemoteCallback or None] a remote callback if 
            one is required.
         bufferType: [@BufferType or None] if specified, this is the buffer 
            type for the buffer if it is created or loaded.  If it is not 
            specified, the buffer type is determined from the file contents 
            and name.
         initialContents: [str or callable<@Buffer>] The intial contents of 
            the buffer if it is an empty file.
      
      returns:
         [@Buffer] the buffer created or selected.
   """
   # get statInfo if it wasn't provided.
   if not statInfo:
      statInfo = getApp().filer.stat(filename)

   # resolve all symlinks
   realname = os.path.realpath(filename)
   
   # see if the file already is in the buffer ring
   buf = top.findBuffer(realname = realname)   
   if buf:
      # it is... is the disk copy more recent?
      if statInfo and statInfo[stat.ST_MTIME] > buf.attrs.get('mtime', 0):
         # give the user the option to reload
         prompt = YesNoPrompt(top, 'File has changed.  Reload?',
                              LoadFileCmd.Reloader(top, buf,
                                                   filename,
                                                   realname,
                                                   callback
                                                   )
                              )
         top.addStatusWidget(prompt)
         return
   else:
      buf = createBufferFromFile(top, filename, realname, bufferType, 
                                 initialContents
                                 )
      saveEditorState(top)   
   
   # add the callback as a destroy handler - callback will be invoked on 
   # buffer destruction.
   if callback:
      buf.addDestroyHandler(callback)
   
   # bring the buffer to the foreground
   top.focus(buf)
   showBufferInfo(top, buf)
         
   return buf

def readBufferFromFile(top, buffer, filename, realname = None, 
                       bufferType = None,
                       initialContents = 'FILE NOT FOUND'
                       ):
   """
      Read the given file into the buffer.
      
      parms:
         top::
            [@Top] toplevel window that will own the new buffer.
         buffer::
            [@Buffer] the buffer
         filename::
            [string] the name of the file to load
         realname::
            [string or None] if specified, this is the fully qualified name of 
            the file that "filename"
         bufferType: [@BufferType or None] if specified, this is the buffer 
            type for the buffer if it is created or loaded.  If it is not 
            specified, the buffer type is determined from the file contents 
            and name.
         initialContents: [str or callable<@Buffer>] The intial contents of 
            the buffer if it is an empty file.
   """
   # resolve symlinks
   if not realname:
      realname = os.path.realpath(filename)

   try:
      filer = getApp().filer
      text = filer.read(realname)
      statInfo = filer.stat(realname)
      buffer.attrs['mtime'] = statInfo[stat.ST_MTIME]
      buffer.attrs['mode'] = statInfo[stat.ST_MODE]
   except IOError:
      if callable(initialContents):
         initialContents(buffer)
         text = ''
      else:
         text = initialContents
      buffer.attrs['mtime'] = 0
      buffer.attrs['mode'] = 0666
   buffer.insert('end', text)
   buffer.setModified(False)
   buffer.attrs['filename'] = filename
   buffer.attrs['realname'] = realname
   buffer.mark_set('insert', '1.0')
   if bufferType:
      bufferType.initialize(buffer, top)
   else:
      getBufferInitializer(buffer)(top, buffer)
   return buffer

def saveBuffer(buffer, filename = None, force = True):
   """
      Save the contents of the buffer to the given filename.
      
      Raises "Abort" if the file has been modified.
      
      parms:
         buffer::
            [@Buffer] the buffer to save
         filename::
            [string or None] the filename to save it under.  If *None*
            the filename of the buffer will be used.
   """
   # if no filename was provided, get it from the buffer
   if filename is None:
      filename = buffer.attrs['realname']
   
   # make sure that we're not overwriting a newer file
   filer = getApp().filer
   statInfo = filer.stat(filename)
   if not force and statInfo and \
      statInfo[stat.ST_MTIME] > buffer.attrs.get('mtime', 0):
      raise Abort('File %s is more recent than the buffer' % filename, 
                  confirm = True
                  )
   
   # if the permissions have changed, read them
   if statInfo and statInfo[stat.ST_MODE] != buffer.attrs.get('mode'):
      buffer.attrs['mode'] = statInfo[stat.ST_MODE]
   
   data = buffer.get('1.0', 'end')
   
   # the text control appears to append an extra newline - remove it
   if data[-1] == '\n':
      data = data[:-1]

   # if we got back unicode, encode the data (UTF-8 by default)
   if isinstance(data, unicode):
      encoded = data.encode(buffer.attrs.get('encoding', 'utf-8'))
   else:
      encoded = data

   try:
      statInfo = filer.write(filename, encoded, buffer.attrs.get('mode', 0666))
      buffer.attrs['mtime'] = statInfo[stat.ST_MTIME]
      buffer.setModified(False)
   except IOError, ex:
      raise Abort('Error writing file: %s' % ex)

def saveEditorState(top, filename = None):
   if not filename:
      path = getApp().getInstDir()
      filename = os.path.join(path, 'state')
  
   # write the state file version
   dst = open(filename, 'wb')
   Pickler(dst).dump('0')
   
   # and write the toplevel state
   top.writeState(dst)

def getRealnameFromDir(filename):
   """
      Returns the realname of a file based on the realpath of the directory of 
      "filename" (a string)
      
      parms:
         filename::
            [string]
   """
   dirname, basename = os.path.split(filename)
   dirname = os.path.realpath(dirname)
   return os.path.join(dirname, basename)

class QuitBufferCmd(Command):
   """
      Close the current buffer, verify if there are changes.
   """
   
   class Callback(YesNoHandler):
      
      def __init__(self, top, buffer):
         self.top = top
         self.buffer = buffer
         
      def handleYes(self):
         result = self.top.destroyBuffer(self.buffer)
         saveEditorState(self.top)
         if not result:
            showBufferInfo(self.top, self.top.currentBuffer)
         return result

      def handleNo(self):
         pass            
      
   def process(self, context):
      buf = context.top.currentBuffer
      if buf.getModified():
         prompt = \
            YesNoPrompt(context.top, 'Delete modified buffer %s?' % buf.name, 
                        self.Callback(context.top, context.buffer)
                        )
         context.top.addStatusWidget(prompt)
      else:
         context.top.destroyBuffer(buf)
         saveEditorState(context.top)
         showBufferInfo(context.top, context.top.currentBuffer)
   
   def destroyTheBuffer(self, buffer):
      context.top.destroyBuffer(buffer)

class QuitCmd(Command):
   """
      Quit the editor.
   """
   
   class Processor(YesNoCancelHandler):
      """
         Handles all of the quit command processing - this may be spread over
         multiple callbacks and yes/no prompts, so this object contains an
         iterator over the set of buffers.
      """
      
      def __init__(self, top, force = False):
         self.top = top
         self.buffer = None
         self.bufferIter = iter(top.buffers)
         self.force = force
         
      def handleYes(self):
         try:
            saveBuffer(self.buffer, force = self.force)
            return self.process()
         except Abort, ex:
            if ex.confirm:
               # if they answer yes to this next one, force the save
               self.force = True
               prompt = \
                  YesNoCancelPrompt(self.top,
                                    ex[0] + '  Continue?',
                                    self
                                    )
               self.top.addStatusWidget(prompt)
            else:
               self.top.showError(ex[0])
            
      def handleNo(self):
         return self.process()
      
      def handleCancel(self):
         # restore the buffer to a usable state
         del self.buffer.attrs['keepStatusWidgets']
         self.buffer.text['state'] = NORMAL
      
      def __iter__(self):
         return self.bufferIter
         
      def process(self):
         # give the user the opportunity to save all of the buffers
         for buffer in self:
            if buffer.getModified() and buffer.attrs.has_key('filename'):
               
               # store this in the object so we can pick it up in handleYes()
               self.buffer = buffer
               
               # normally, bringing the buffer into focus would remove the 
               # status widgets.  We want to keep them, and we want to make it 
               # so that the user can't edit this.
               buffer.attrs['keepStatusWidgets'] = True
               buffer.text['state'] = DISABLED
               
               # start the prompt to give the user the opportunity to save.
               self.top.focus(buffer)
               prompt = \
                  YesNoCancelPrompt(self.top, 
                                    'Save modified file %s?' % 
                                     buffer.attrs['filename'],
                                    self
                                    )
               self.top.addStatusWidget(prompt)
               
               # quit here.  We will resume processing when the user addresses 
               # the prompt
               break
         else:
            # all buffers can be safely disposed of
            
            # run through the buffers again to call the destroy handlers
            for buffer in self.top.buffers:
               buffer.callDestroyHandlers()
            
            # just destroy the toplevel
            self.top.top.destroy()
            return True

   def process(self, context):
      self.Processor(context.top).process()

class SaveCmd(Command):
   """
      Save the contents of the current buffer.
   """

   class Callback:
      def __init__(self, top, buffer, cmd):
         self.top = top
         self.buffer = buffer
         self.cmd = cmd
      def handleData(self, data):
         self.buffer.attrs['filename'] = data
         self.buffer.attrs['realname'] = getRealnameFromDir(data)
         self.cmd.saveAs(self.top, self.buffer, data)
      def handleCancel(self):
         pass
   
   class YesNo(YesNoHandler):
      def __init__(self, top, buffer, cmd, filename):
         self.top = top
         self.buffer = buffer
         self.cmd = cmd
         self.filename = filename
      def handleYes(self):
         self.cmd.saveAs(self.top, self.buffer, self.filename, force = True)
      def handleNo(self):
         pass

   # controls whether we're deleting the buffer afterwards
   _delete = False
   
   def process(self, context):
      # make sure that the buffer has a filename
      if not context.buffer.attrs.has_key('realname'):
         context.top.addStatusWidget(Prompt(context.top, 'Filename:',
                                            self.Callback(context.top,
                                                          context.buffer, 
                                                          self
                                                          )
                                            )
                                     )

      else:
         # just call the save function directly
         self.saveAs(context.top, context.buffer, 
                     context.buffer.attrs['realname']
                     )

   def saveAs(self, top, buffer, filename, force = False):
      try:
         saveBuffer(buffer, filename, force)
         
         # if we're in "delete" mode, delete the buffer.  Otherwise just 
         # update the buffer info
         if self._delete:
            top.destroyBuffer(buffer)
            saveEditorState(top)
            showBufferInfo(top, top.currentBuffer)
         else:
            showBufferInfo(top, buffer)

      except Abort, ex:
         if ex.confirm:
            prompt = YesNoPrompt(top, ex[0] + '  Continue?',
                                 self.YesNo(top, buffer, self, filename)
                                 )
            top.addStatusWidget(prompt)
         else:
            top.showError(ex[0])

class SaveAndQuitCmd(SaveCmd):
   """
      Command to save the buffer and destroy it.
   """
   
   # This is rather unaesthetic, but the best way to do this at this time is
   # to have a flag in SaveCmd and just override it for the save and quit 
   # flavor.
   _delete = True

class NewBufferCmd(Command):
   """
      Create a new empty buffer.
   """

   def process(self, context):
      buffer = context.top.createBuffer('noname')
      buffer.text.focus()

class NextBufferCmd(Command):
   """
      Cycle to the next buffer.
   """

   def process(self, context):
      buf = context.top.nextBuffer()
      saveEditorState(context.top)
      showBufferInfo(context.top, buf)

class PrevBufferCmd(Command):
   """
      Cycle to the previous buffer.
   """

   def process(self, context):
      buf = context.top.nextBuffer(False)
      saveEditorState(context.top)
      showBufferInfo(context.top, buf)

class GoToParentCmd(Command):
   """
      Navigate to the parent buffer.
   """
   
   def process(self, context):
      parent = context.buffer.attrs.get('parent')
      if not parent:
         context.top.showError('Buffer has no parent.')
         return
      buf = context.top.selectBuffer(parent)

class BufferTypeAheadHandler(TypeAheadHandler):

   def __init__(self, top):
      self.top = top

   def getOptions(self, text):
      options = []
      for buffer in self.top.buffers:
         if text in buffer.name:
            options.append(buffer.name)
      return options
   
   def getCommonText(self, options, text):
      # find the longest common subsequence

      # create a list of the substrings beginning with "text"
      substrings = []
      for opt in options:
         start = opt.find(text)
         substrings.append(opt[start:])
      
      # return the longest common sustring
      return os.path.commonprefix(substrings)

class SelectBufferCmd(Command):
   """
      Allow the user to enter a buffer name with typeahead.
   """

   # set this to true if we want to relocate the buffer to after the current 
   # buffer
   changePosition = False

   class Callback(PromptHandler):
      def __init__(self, top, changePosition):
         self.top = top
         self.changePosition = changePosition
         
      def handleData(self, bufferName):
         last = self.top.currentBuffer
         buf = self.top.selectBuffer(bufferName, True)
         if buf:
            saveEditorState(self.top)
            showBufferInfo(self.top, buf)
            if self.changePosition:
               self.top.insertBuffer(last, buf)
      
      def handleCancel(self):
         pass

   def process(self, context):
      prompt = TypeAheadPrompt(context.top, 'Buffer name:',
                               self.Callback(context.top,
                                             self.changePosition
                                             ),
                               typeAheadHandler =
                                 BufferTypeAheadHandler(context.top)
                               )
      context.top.addStatusWidget(prompt)

class MoveBufferCmd(SelectBufferCmd):
   """
      Allows the user to enter a buffer name with typeahead, the buffer is 
      then moved to after the current buffer in the ring and selected.
   """
   changePosition = True
      
class ReturnCmd(Command):
   """
      Overrides default return functionality to do smart automatic
      indentation.
   """

   def process(self, context):
      prefix = context.buffer.rules.calcNewlinePrefix(context.buffer, 
                                                      hardReturn = True
                                                      )
      context.buffer.insert('insert', '\n' + prefix)
      context.buffer.see('insert')
      return True

class KeyAliasCmd(Command):

   def __init__(self, top, key):
      Command.__init__(self, top)
      self.key = key

   def process(self, context):
      context.buffer.event_generate(self.key)
      return True

class KeyAliasCmdFactory:

   def __init__(self, key):
      self.key = key

   def __call__(self, top):
      return KeyAliasCmd(top, self.key)

class BufferOutput:
   """
      Emulates an output stream, inserts everything written to the stream into 
      a buffer.
   """

   def __init__(self, text, tag, insertHandler):
      """
         Parms:
            text::
               [@Buffer or None] the buffer to write to
            tag::
               [string] the tag to mark the text with while inserting it.
            insertHandler::
               [callable<string> or None] an arbitrary function to be called 
               with the newly inserted text.
      """
      self.text = text
      self.tag = tag
      self.insertHandler = insertHandler

   def write(self, data):      
      self.text.insert('end', data, self.tag, ignoreRules = True)
      self.text.mark_set('insert', 'end')
      self.text.see('insert')
      if self.insertHandler: self.insertHandler(self.text)

   def flush(self):
      pass

class AlertOutput(BufferOutput):
   """
      This is just like BufferOutput, only it will create a buffer and bring 
      it to the foreground if it the buffer is not already in existence or in 
      the foreground.
      
      The intent of this is to capture streams that require the users's 
      immediate attention, such as standard error.
   """
   
   def __init__(self, top, creator, query, tag, insertHandler):
      """
         Parms:
            top::
               [@Top] the toplevel window
            creator::
               [callable<@Top>] function that can be used to create the buffer 
               if it doesn't exist.
            query::
               [dict<string, any>] a query that can be used to locate the 
               buffer (passed as the arguments to @Top.findBuffer())
            tag::
               [string] the tag to mark the text with while inserting it.
            insertHandler::
               [callable<string> or None] an arbitrary function to be called 
               with the newly inserted text.
      """
      BufferOutput.__init__(self, None, tag, insertHandler)
      self.top = top
      self.query = query
      self.creator = creator
   
   def write(self, text):
      # find or create the buffer
      if not self.text:
         self.text = self.top.findBuffer(**self.query)
         if not self.text:
            self.text = self.creator(self.top)
         self.text.addDestroyHandler(self.__handleBufferDestroyed)
      
      # make sure it has the focus
      if self.top.currentBuffer is not self.text:
         self.top.focus(self.text)
         showBufferInfo(self.top, self.text)
         
      # do the standard write to it.
      BufferOutput.write(self, text)      
      
   def __handleBufferDestroyed(self, buf):
      "Called when the buffer is destroyed - sets text to None"
      self.text = None

class FileTypeAheadHandler(TypeAheadHandler):

   def getOptions(self, text):
      filer = getApp().filer
      return filer.getPrefixed(text)
   
   def getCommonText(self, options, text):
      prefix = os.path.commonprefix([f.fullname for f in options])
      
      # if there is only one option, and it is a directory, add the trailing
      # slash to the common text
      if len(options) == 1 and getApp().filer.isdir(prefix):
         return prefix + os.sep
      else:
         return prefix

class LoadFileCmd(Command):
   """
      Bring up a filename prompt and load a file into a new buffer.
   """
   
   class Creator(YesNoHandler):

      def __init__(self, top, filename, realname, callback, bufferType,
                   initialContents
                   ):
         self.top = top
         self.filename = filename
         self.realname = realname
         self.callback = callback
         self.bufferType = bufferType
         self.initialContents = initialContents

      def handleYes(self):
         buffer = self.top.createBuffer(self.filename)
         if callable(self.initialContents):
            self.initialContents(buffer)
         else:
            buffer.insert('end', self.initialContents)
         self.top.focus(buffer)
         buffer.attrs['filename'] = self.filename
         buffer.attrs['realname'] = self.realname
         if self.bufferType:
            self.bufferType.initialize(buffer, self.top)
         else:
            getBufferInitializer(buffer)(self.top, buffer)
         saveEditorState(self.top)
         showBufferInfo(self.top, buffer)
         
         if self.callback:
            buffer.addDestroyHandler(self.callback)

      def handleNo(self):
         if self.callback: 
            self.callback.handleFailure('user denied create request')

   class Reloader(YesNoHandler):
      def __init__(self, top, buffer, filename, realname, callback):
         self.top = top
         self.buffer = buffer
         self.filename = filename
         self.realname = realname
         self.callback = callback
      
      def handleYes(self):
         # reload the file
         self.buffer.delete('1.0', 'end')
         self.buffer.clearUndoStack()
         readBufferFromFile(self.top, self.buffer, self.filename, 
                            self.realname
                            )
         self.top.focus(self.buffer)
         
         # add the callback to the buffer's destroy handlers
         if self.callback:
            self.buffer.addDestroyHandler(self.callback)
      
      def handleNo(self):
         # just bring the buffer to the foreground
         self.top.focus(self.buffer)
         
         # add the callback to the buffer's destroy handlers
         if self.callback:
            self.buffer.addDestroyHandler(self.callback)
      
   class Loader:

      def __init__(self, top, callback = None, bufferType = None, 
                   initialContents = ''
                   ):
         self.top = top
         self.callback = callback
         self.bufferType = bufferType
         self.initialContents = initialContents

      def handleData(self, filename):
         filer = getApp().filer
         statInfo = filer.stat(filename)
         if not statInfo:
            # the file doesn't exist - create it
            
            # figure out the realname based on the realname of the directory
            realname = getRealnameFromDir(filename)

            creator = LoadFileCmd.Creator(self.top, filename, realname, 
                                          self.callback,
                                          self.bufferType,
                                          self.initialContents
                                          )
            
            self.top.addStatusWidget(YesNoPrompt(self.top,
                                                 'File not found.  Create?',
                                                 creator
                                                 )
                                     )
         else:
            # the file exists - select the buffer if an up-to-date copy is in 
            # memory or load it if not.      
            selectBufferFromFilename(self.top, filename, statInfo, 
                                     self.callback
                                     )

      def handleCancel(self):
         pass

   def process(self, context):
      # if the current buffer has a filename, use that as the current directory
      path = context.buffer.attrs.get('filename')
      if path:
         cwd = os.path.dirname(os.path.abspath(path))
      else:
         cwd = os.getcwd()
      
      # make sure it's slash terminated
      cwd = os.path.join(cwd, '')
      self.load(context, cwd)
   
   @classmethod
   def load(cls, context, dir, bufferType = None, initialContents = ''):
      """
         Bring up the filename selector widget to load a file.  Can be used by 
         other commands to initiate file loading.
         
         args:
            context: [@yed.core.CommandContext] the command context.
            dir: [str] the directory where we are looking for files.  This 
               _must_ end with a slash.
            bufferType: [@BufferType or None] if specified, this is the buffer 
               type for the buffer if it is created or loaded.  If it is not 
               specified, the buffer type is determined from the file contents 
               and name.
            initialContents: [str or callable<@Buffer>] The intial contents of 
               the buffer if the file is being created.
      """
      loader = cls.Loader(context.top, bufferType = bufferType, 
                          initialContents = initialContents
                          )
      context.top.addStatusWidget(TypeAheadPrompt(context.top, 'Filename:', 
                                                  loader,
                                                  dir,
                                                  typeAheadHandler = 
                                                   FileTypeAheadHandler()
                                                  )
                                  )
   
   class RemoteCallback:
      
      def __init__(self, request):
         self.request = request
         
      def handleSuccess(self):
         self.request.sendReply('ok')
      
      def __call__(self, buffer):
         self.handleSuccess()
      
      def handleFailure(self, error):
         self.request.sendReply('error: %s' % error)
         
   def processAsRemote(self, request):
      if not request.args.has_key('filename'):
         request.sendReply('error: the edit command requires a filename '
                            'argument')
      else:
         blocking = request.args.get('block') != str(False)
         loader = self.Loader(request.top, 
                              blocking and self.RemoteCallback(request or None)
                              )
         loader.handleData(request.args['filename'])
         if not blocking:
            request.sendReply('ok')

class MovementCommandFactory(CommandFactory):
   """
      Factory for movement commands - these can all be parameterized with a
      "select" option, indicating that the selection should also move.
   """
   
   def __init__(self, factory, select):
      """
         parms:
            factory::
               [callable<@Top, boolean>] Should create a command.  First argument
               is the toplevel, the second indicates whether "select" is turned
               on for the command.
            select::
               [boolean] indicates whether select is turned on.
      """
      self.factory = factory
      self.select = select
   
   def __call__(self, top):
      return self.factory(top, self.select)

class MovementCommand(Command):
   """
      Commands that move the cursor, and as such can have selection enabled.
      
      These commands include a "select" instance variable.
   """
   
   def __init__(self, top, select):
      Command.__init__(self, top)
      self.select = select
      
class SearchCommandFactory(MovementCommandFactory):
   """
      Factory for creating a search command, which has some additional 
      parameters not used by a movement command.
   """
   
   def __init__(self, factory, select, forward, caseSensitive):
      """
         parms:
            factory::
               [callable<@Top, boolean, boolean, boolean>] Should create a 
               command.  First argument is the toplevel, the remaining 
               arguments mirror those of this constructor: select, forward and 
               caseSensitive.
            select::
               [boolean] indicates whether select is turned on.
            forward::
               [boolean] true indicates forward search, false indicates 
               reverse search.
            caseSensitive::
               [boolean] true indictes that the search is case-sensitive, 
               false indicates that it is not.
      """
      self.factory = factory
      self.select = select
      self.forward = forward
      self.caseSensitive = caseSensitive
   
   def __call__(self, top):
      return self.factory(top, self.select, self.forward, self.caseSensitive)

class SearchUtils:
   """
      Mix-in for implementing search functionality.
   """
   
   def __init__(self, buffer):
      self.buffer = buffer
   
   def _clearSearchTag(self):
      ranges = self.buffer.tag_ranges('search_match')
      while ranges:
         start, end = ranges[:2]
         ranges = ranges[2:]
         self.buffer.tag_remove('search_match', start, end)
   
   def _addSearchTag(self, index, length):
      """
         Tags the text at the index with the "search tag" ("search_match").  
         Also sets the insertion point and makes it visible.  This has the 
         effect of providing a visual indicator of the search result.
         
         parms:
            index::
               [string] index of the matching text
            length::
               [int] length of the matching text
      """
      end = '%s + %s chars' % (index, length)
      self.buffer.mark_set('insert', end)
      self.buffer.tag_add('search_match', index, end)
      self.buffer.tag_raise('search_match')
      self.buffer.see('insert')

class SearchCmd(MovementCommand):

   class Handler(SearchUtils):

      def __init__(self, buffer, select, forward, caseSensitive):
         SearchUtils.__init__(self, buffer)
         self.select = select
         self.forward = forward
         self.caseSensitive = caseSensitive
         self.lastIndex = None
         self.searchText = ''
         
         # references the prompt widget so that we can change its label
         self.prompt = None

         # store the current position 
         buffer.mark_set('search_start', 'insert')

      def handleChange(self, data):

         # if the search string is empty, don't try to search for it
         if not data:
            self._clearSearchTag()
            return
         
         return self.search('search_start', data)

      def search(self, start, text):
         global lastSearch
         parms = {}
         if self.forward:
            parms['forwards'] = True
         else:
            parms['backwards'] = True
            
         index = self.buffer.search(text, start, 
                                    nocase = not self.caseSensitive,
                                    **parms
                                    )
         if index:
            # we found a match...
            if self.lastIndex:
               self._clearSearchTag()
               self.lastIndex = None
            self.searchText = text
            self.lastIndex = index
            self._addSearchTag(index, len(text))
            
            # store the last search result globally
            lastSearch = text
            
            return True
         else:
            # no matching string...
            return False

      def handleData(self, data):

         # if we've moved, clear the search highlights and set the insert
         # position to the final position
         if self.lastIndex:
            self._clearSearchTag()
            
            # if we're selecting, extend the selection
            if self.select:
               self.buffer.moveSelection('search_start', self.lastIndex)
               
            self.buffer.mark_set('insert', self.lastIndex)
               
         self.buffer.text.focus()

      def handleCancel(self):

         # if we've moved, remove the search highlight and restore the cursor
         # to the start position
         if self.lastIndex:
            self._clearSearchTag()
            self.buffer.mark_set('insert', 'search_start')
            self.buffer.see('insert')

         self.buffer.text.focus()
      
      def handleSearchAgain(self):
         # abort if the search text is currently empty
         if not self.searchText:
            return

         self._clearSearchTag()

         # if we're searching forward, find the index to the _end_ of the 
         # last match.  Searching bacwards starts from the beginning
         if self.forward:
            index = '%s + %d chars' % (self.lastIndex, len(self.searchText))
         else:
            index = self.lastIndex

         # and search for the next occurence...
         self.search(index, self.searchText)
      
      def handleSearchAgainFwd(self, event):
         """
            Set to forward search and repeat search.
         """
         self.forward = True
         self.handleSearchAgain()
         
         return 'break'
      
      def handleSearchAgainRev(self, event):
         """
            Set to forward search and repeat search.
         """
         self.forward = False
         self.handleSearchAgain()
         
         return 'break'
      
      def handleLastSearch(self, event):
         """
            Search for the last found search string.
         """
         global lastSearch
         
         # make sure there was a last search
         if not lastSearch:
            self.prompt.bell()
            return 'break'
            
         if self.lastIndex:
            self._clearSearchTag()
         if self.search('search_start', lastSearch):
            self.prompt.entry.delete(0, 'end')
            self.prompt.entry.insert(0, lastSearch)
         else:
            self.prompt.label.configure(text = self.createLabel(lastSearch))
      
      def handleCaseToggle(self, event):
         self.caseSensitive = not self.caseSensitive
         self.prompt.label.configure(text = self.createLabel())
         
         return 'break'
      
      def createLabel(self, failedSearchText = None):
         "Creates a label based on the search settings."
         mods = ''
         if not self.caseSensitive:
            mods += 'I'
         
         if mods:
            mods = '[%s]' % mods
         
         if failedSearchText:
            error = "Can't find %s. " % repr(failedSearchText)
         else:
            error = ''
         
         return '%sFind%s:' % (error, mods)

   def __init__(self, top, select, forward, caseSensitive):
      MovementCommand.__init__(self, top, select)
      self.forward = forward
      self.caseSensitive = caseSensitive

   def process(self, context):
      handler = \
         self.Handler(context.buffer, self.select, self.forward, 
                      self.caseSensitive
                      )
      
      # create the label modifiers
      label = handler.createLabel()
      prompt = SearchPrompt(context.top, label, handler)
      handler.prompt = prompt
      prompt.entry.bind('<Control-f>', handler.handleSearchAgainFwd)
      prompt.entry.bind('<Control-r>', handler.handleSearchAgainRev)
      prompt.entry.bind('<Control-i>', handler.handleCaseToggle)
      prompt.entry.bind('<Control-g>', handler.handleLastSearch)
      context.top.addStatusWidget(prompt)
      return True

class Replacer:
   """
      The compiled form a of a replacement string - used to perform group 
      substitution and expression interpolation.
   """

   plainRx = re.compile(r'[^\\]+')
   
   class GroupReplacer:
      """
         Inserts the value of a single match group.
      """
      
      def __init__(self, groupIndex):
         self.index = groupIndex
      
      def __call__(self, match):
         return match.group(self.index)
   
   class ExprReplacer:
      """
         Evaluates a python expression and inserts the result.
      """
      
      def __init__(self, expr):
         self.expr = compile(expr, '<replace>', 'eval')
      
      def __call__(self, match):
         result = eval(self.expr, {'groups': (match.group(),) + match.groups()})
         return str(result)
      
      __tokRx = re.compile(r'\(|\)|("([^"\\]|(\\.))*")|(\'([^\'\\]|(\\.))*\')'
                           r'|([^\(\)"\']+)'
                           )
      @staticmethod
      def parse(sourceStr):
         """
            Parse an expression from the source string.  The expression should 
            begin after the paren after the initial backslash and will include 
            the final paren (so there's one extra closing paren).
            
            Returns an @ExprReplacer instance and the size of the parsed 
            expression.
         """
         
         # we have to do a little bit of parsing for this because of the 
         # possibility that there could be nested parenthesis.  Counting depth 
         # works.
         depth = 1
         parsed = ''
         while depth:
            m = Replacer.ExprReplacer.__tokRx.match(sourceStr)
            if not m:
               raise Exception('Error parsing expression in replacement string')
            tok = m.group()
            sourceStr = sourceStr[m.end():]
            if tok == '(':
               depth += 1
            elif tok == ')':
               depth -= 1

            if depth: parsed += tok
         
         return Replacer.ExprReplacer(parsed), len(parsed) + 1
   
   def __init__(self, replaceStr):
      """
         Constructs an instance from its string representation.
      """

      # the replacement string as a list of elements - consists of constant 
      # strings and functors that produce a string when called with a regex 
      # match argument.
      self.__elems = []
      while replaceStr:
         
         # check for plain strings
         m = self.plainRx.match(replaceStr)
         if m:
            self.__elems.append(m.group())
            replaceStr = replaceStr[m.end():]
            continue
         
         # check for a backslash expression
         if len(replaceStr) == 1:
            # if this is just a backslash at the end of the string, we treat 
            # it as literal
            self.__elems.append(replaceStr)
         if replaceStr[1] in string.digits:
            # check for \digit
            self.__elems.append(self.GroupReplacer(int(replaceStr[1])))
            replaceStr = replaceStr[2:]
         elif replaceStr[1] == '(':
            # parse a parenthesized expression
            replacer, end = self.ExprReplacer.parse(replaceStr[2:])
            self.__elems.append(replacer)
            replaceStr = replaceStr[end + 2:]
         else:
            # all other values are just passed through
            self.__elems.append(replaceStr[1])
            replaceStr = replaceStr[2:]
   
   def __call__(self, match):
      result = ''
      for elem in self.__elems:
         if isinstance(elem, str):
            result += elem
         else:
            result += elem(match)               
      
      return result

class ReplaceCmd(RangeCmd):
   """
      Regular expression search and replace command.
   """

   lastRegex = ''
   lastReplace = ''
   
   class ReplaceEntryPrompt(StatusWidget):
      def __init__(self, parent, buffer, start, end):
         StatusWidget.__init__(self, parent)
         self.parent = parent
         self.buffer = buffer
         self.start = start
         self.end = end
         
         self.regexLbl = Label(self, text = 'Regex:')
         self.regexLbl.pack(side = LEFT)
         
         self.regexEntry = Entry(self)
         self.regexEntry.pack(side = LEFT, expand = 1, fill = BOTH)
         self.regexEntry.insert(0, ReplaceCmd.lastRegex)
         self.regexEntry.focus()
         
         self.replaceLbl = Label(self, text = 'Replace:')
         self.replaceLbl.pack(side = LEFT)
         
         self.replaceEntry = Entry(self)
         self.replaceEntry.pack(side = LEFT, expand = 1, fill = BOTH)
         self.replaceEntry.insert(0, ReplaceCmd.lastReplace)
         
         for entry in (self.regexEntry, self.replaceEntry):
            entry.bind('<Return>', self.handleReturn)
            entry.bind('<Escape>', self.handleEscape)
      
      def handleReturn(self, event):
         ReplaceCmd.lastRegex = regex = self.regexEntry.get()
         ReplaceCmd.lastReplace = self.replaceEntry.get()
         
         # try to create a prompt and add it, if there were errors in the user 
         # input just show an error message and be done with it.
         try:
            replace = Replacer(ReplaceCmd.lastReplace)
            prompt = \
               ReplaceCmd.ReplacePrompt(self.parent, 
                                        self.buffer,
                                        regex,
                                        replace,
                                        self.start,
                                        self.end
                                        )
            self.parent.addStatusWidget(prompt)
         except Exception, ex:
            self.parent.setStatusText('Argument error: %s' % ex)

         self.parent.removeStatusWidget(self)
      
      def handleEscape(self, event):
         self.parent.removeStatusWidget(self)
   
   class ReplacePromptHandler(PromptHandler):
      """
         This handler just relays return and cancel processing back to the 
         handleYes() and handleCancel() methods of @ReplacePrompt.
      """
      
      def __init__(self, prompt):
         self.prompt = prompt
      
      def handleData(self, data):
         self.prompt.handleYes()
      
      def handleCancel(self):
         self.prompt.handleCancel()
         
   class ReplacePrompt(Prompt, SearchUtils):
      """
         This prompt manages the yes/no/all/cancel loop of a search/replace 
         operation.
      """
      
      def __init__(self, parent, buffer, regex, replace, start, end):
         Prompt.__init__(self, parent, 
                         'replace? (y)es, (n)o, (a)ll, (c)ancel:',
                         ReplaceCmd.ReplacePromptHandler(self)
                         )
         SearchUtils.__init__(self, buffer)
         self.entry.bind('<KeyPress>', self.handleKey)
         
         # all modifications in the search/replace are atomic
         self.buffer.edit_separator()
         self.buffer.setAutoSeparators(False)

         self.orgPos = buffer.index('insert')
         self.pos = start or self.orgPos
         self.end = end
         self.match = None
         self.length = 0
         self.regex = regex
         self.rx = re.compile(self.regex)
         self.replace = replace
         self.search()
      
      def endReplace(self):
         """
            Restores the insertion point to the start position and closes the 
            atomic edit action.
         """
         self.buffer.mark_set('insert', self.orgPos)
         self.buffer.see('insert')
         self.buffer.setAutoSeparators(True)
         self.buffer.edit_separator()
      
      def abort(self):
         """
            Terminates the search without changing the status text.
         """
         self._clearSearchTag()
         self.endReplace()
         self.handler = None  # so we don't "handleCancel()"
         self.parent.removeStatusWidget(self)

      def search(self):
         self.pos = pos = \
            self.buffer.search(self.regex, self.pos, self.end, regexp = True)
         self._clearSearchTag()
         
         if not pos:
            self.parent.setStatusText('No remaining matches')
            self.abort()
            return False
         
         # figure out the length of the match
         line = self.buffer.get(pos, '%s lineend' % pos)
         self.match = m = self.rx.match(line)
         self.length = len(m.group())
         
         # tag the result
         self._addSearchTag(pos, self.length)
         return True

      def replaceCurrentMatch(self):
         # evaluate the replacement text - if we get an exception, abort
         try:
            replacement = self.replace(self.match)
         except Exception, ex:
            self.parent.setStatusText('Error calculating replacement: %s' % ex)
            return False

         self.buffer.delete(self.pos, 
                            '%s + %d chars' % (self.pos, self.length)
                            )
         tempMark = 'yed.ReplaceCmd.matchEnd'
         self.buffer.mark_set(tempMark, self.pos)
         self.buffer.mark_gravity(tempMark, 'right')
         
         # insert the replacement text
         self.buffer.insert(self.pos, replacement)
         
         # next search position starts after the replacement
         self.pos = self.buffer.index(tempMark)
         self.buffer.mark_unset(tempMark)
         
         return True
      
      def handleYes(self):
         # replace the match
         if self.replaceCurrentMatch():
            self.search()
         else:
            self.abort()
         
      def handleNo(self):
         self.pos = self.buffer.index('%s + 1 chars' % self.pos)
         self.search()
         
      def handleAll(self):
         while self.search():
            if not self.replaceCurrentMatch():
               self.abort()
               return

      def handleCancel(self):
         self._clearSearchTag()
         self.endReplace()
         showBufferInfo(self.parent, self.buffer)
         
      def handleKey(self, event):
         if event.char:
            if event.char == 'y':
               self.handleYes()
               return 'break'
            elif event.char == 'n':
               self.handleNo()
               return 'break'
            elif event.char == 'a':
               self.handleAll()
               return 'break'               
            elif event.char == 'c':
               self.parent.removeStatusWidget(self)
               return 'break'
      
   def processNonRange(self, context):
      prompt = self.ReplaceEntryPrompt(context.top, context.buffer, 
                                       '1.0',
                                       'end'
                                       )
      context.top.addStatusWidget(prompt)
   
   def processRange(self, context, start, end):
      prompt = self.ReplaceEntryPrompt(context.top, context.buffer, 
                                       start,
                                       end
                                       )
      context.top.addStatusWidget(prompt)

class GotoCmd(MovementCommand):
   """
      Prompt the user for a line number then jump to that line.
   """
   
   class Callback(PromptHandler):
      def __init__(self, top, buffer, select):
         self.top = top
         self.buffer = buffer
         self.select = select
         
      def handleData(self, data):
         try:
            lineNum = int(data)
         except ValueError:
            self.top.showError('%s is not a valid line number' % lineNum)
         
         # calculate the index
         index = '%d.0' % lineNum
         
         # if we are in select mode, add the difference between the current
         # position and the new position to the selection text
         if self.select:
            self.buffer.moveSelection('insert', index)
            
         # move the insertion cursor, make sure it's visible
         self.buffer.mark_set('insert', index)
         self.buffer.see('insert')
         self.buffer.text.focus()
         
         self.top.setStatusText('Line %d' % lineNum)
         
      def handleCancel(self):
         pass
         
   def process(self, context):
      prompt = Prompt(context.top, 'Line number:', 
                      self.Callback(context.top, context.buffer, self.select)
                      )
      context.top.addStatusWidget(prompt)

class AlignToTopCmd(Command):
   """
      Moves the view so that the line that the cursor is on is aligned with 
      the top of the view.
   """
   
   def process(self, context):
      # calculate the cursor row and end row as "rows since the top of the 
      # buffer" - note that the end pointer is always at last line number + 1, 
      # so we subtract one to get the number of lines in the buffer
      row = context.buffer.getRow('insert') - 1
      endRow = context.buffer.getRow('end') - 1
      context.buffer.yview_moveto(float(row) / endRow)

class AlignToBottomCmd(Command):
   """
      Moves the view so that the line that the cursor is on is aligned with 
      the bottom of the view.
   """
   
   def process(self, context):
      # figure out the current fraction of view size/buffer size
      viewTop, viewBottom = context.buffer.yview()
      viewable = viewBottom - viewTop
      
      # figure out the current row as a fraction of the number of lines
      row = context.buffer.getRow('insert')
      endRow = context.buffer.getRow('end') - 1
      pos = float(row) / endRow
      
      # figure out the desired end position as a fraction of total size
      desiredEndPos = pos - viewable 
      
      # if this is positive, move to it.  Otherwise just move to the beginning 
      # of the buffer
      if desiredEndPos >= 0:
         context.buffer.yview_moveto(desiredEndPos)
      else:
         context.buffer.yview_moveto(0)

def handleSelect(offset, length):
   global copyBuffer
   return copyBuffer[int(offset):int(offset + length)]

class CopyCmd(RangeCmd):
   """
      Command to copy text to the copy buffer.
   """
   def processRange(self, context, start, end):
      
      # copy to yed's internal copy buffer
      global copyBuffer
      copyBuffer = context.buffer.get(start, end)
      
      # copy to the clipboard
      context.buffer.selection_handle(handleSelect)
      context.buffer.selection_own()
      
      return True
      
class CutCmd(RangeCmd):
   """
      Command to copy text to the copy buffer.
   """
   def processRange(self, context, start, end):
      global copyBuffer
      copyBuffer = context.buffer.get(start, end)
      context.buffer.delete(start, end)
      return True
      
class PasteCmd(Command):
   """
      Insert text from the copy buffer.
   """
   
   def process(self, context):
      global copyBuffer
      context.buffer.insert('insert', copyBuffer)
      context.buffer.see('insert')
      return True

class BackSpaceCmd(RangeCmd):
   """
      Override backspace processing to backspace over tabs.
   """
   
   def processRange(self, context, start, end):
      """
         When used with a range, deletes the range.
      """
      context.buffer.delete(start, end)
      return True
      
   def processNonRange(self, context):
      line = context.buffer.get('insert linestart', 'insert')
      expanded = line.expandtabs(context.buffer.getTabWidth())
      m = wsRx.match(expanded)
      # if the leading whitespace is evenly divisble by the tab width,
      # delete a full tab width
      tabs = context.buffer.attrs['tabs']
      if m and len(line) == len(m.group()) and not (len(m.group()) % tabs):
            
         # if we had to expand tabs, replace the entire beginning of the
         # line
         if expanded != line:
            context.buffer.delete('insert linestart', 'insert')
            context.buffer.insert('insert', expanded[:-tabs])
         else:
            context.buffer.delete('insert - %d chars' % tabs, 'insert')
      else:
         # remove a single character
         context.buffer.delete('insert - 1 chars')
      
      return True

class TabCmd(RangeCmd):

   unselectAfterProcessing = False
   
   def processRange(self, context, start, end):
      # when applied to a range, indent the entire range
      
      # add a separator, the turn autoseparators off so that the outdent will
      # be regarded as a single action
      context.buffer.edit_separator()
      context.buffer.setAutoSeparators(False)
      
      cur = start
      while context.buffer.compare(cur, '<', '%s lineend' % end):
         # get the current line
         line = context.buffer.get('%s linestart' % cur, '%s lineend' % cur)
         
         # expand tabs
         expanded = line.expandtabs(context.buffer.getTabWidth())
         
         # append a tab to the leading whitespace
         m = wsRx.match(expanded)
         if m:
            tabs = context.buffer.attrs['tabs']
            line = m.group() + (tabs - m.end() % tabs) * ' ' + \
               expanded[m.end():]
         else:
            line = ' ' * context.buffer.attrs['tabs'] + line
         
         # reinsert the line
         context.buffer.delete('%s linestart' % cur, '%s lineend' % cur)
         context.buffer.insert('%s linestart' % cur, line)
         
         # increment the current pointer
         cur = context.buffer.index('%s + 1 lines' % cur)
      
      # turn autoseparators back on and add one to mark off the end of the 
      # action
      context.buffer.setAutoSeparators(True)
      context.buffer.edit_separator()
      
      context.buffer.tag_add('sel', '%s linestart' % start, 
                             '%s lineend' % end
                             )
      
      return True
         
   def processNonRange(self, context):
      # get the index of the insertion cursor
      row, col = \
         map(int, tuple(context.buffer.index('insert').split('.')))

      # insert enough characters to offset to the next whitespace
      tabs = context.buffer.attrs['tabs']
      context.buffer.insert('insert', ' ' * (tabs - col % tabs))
      return True

class UntabCmd(RangeCmd):
   """
      back-tab untabs the region.
   """
   
   unselectAfterProcessing = False
   
   def processRange(self, context, start, end):
      cur = start
      tabs = context.buffer.attrs['tabs']
      
      # add a separator, the turn autoseparators off so that the outdent will
      # be regarded as a single action
      context.buffer.edit_separator()
      context.buffer.setAutoSeparators(False)
      
      while context.buffer.compare(cur, '<', '%s lineend' % end):
         
         # get the current line
         line = context.buffer.get('%s linestart' % cur, '%s lineend' % cur)
         
         # see if it starts with at least one tabstop
         expanded = line.expandtabs(context.buffer.getTabWidth())
         m = wsRx.match(expanded)
         if m and len(m.group()) >= tabs:
            # remove the last level
            line = m.group()[:-tabs] + expanded[m.end():]
            
            # rewrite the line
            context.buffer.delete('%s linestart' % cur, '%s lineend' % cur)
            context.buffer.insert('%s linestart' % cur, line)
         
         # increment to the next line
         cur = context.buffer.index('%s + 1 lines' % cur)
      
      # turn autoseparators back on and add one to mark off the end of the 
      # action
      context.buffer.setAutoSeparators(True)
      context.buffer.edit_separator()
      
      # select the complete lines
      context.buffer.tag_add('sel', '%s linestart' % start, 
                             '%s lineend' % end
                             )
      
      return True

class ReflowCmd(RangeCmd):
   """
      Reflows the selected text according to the formatting rules for the 
      buffer.
   """
   
   atomicEdits = True
   
   def processRange(self, context, start, end):
      
      # store some commonly used variables
      buffer = context.buffer
      textwidth = buffer.attrs.get('textwidth', 78)
      buffer.rules.reflow(buffer, start, end, textwidth)
      
      return True

class PipeCmd(RangeCmd):
   """
      Write the marked region to a command, then read the command output back 
      into the buffer.
   """
   
   class Callback(PromptHandler):
   
      def __init__(self, buffer, start, end):
         self.buffer = buffer
         self.start = start
         self.end = end
         
      def handleData(self, command):
         childOut, childIn = popen2.popen2(command)
         
         # get the input data
         inputData = self.buffer.get(self.start, self.end)
         childIn.write(inputData)
         childIn.close()
         data = childOut.read()
         
         # have to do the atomic edits stuff the hard way, since we're no 
         # longer in the handler
         self.buffer.edit_separator()
         self.buffer.setAutoSeparators(False)
         
         # delete the range, replace with the output
         self.buffer.delete(self.start, self.end)
         self.buffer.insert(self.start, data)
         
         self.buffer.setAutoSeparators(True)
         self.buffer.edit_separator()
      
      def handleCancel(self):
         pass

   __history = []

   def __init__(self, *args):
      RangeCmd.__init__(self, *args)
         
   def processRange(self, context, start, end):
      handler = self.Callback(context.buffer, start, end)
      context.top.addStatusWidget(HistoryPrompt(context.top, 'Command:', 
                                                handler,
                                                history = self.__history
                                                )
                                  )

class BlockInsertCmd(RangeCmd):
   """
      Prompt the user for text, insert the text at the start of the range and 
      the same column on every subsequent line until the end of the 
      range.
   """
   class Callback(PromptHandler):
   
      def __init__(self, buffer, start, end):
         self.buffer = buffer
         self.start = start
         self.end = end
         
      def handleData(self, text):
         cur = self.start
         while self.buffer.compare(cur, '<=', self.end):
            self.buffer.insert(cur, text)
            cur = self.buffer.index('%s + 1 lines' % cur)
      
      def handleCancel(self):
         pass

   def processRange(self, context, start, end):
      handler = self.Callback(context.buffer, start, end)
      context.top.addStatusWidget(Prompt(context.top, 'Text to insert:',
                                         handler
                                         )
                                  )

class BlockDeleteCmd(RangeCmd):
   """ 
      Prompts the user for a line of text that will be deleted if it exists 
      at the start of the range and on the same column of every subsequent 
      line until the end of the range. 
   """
   class Callback(PromptHandler):
   
      def __init__(self, buffer, start, end):
         self.buffer = buffer
         self.start = start
         self.end = end
         
      def handleData(self, text):
         cur = self.start
         textLen = len(text)
         while self.buffer.compare(cur, '<=', self.end):
            end = '%s + %d chars' % (cur, textLen)
            if self.buffer.get(cur, end) == text:
               self.buffer.delete(cur, end)
            cur = self.buffer.index('%s + 1 lines' % cur)
            
      def handleCancel(self):
         pass

   def processRange(self, context, start, end):
      handler = self.Callback(context.buffer, start, end)
      context.top.addStatusWidget(Prompt(context.top, 'Text to delete:',
                                         handler
                                         )
                                  )

class YedletCommandFactory(CommandFactory):
   """
      Creates a YedletCmd, keeps track of the code and whether it needs to be 
      reloaded.
   """
   
   def __init__(self, path, name):
      """
         parms:
            path: [str] the yedlet directory.
            name: [str] the command name
      """
      self.__path = path
      self.__name = name
      self.__code = None
      self.__ts = 0
   
   def __call__(self, top):
      return YedletCmd(self, top)
   
   def getCode(self):
      """
         Returns the codeblock for the yedlet, loading or reloading the source 
         file as necessary.
      """
      
      # reload the file if its timestamp is more recent than the last we have 
      # loaded.
      filename = os.path.join(self.__path, self.__name)
      fileTime = os.path.getmtime(filename)
      if fileTime > self.__ts:
         self.__code = compile(file(filename).read(), filename, 'exec')
         self.__ts = fileTime
      
      return self.__code

class YedletCmd(RangeCmd):
   """
      This command runs an arbitrary "yedlet" - a script in the .yed directory 
      that is sort of a cross between a pipe command and a plugin.
   """

   class OutputRegion(StringIO):
      """
         StringIO object that keeps track of whehter it has been written.
      """
      
      def __init__(self):
         
         # tracks whether we've done a write
         self.written = False
         StringIO.__init__(self)
         
      def write(self, data):
         self.written = True
         StringIO.write(self, data)
      
   def __init__(self, factory, top):
      RangeCmd.__init__(self, top)
      self.__factory = factory

   def processRange(self, context, start, end):
      code = self.__factory.getCode()
      output = self.OutputRegion()
      exec code in {
         'src': StringIO(context.buffer.get(start, end)),
         'dst': output,
         'buf': context.buffer,
         'top': context.top,
         'start': start,
         'end': end,
         'args': context.args,
         'context': context,
      }
      
      # if we wrote to the output, replace the marked region with the output
      if output.written:
         context.buffer.delete(start, end)
         context.buffer.insert(start, output.getvalue())
      
      return context.result
   
   def processNonRange(self, context):
      insert = context.buffer.index('insert')
      return self.processRange(context, insert, insert)
   
class TestCmd(Command):
   def process(self, context):
      buffer = context.buffer
      buffer.edit_separator()
      buffer.setAutoSeparators(False)
      buffer.insert('insert', 'this is a bogus')
      buffer.delete('insert - 8 chars', 'insert')
      buffer.insert('insert', ' a test')
      buffer.setAutoSeparators(True)
      buffer.edit_separator()
            
class DumpCmd(Command):
   
   def process(self, context):
      sys.stdout.write(repr(context.buffer.dump('1.0', 'end', mark = True, tag = True)))

class CommandHandler:
   """
      Callback interface for CommandLineCmd.
   """

   def handleCommand(self, context, command):
      """
         Called to process a command.
         
         parms:
            context::
               [@yed.core.CommandContext] the command context.
            command::
               [string] command and all arguments.
      """
      raise NotImplementedError()
      
class CommandLineCommandFactory(CommandFactory):
   """
      Factory to create commands that process command lines.
   """
   def __init__(self, label, commandHandler):
      self.__label = label
      self.__commandHandler = commandHandler
      
      # create this with a new history list - we therefore assume that command 
      # histories will be partitioned by the factory where they originated
      self.__history = []
   
   def __call__(self, *args):
      return CommandLineCmd(self.__label, self.__commandHandler, 
                            self.__history, 
                            *args
                            )

class CommandLineCmd(Command):
   
   class Callback(PromptHandler):
      def __init__(self, context, commandHandler):
         self.context = context
         self.commandHandler = commandHandler
      
      def handleData(self, data):
         self.commandHandler.handleCommand(self.context, data)
      
      def handleCancel(self):
         pass

   def __init__(self, label, commandHandler, history, *args):
      Command.__init__(self, *args)
      self.__label = label
      self.__commandHandler = commandHandler
      self.__history = history

   def process(self, context):
      handler = self.Callback(context, self.__commandHandler)
      context.top.addStatusWidget(HistoryPrompt(context.top, self.__label, 
                                                handler, 
                                                history = self.__history
                                                )
                                  )

class ExecHandler(CommandHandler):
   """
      Execute a single line python command.
   """
   
   def handleCommand(self, context, command):
      ctx = {'top': context.top, 'buffer': context.buffer}
      try:
         exec command in ctx
      except:
         tp, ex, tb = sys.exc_info()
         context.top.setStatusText(str(ex))

class YedCmdHandler(CommandHandler):
   """
      Command line handler that executes a yed command.
   """
   
   def handleCommand(self, context, command):
      # split it up into words
      parsedCommand = command.strip().split()
      context.args = parsedCommand[1:]
      
      # and try to resolve it, first in the buffer rules, then 
      cmd = context.buffer.rules.lookupCommand(parsedCommand[0], context) or \
            getApp().lookupCommand(parsedCommand[0], context)
      if cmd:
         context.args = parsedCommand[1:]
         cmd.process(context)
      else:
         context.top.showError('Invalid command: %s' % parsedCommand[0])
         
         
      return True

#class KeySequenceCommandFactory:
#   
#   def __init__(self, sequence, 
#   def __call__(self, *args):
#      return KeySequenceCmd(
#class KeySequenceCmd(Command):
#   """
#      This command serves as the root for all commands that are sequences of 
#      multiple keystrokes.
#   """
#   
#   class KeySequencePrompt(Prompt):
#      
#   class Callback(PromptHandler):
#      
#   
#   def __init__(self, 
#   def process(self, context):
#      handler = self.Callback(context.top, context.buffer)
#      context.top.addStatusWidget(Prompt(context.top, '
         

def applyBindings(widget, top, bindings):
   for key, factory in bindings.items():
      widget.bind(key, factory(top))

wsRx = re.compile(r'[ \t]+')
def defaultInitializer(top, buffer):

   # try to figure out the filetype
   filename = buffer.attrs.get('filename', '')
   firstline = buffer.get('1.0', '1.0 lineend')

   # first try a lookup
   bufferType = getBufferType(filename, firstline)
   bufferType.initialize(buffer, top)

class Parser:
   
   def __init__(self, buffer, start = '1.0', end = 'end'):
      self.__buffer = buffer
      self.__cur = start
      self.__end = end
      self.__tags = []
      
   def next(self):
      """
         Returns a tuple consisting of the token type, its start position and
         its end position.
      """
      start = self.__cur
      if self.__buffer.compare(start, '>=', self.__end):
         raise StopIteration()
         
      restOfLine = self.__buffer.get(self.__cur, 
                                     self.__cur + ' lineend + 1 chars'
                                     )
      for type, rx, ender in self.tokens:
         
         # see if the token matches
         m = rx.match(restOfLine)
         if m:
            # point the cursor to the end of the token
            cur = self.__cur = \
               self.__buffer.index('%s + %d chars' % (start, m.end()))
            
            # if there's no "ender", we're done - otherwise, move on to phase
            # two
            if not ender:
               return type, start, self.__cur
            else:
               break
      else:
         self.__cur = self.__end
         return 'error', start, self.__end
      
      # Phase 2: find the ender
      while self.__buffer.compare(self.__cur, '<', self.__end):
         restOfLine = self.__buffer.get(self.__cur, self.__cur + ' lineend')
         m = ender.match(restOfLine)
         if m:
            # point to the end of the match
            self.__cur = '%s + %d chars' % (self.__cur, m.end())
            return type, start, self.__cur
         
         # next line
         self.__cur = '%s + 1 line linestart' % self.__cur
      else:
         return 'error', start, self.__end
   
   def __iter__(self):
      return self
   
   def getTags(cls):
      """
         Returns all of the tag names (list<string>) used by the parser.
      """
      if not hasattr(cls, '_Parser__tags'):
         
         # create the tags for this parser class - start with "error" which
         # is effectively a syntax tag.
         cls.__tags = ['error']
         for tag, rx, ender in cls.tokens:
            cls.__tags.append(tag)
      
      return cls.__tags
   getTags = classmethod(getTags)

class CLikeParser(Parser):
   tokens = [
      ('symbol', re.compile(r'[\{\}\[\]\(\)!@$%\^&\*\-\+=\|\\:;<>,\.\?\/\~]'),
       None
       ),
      ('string', re.compile(r'"([^"\\]|\\.)*"|\'([^\'\\]|\\.)*\''), None),
      (None, re.compile(r'[^\{\}\[\]\(\)!@$%\^&\*-\+=\|\\:;<>,\.\?\/\~#"\']+'),
       None
       ),
   ]

class ParsingRules(Rules):
   """
      Rules for buffers that contain formal, parseable languages.
   """

   def getParser(self, buffer, start, end):
      return self.getParserClass()(buffer, start, end)
   
   def getParserClass(self):
      raise NotImplementedError()

   def initialize(self, buffer):
      self.retagRegion(buffer, '1.0', 'end', calcTabs = True)
   
   def retagRegion(self, buffer, start, end, calcTabs = False):
      """
         Implements @Rules.retagRegion().  Adds one special argument, 
         "calcTabs", which if true indicates that the method should
         try to establish the indentation width for the buffer based
         on its contents.
      """
         
      parser = self.getParser(buffer, start, end)
      
      # keep track of the level of parenthesis nesting
      parenLevel = 0
      
      for tok in parser:            
         if tok[0]:
            buffer.tag_add(tok[0], tok[1], tok[2])

            # if we're calculating tabs, keep track of the parenthesis level.
            # There's a subtle pragmatic reason for this - indentation in 
            # parameter lists is often aligned to the first parenthesis.  
            # Since formal argument lists are often the first indentation in a 
            # file in C++, this ends up screwing up the indentation because it 
            # makes the parser thing that the file indentation is based on the 
            # indentation of the first parameter list.
            if calcTabs and tok[0] == 'symbol':
               tokText = buffer.get(tok[1], tok[2])
               if tokText == '(':
                  parenLevel += 1
               elif tokText == ')':
                  parenLevel -= 1
         
         # whitespace is currently one of the "typeless" tokens, so if we
         # got something typeless and we're calculating tabs and this token
         # is at the start of the line and some whitespace is at the start of
         # the line, use that as the indent level
         elif calcTabs and \
              buffer.compare(tok[1], '==', '%s linestart' % tok[1]) and \
              not parenLevel:
            m = wsRx.match(buffer.get(tok[1], tok[2]))
            if m:
               tabWidth = buffer.getTabWidth()
               buffer.attrs['tabs'] = len(m.group().expandtabs(tabWidth))
               calcTabs = False
      
      # if we still need to calculate the tabs, set it to the default
      if calcTabs:
         buffer.attrs['tabs'] = getApp().get('tabs', 4)
   
   def unmatchedParen(self, line):
      diff = line.count('(') - line.count(')')
      if diff > 0:
         # "diff" is now the number of parenthesis that we want to align, with,
         # so find the "diff"th parenthesis
         index = 0
         while diff:
            index = line.find('(', index) + 1
            diff -= 1
         return index
      return 0
   
   def getCommentPrefix(self, line, hardReturn):
      """
         If the line begins with a full-line comment indicator (such as 
         '#' in shell/python/perl or '//' in C++), returns the expected
         prefix of the following line.  Otherwise returns something false.
         
         Must be implemented by derived classes.
         
         parms:
            line: [str] the existing line to be checked for a comment
            hardReturn: [bool] true if the user has explicitly pressed the 
               return key, false if a return was inferred due to line length.
      """
      raise NotImplementedError()
      
   wsRx = re.compile(r'\s+')
   def calcIndent(self, line, tabs, hardReturn, stringLine):
   
      # empty lines are always followed by no indent
      if not line:
         return ''

      # start with the existing indentation
      m = self.wsRx.match(line)
      if m:
         indent = m.group()
      else:
         indent = ''
      
      # check for a line that explicitly starts with a comment
      commentPrefix = self.getCommentPrefix(line, hardReturn)
      if commentPrefix is not None:
         return commentPrefix
      
      # circumvent all syntax checks if the line begins in a string token
      if stringLine:
         return indent
      
      # check for an unmatched paren on the line
      unmatchedParen = self.unmatchedParen(line)
      if unmatchedParen:
         indent = ' ' * unmatchedParen
         return indent
   
      # check for any of the special characters that suggest that are normally
      # followed by an additional indent
      if line[-1] in (':', '{', '[', '('):
         indent = indent + ' ' * (tabs - len(indent) % tabs)

      return indent
   
   def calcNewlinePrefix(self, buffer, index = 'insert', hardReturn = False):
      line = buffer.get('%s linestart' % index, index)
      inString = 'string' in buffer.tag_names(index)
      prefix = self.calcIndent(line, buffer.attrs['tabs'], hardReturn, inString)
      return prefix
   
   def isJoinable(self, buffer, index = 'insert'):
      
      # first verify that we've got a next line
      gotNextLine = \
         buffer.compare(buffer.index('%s + 1 lines' % index), '<', 'end')
      if not gotNextLine:
         return False, 0

      # get the current line and the next line and see if they are both 
      # comments.

      line = buffer.get('%s linestart' % index, '%s lineend' % index)
      nextLineIndex = buffer.index('%s + 1 lines' % index)
      linePrefix = self.getCommentPrefix(line, False)
      nextLine = buffer.get('%s linestart' % nextLineIndex,
                            '%s lineend' % nextLineIndex
                            )      
      nextLinePrefix = self.getCommentPrefix(nextLine, False)
      
      # if the current line and the next line are both comments, we can join 
      # up to the beginning of the comment text on the next line
      if linePrefix and nextLinePrefix:
         return (True, len(nextLinePrefix))
      
      # if neither line is a comment, we can join up through the first 
      # whitespace in the next line
      elif not linePrefix and not nextLinePrefix:
         ws = wsRx.match(nextLine)
         return True, ws and len(ws.group()) or 0
      
      else:
         # one line is a comment, one is not.  We can not join
         return False, 0

   def getSyntaxSensitiveTags(self):
      return self.getParserClass().getTags()
      
class ShellCommentRules(ParsingRules):
   """
      Base class for rules classes that have shell-like comments.
   """
   __commentRx = re.compile(r'\s*#\s*')
   def getCommentPrefix(self, line, hardReturn):
      m = self.__commentRx.match(line)
      if m and not hardReturn:
         return m.group()
      
      return None

class TextRules(Rules):
   """Basic rules applicable for all text files."""

   def initialize(self, buffer):   
      # try to figure out the tabstop from the first indented line
      i = 1
      while i > 0:
         line = buffer.get('%d.0' % i, '%d.0' % (i + 1))
         if not line:
            # set the index to zero and continue to force the "else" condition
            # of the while
            i = 0
            continue
         m = wsRx.match(line)
         if m:
            line.expandtabs(buffer.getTabWidth())
            buffer.attrs['tabs'] = len(m.group())
            break
         i += 1
      else:
         # set the default tabs
         buffer.attrs['tabs'] = getApp().get('tabs', 4)

   def calcNewlinePrefix(self, buffer, index = 'insert', hardReturn = False):
      line = buffer.get('%s linestart' % index, '%s lineend' % index)
      m = wsRx.match(line)
      if m:
         return m.group()
      else:
         return ''

def getBufferInitializer(buffer):
   """
      Figure out the buffer type and return the appropriate initializer (a
      callable<Buffer>).
   """
   return defaultInitializer

topBindings = {
   '<F2>': SaveCmd,
   '<F3>': QuitBufferCmd,
   '<Control-F3>': QuitCmd,
   '<F4>': SaveAndQuitCmd,
   '<F6>': GoToParentCmd,
   '<F8>': LoadFileCmd,
   '<F10>': SelectBufferCmd,
   '<Control-F10>': MoveBufferCmd,
   '<F12>': NextBufferCmd,
   '<F11>': PrevBufferCmd,
   '<Control-F9>': CommandLineCommandFactory('Python Stmt:', ExecHandler()),
}

def remotePing(request):
   request.sendReply('ok')

def remoteLs(request):
   buffers = []
   for buffer in getApp().top.buffers:
      buffers.append(buffer.name)
   request.sendReply('\n'.join(buffers))

def remoteSaveAll(request):
   buffers = []
   for buffer in getApp().top.buffers:
      try:
         modified = buffer.getModified()
         hasRealname = buffer.attrs.has_key('realname')
         if modified and hasRealname:
            saveBuffer(buffer, buffer.attrs['realname'], False)
            buffers.append('%s: saved' % repr(buffer.name))
         elif not hasRealname:
            buffers.append('%s: no realname' % repr(buffer.name))
         else:
            buffers.append('%s: unmodified' % repr(buffer.name))
      except Abort, ex:
         buffers.append('%s: error %s' % (repr(buffer.name), repr(str(ex))))
   
   request.sendReply('\n'.join(buffers))

def remoteReloadAll(request):
   buffers = []
   top = getApp().top
   for buffer in top.buffers:
      try:
         
         # if the file exists and is newer than the buffer, reload it
         realname = buffer.attrs.get('realname')
         if realname is None:
            continue
         filer = getApp().filer
         statInfo = filer.stat(realname)
         if statInfo and statInfo[stat.ST_MTIME] > buffer.attrs['mtime']:
               
            buffer.delete('1.0', 'end')
            buffer.clearUndoStack()
            readBufferFromFile(top, buffer, buffer.attrs['filename'],
                               buffer.attrs['realname']
                               )
            buffers.append('%s: reloaded' % buffer.name)
         else:
            buffers.append('%s: up-to-date' % repr(buffer.name))
      except Abort, ex:
         buffers.append('%s: error %s' % (repr(buffer.name), repr(str(ex))))
   request.sendReply('\n'.join(buffers))

def remoteReadFile(request):
   filename = request.args.get('filename')
   if not filename:
      request.sendReply('error: readFile requires a filename argument')
   
   # if the file is buffered, use the contents of the buffer
   app = getApp()
   buf = app.top and app.top.findBuffer(realname = filename)
   if buf:
      # if the file is buffered, we pretty much always want to get from the 
      # buffer. XXX not sure how best to deal if the file is 
      # more recent than the buffer
      contents = buf.get('1.0', 'end')
      request.sendReply(contents)
   
   try:
      contents = app.filer.read(filename)
      request.sendReply(contents)
   except IOError, ex:
      request.sendReply('error: %s' % ex)

def remoteWriteFile(request):
   filename = request.args.get('filename')
   if not filename:
      request.sendReply('error: writeFile requires a filename argument')
   
   mode = request.args.get('mode')
   if mode is not None:
      try:
         mode = int(mode)
      except ValueError:
         request.sendReply('error: writeFile mode argument must be an '
                            'integer.'
                           )

   try:
      # write the file
      app = getApp()
      statInfo = app.filer.write(filename, request.body, mode)
      
      # if the file is buffered, we have to store the new contents in the 
      # buffer XXX what if the buffer is more recent than the file?
      top = app.top
      buf = top and top.findBuffer(realname = filename)
      if buf:
         buf.setAutoSeparators(False)
         buf.delete('1.0', 'end')
         buf.insert('end', request.body)
         buf.setAutoSeparators(True)
         buf.setModified(False)
         buf.attrs['mtime'] = statInfo[stat.ST_MTIME]
         buf.attrs['mode'] = statInfo[stat.ST_MODE]
      
      request.sendReply('ok %s' % repr(os.stat(filename)))
   except IOError, ex:
      request.sendReply('error: %s' % ex)

def remoteListdir(request):
   dir = request.args.get('dir')
   if dir is None:
      dir = os.getcwd()
   
   # build a dictionary mapping the filename to its "stat" info
   entries = {}
   for entry in os.listdir(dir):
      try:
         entries[entry] = os.lstat(os.path.join(dir, entry))
      except OSError:
         # lstat fails - the entry must no longer exist
         pass
      
   request.sendReply(pprint.pformat(entries))

def remoteStat(request):
   filename = request.args.get('filename')
   if not filename:
      request.sendReply('error: stat requires a filename argument')
      return
   
   try:
      # XXX need to 1) use filer.stat() and 2) mutate the mtime with the 
      # _time of the last modification of the buffer_ (which we don't 
      # currently store!)
      statinfo = os.stat(filename)
   except OSError:
      request.sendReply('enoent')
      return
   
   request.sendReply('ok %s' % repr(statinfo))

def remoteGlob(request):
   pattern = request.args.get('pattern')
   if not pattern:
      request.sendReply('error: glob requires a pattern argument')
      return
   
   results = []
   for file in glob.glob(pattern):
      results.append(file)
   request.sendReply('\n'.join(results))

remoteCommands = {
   'edit': LoadFileCmd(None).processAsRemote,
   'ping': remotePing,
   'ls': remoteLs,
   'saveAll': remoteSaveAll,
   'reloadAll': remoteReloadAll,
   'readFile': remoteReadFile,
   'writeFile': remoteWriteFile,
   'listDir': remoteListdir,
   'stat': remoteStat,
   'glob': remoteGlob,
}
