#==============================================================================
#
#  $Id$
#
"""
    Application main module.  
"""
#
#   Copyright (C) 2007 Michael A. Muller
#   Copyright (C) 2009 Google Inc.
#
#   This file is part of SpugEdit.
#
#   SpugEdit is free software: you can redistribute it and/or modify it under
#   the terms of the GNU Lesser General Public License as published by the
#   Free Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   SpugEdit is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with SpugEdit.  If not, see <http://www.gnu.org/licenses/>.
#
#==============================================================================

import sys, os, traceback, getopt
from pickle import Unpickler
from types import ClassType
from Tkinter import Tk, Text
from core import Command, CommandFactory, initializers, YesNoHandler, \
    YesNoPrompt
from yed import AlertOutput, createBufferFromFile, getBufferInitializer, \
    getApp, LoadFileCmd, remoteCommands, saveEditorState, \
    selectBufferFromFilename, showBufferInfo, setApp, Top, topBindings, \
    YedletCommandFactory
from comm import Communicator, getRemote, getTCPRemote, BasicAuthenticator, \
    TimeoutError
from filer import RemoteFiler, LocalFiler, CompoundFiler
import shell
import log
import buf.exo
import buf.shell
import buf.cxx
import buf.term
import buf.python
from buf.python import createPythonCommandBuffer, PythonBufferType
import buf.help

class MountCommand(Command):
    """
        mount <mount-point> <host:port> <auth-token>

        Mount another instance of yed.  This will have two effects:
        -   mount the remote instance's filesystem on "mount-point"
        -   reroute commands sent to the remote instance to this instance.
            (XXX not implemented yet)
    """
    
    def process(self, context):
        if len(context.args) != 3:
            context.top.showError('Usage: mount <mount-point> <host:port> '
                                   '<auth-token>'
                                  )
            return
        
        mountPoint, addr, token = context.args
        
        # parse the remote host:port
        try:
            host, port = addr.split(':')
            port = int(port)
        except ValueError, ex:
            context.top.showError('Invalid remote id: %s' % repr(addr))
        
        app = getApp()
        remote = getTCPRemote(app, host, port)
        remote.authToken = token
        filer = RemoteFiler(remote)
        
        try:
            app.filer.mount(mountPoint, filer)
        except KeyError, ex:
            context.top.showError('Already mounted: %s' % ex[0])

class UnmountCommand(Command):
    """
        umount <mount-point>
        
        Unmount a filesystem mounted with the mount command.
    """
    def process(self, context):
        if len(context.args) != 1:
            context.top.showError('Usage: mount <mount-point>')
            return
        
        mountPoint, = context.args
        app = getApp()
        try:
            app.filer.unmount(mountPoint)
        except KeyError:
            context.top.showError('%s is not mounted' % mountPoint)

class YedletCommand(Command):
    """
        yedlet [-ide] <name>
        
        Create or edit a yedlet.
        
        Options:
            -i  Refer to the instance directory instead of the global 
                yedlet directory.
            -d  Delete the specified yedlet.
            -e  Bring up the edit file command.
    """

    INITIAL_CONTENTS = '# -*- Python -*-'    

    class Deleter(YesNoHandler):
        
        def __init__(self, top, filename):
            self.__top = top
            self.__filename = filename
        
        def handleYes(self):
            os.remove(self.__filename)
        
        def handleNo(self):
            self.__top.showError('Delete aborted')

    def process(self, context):
        
        # parse options
        inst = False
        delete = False
        edit = False
        opts, args = getopt.gnu_getopt(context.args, 'ide')
        for opt, arg in opts:
            if opt == '-i':
                inst = True
            elif opt == '-d':
                delete = True
            elif opt == '-e':
                edit = True

        # get the instance yedlet dir or the global yedlet dir
        if inst:
            dir = getApp().getYedletDir(True)
        else:
            dir = getApp().getYedletDir()

        if edit and delete:
            context.top.showError('-e and -d cannot be used together')
            return
        
        if edit:
            if args:
                context.top.showError('"yedlet -e" accepts no additional '
                                       'arguments'
                                      )
                return
            
            LoadFileCmd.load(context, dir + '/', PythonBufferType(), 
                             self.INITIAL_CONTENTS
                             )
            return
        
        if delete:
            if len(args) != 1:
                context.top.showError('usage: yedlet -d <name>')
                return
            yedlet = os.path.join(dir, args[0])
            
            if not os.path.exists(yedlet):
                context.top.showError('%s does not exist' % repr(yedlet))
                return
            
            prompt = YesNoPrompt(context.top, 'Delete %s?' % repr(yedlet),
                                 self.Deleter(context.top, yedlet)
                                 )
            context.top.addStatusWidget(prompt)
            return

        if len(args) != 1:
            context.top.showError('usage: yedlet <name>')
            return

        name = args[0]
        fullname = os.path.join(dir, name)
        selectBufferFromFilename(context.top, fullname,
                                 bufferType = buf.python.PythonBufferType(),
                                 initialContents = self.INITIAL_CONTENTS
                                 )

class YedApp(dict):
    """
        Bundles up the application in a nice clean package.
        
        Derives from dict to provide a dictionary interface for global 
        paramters.
    """
    
    def __init__(self, files, instanceName = 'Default', block = False,
                 serve = False,
                 master = None,
                 authToken = None,
                 commands = None,
                 captureStderr = True,
                 yedDir = None
                 ):
        dict.__init__(self)
        self.__instanceName = instanceName
        self.__yedlets = {}
        self.__serve = serve
        self.__createdYedDir = False
        self.__yedDir = yedDir
        if yedDir is None:
            self.__yedDir = os.path.expanduser('~/.yed')
        
        # register yourself - this must be done prior to creating any external
        # components
        setApp(self)
        
        # create the communicator
        self.auth = BasicAuthenticator()
        self.communicator = Communicator(remoteCommands, self.auth)

        # see if we are in remote mode
        try:
            self.__remote = getRemote(self)
        except TimeoutError:
            self.communicator.stop()
            raise
        
        # now we decide what mode we've ended up in and what we're going to do

        if files and self.__remote:
            print 'adding files to remote instance'
            self.communicator.start()
            try:
                for file in files:
                    self.__remote.edit(os.path.abspath(file), block)
                return
            finally:
                self.__remote.close()
                self.communicator.stop()

        elif self.__remote:

            # if we're in remote mode, execute commands
            self.communicator.start()
            try:
                # make sure we got some commands
                if not commands:
                    print 'Yed instance "%s" is already running.  To add ' \
                          'files to it, "yed <filename> ...".  To start a ' \
                          'new instance, "yed -i <instance-name>".' % \
                          instanceName
                    return

                for cmd in commands:
                    if cmd == 'ls':
                        print self.__remote.ls()
                    elif cmd == 'saveAll':
                        print self.__remote.saveAll()
                    elif cmd == 'reloadAll':
                        print self.__remote.reloadAll()
                    elif cmd == 'getToken':
                        print self.__remote.getToken()
                    elif cmd == 'readFile':
                        print self.__remote.readFile('yed')
                    else:
                        assert False, 'Unimplemented command: %s' % cmd
            finally:
                self.__remote.close()
                self.communicator.stop()

            return

        elif not self.__serve:
            
            # make sure that we didn't get any commands - that would mean the 
            # user tried to send remote commands and didn't find a remote 
            # object
            if commands:
                assert False, 'Unable to connect to remote server'
            
            # if a remote master was specified, override the default file 
            # interface.
            if master:
                host, port = master.split(':')
                port = int(port)
                self.communicator.start()
                remote = getTCPRemote(self, host, port)
                if authToken:
                    remote.authToken = authToken
                self.filer = RemoteFiler(remote)
            else:
                self.filer = CompoundFiler()
                self.filer.mount('/', LocalFiler())
            
            self.loadExtensions()
            
            # try to create from the last statefile
            self.top = self.createEditor()

            self.top.update()

            log.trace('Initializing commands\n')
            self.__initCommands()

            # begin redirection of standard error
            self.__orgStderr = sys.stderr
            if captureStderr:
                sys.stderr = AlertOutput(self.top, createPythonCommandBuffer, 
                                        {'isPythonCommandBuffer': True},
                                        'error',
                                        lambda text: 
                                            text.mark_set('execstart', 
                                                          'insert'
                                                          )
                                        )
            
            # add all files specified on the command line to the editor
            for file in files:
                self.console.insert('end', 'Loading %s\n' % file)
                self.top.update()
                # resolve the realpath of the file, see if it already is 
                # loaded
                realname = os.path.realpath(file)
                buffer = self.top.findBuffer(realname = realname)
                if buffer:
                    self.top.focus(buffer)
                    showBufferInfo(self.top, buffer)
                else:
                    buffer = createBufferFromFile(self.top, file)
            
            # if we ended up with no buffers in the toplevel, create a help 
            # screen (if this is the first instantiation) or a python 
            # command line
            if not self.top.buffers:
                if self.__createdYedDir:
                    log.trace('creating help buffer')
                    buf.help.showHelpBuffer(self.top)
                else:
                    log.trace('Creating command line buffer')
                    buffer = createPythonCommandBuffer(self.top)
                    self.top.focus(buffer)
            
            # disable console logging and destroy the console.
            log.setTracer(lambda msg: None)
            self.console.destroy()
            
            # need to save the editor state if we changed it.  We kick this 
            # off in idle time because part of what we're saving is the window 
            # geometry, and that doesn't exist until the windows is mapped.
            if files: 
                self.top.after_idle(lambda x = self: saveEditorState(x.top))

        else:
            # server mode - initialize the extensions and the filer
            self.top = None
            self.loadExtensions()
            self.filer = CompoundFiler()
            self.filer.mount('/', LocalFiler())

        self.communicator.startServers()

    def loadExtensionsFrom(self, path):
        """
            Loads all extensions from the specified path.  Extensions are 
            subdirectories with a "init" file, which is a simple python script 
            to be executed using execfile().
            
            While running the init script, the module path is temporarily 
            modified to the "lib" subdirectory of the extension.  Any heavy 
            lifting should be performed in modules in this directory, since 
            modules don't have to be compiled and are inherently more 
            efficient than scripts.
            
            BTW, since we are mucking about with global variables, this method 
            is inherently not threadsafe.  It should only be run during 
            initialization from the main thread.
        """
        if os.path.isdir(path):
            for extDir in os.listdir(path):
                extName = extDir
                extDir = os.path.join(path, extDir)
                initFile = os.path.join(extDir, 'init')
                
                # see if we've got an init file
                if os.path.isfile(initFile):
                    
                    # add the library path, execute, and remove the library 
                    # path
                    sys.path.insert(0, os.path.join(extDir, 'lib'))
                    try:
                        execfile(initFile, 
                                 {'getApp': getApp, 'extName': extName})
                    except:
                        traceback.print_exc()
                    sys.path.pop(0)
                        
                else:
                    print '%s: bad or missing init file' % extDir

    def loadExtensions(self):
        """
            Load all extensions, first the global ones and then the instance 
            specific ones.
        """
        self.loadExtensionsFrom(os.path.join(self.getYedDir(), 'ext'))
        self.loadExtensionsFrom(os.path.join(self.getInstDir(), 'ext'))
    
    def __initCommands(self):
        """
            Initialize all of the application-level commands.
        """
        # XXX move mount/umount into some different module.
        if isinstance(self.filer, CompoundFiler):
            self['mount'] = MountCommand(None)
            self['umount'] = UnmountCommand(None)
            self['yedlet'] = YedletCommand(None)
        for initializer in initializers:
            initializer()
            
    def getYedDir(self):
        """
            Gets the editor's state directory, returns the full path to it.            
        """
        path = self.__yedDir
        
        if not os.path.exists(path):
            os.mkdir(path)
            self.__createdYedDir = True
        
        return path
        
    def getInstDir(self):
        """
            Creates the "instance" directory for the editor - multiple yed 
            instances can run for a single users.
            
            Creates the directory if it does not already exist.
            
            Returns the full path to this directory.            
        """
        yedDir = self.getYedDir()

        # make sure there is an "instances" directory
        path = os.path.join(yedDir, 'instances')
        if not os.path.exists(path):
            os.mkdir(path)
        
        # make sure there is a subdirectory for the current instance
        path = os.path.join(path, self.__instanceName)
        if not os.path.exists(path):
            os.mkdir(path)
        
        return path
    
    def getYedletDir(self, useInstDir = False):
        """
            Returns the full path to the yedlet directory.  Creates this 
            directory if it does not exist.
            
            parms:
                useInstDir: [bool] if true, use the instance yedlet directory 
                    instead of the global yedlet directory.
        """
        if useInstDir:
            baseDir = self.getInstDir()
        else:
            baseDir = self.getYedDir()

        yedletDir = os.path.join(baseDir, 'yedlet')
        if not os.path.exists(yedletDir):
            os.mkdir(yedletDir)
        
        return yedletDir
    
    def getInstName(self):
        """
            Returns the instance name.
        """
        return self.__instanceName

    def createEditor(self, stateFile = None):
        """
            Creates the editor from the statefile, if it exists.
        """
        if not stateFile:
            stateFile = os.path.join(self.getInstDir(), 'state')

        # create the toplevel and set its title
        tk = Tk()
        top = Top(tk, topBindings, self)
        top.winfo_toplevel().title('Yed - instance %s' % 
                                   self.__instanceName
                                   )
        
        # create a special console window to display initialization messages 
        # and hide the main panels while they are under construction.
        self.console = Text(tk)
        self.console.grid(row = 0, column = 0)

        # redirect all log.trace() output to the console        
        def preTrace(msg):
            self.console.insert('end', msg + '\n')
            self.console.see('end')
            top.update()
        log.setTracer(preTrace)
        log.trace('Loading Yed...\n')
    
        if os.path.exists(stateFile):
            
            try:
                src = open(stateFile, 'rb')
                assert Unpickler(src).load() == '0'
                #print 'loaded 0'
                top.readFrom(src)
            except Exception:
                buf = top.createBuffer('Error')
                buf.insert('1.0', '''
Error Loading Statefile

There was an error loading the file %s, which yed uses to store its state
information.  The simplest fix to this is simply to delete the file, although
you will lose your state information.  The traceback follows:

%s''' % (stateFile, traceback.format_exc()))
                getBufferInitializer(buf)(top, buf)
    
        return top

    def handlePoll(self):
        """
            Method to poll the communicator thread.
        """
        self.communicator.processUIQueue(self.top)
        self.top.after(250, self.handlePoll)
    
    def handleShutdown(self):
        """
            Called immediately prior to shutting down the toplevel.  Allows us 
            to restore stderr.
        """
        sys.stderr = self.__orgStderr
        
    def run(self):

        if self.__serve:
            # if we're in "serve" mode, just do the UI queue
            try:
                self.communicator.start()
                self.communicator.runUIQueue()
            except KeyboardInterrupt:
                self.communicator.stop()
                raise

        # don't have to run anything if we are communicating to a remote 
        # instance
        elif not self.__remote:
            self.communicator.start()
            self.top.after(250, self.handlePoll)
            self.top.mainloop()
            self.communicator.stop()
    
    def __gotYedlet(self, name):
        yedlet = self.__yedlets.get(name)
        if yedlet:
            return yedlet
        
        # check the instance yedlet dir, then the global one
        for path in (self.getYedletDir(True), self.getYedletDir()):
            if os.path.exists(os.path.join(path, name)):
                yedlet = self.__yedlets[name] = \
                    YedletCommandFactory(path, name)
                return yedlet
        else:
            return None
    
    def lookupCommand(self, commandName, context):
        """
            Looks up a name (using normal dictionary lookup semantics).  If 
            the name is bound to a command, returns the command.  Otherwise 
            returns *None*.
        """
        cmd = self.get(commandName)
        if isinstance(cmd, Command):
            return cmd
        elif isinstance(cmd, ClassType) and issubclass(cmd, Command) or \
             isinstance(cmd, CommandFactory):
            return cmd(context.top)
        elif self.__gotYedlet(commandName):
            return self.__yedlets[commandName](context.top)
        else:
            return None

usage = '''Yed - one crazy-assed python editor.
Options:
    -b, --block
        When loading files into an existing instance, block for as long as the 
        files are buffered in the editor.  This should be used when yed is 
        assigned to the EDITOR variable because most commands that use EDITOR 
        expect the editor to block until after the user has modified the file.
    -i <instance>, --instance=<instance>
        Specify the instance name.  If an editor session is already running 
        under the instance name, all actions will be applied to the running 
        instance.  If not, a new editor sesion will be created.
    -s, --serve
        Run in dedicated server mode - no user interface will be created, 
        slaves may connect to the server instance with the --master flag.
    -m <host:port>, --master=<host:port>
        Connect to the instance running on host:port.
    -a <auth-token>, --auth=<auth-token>
        Specify an authentication token (for use with the --master flag to 
        securely connect to a remote instance over TCP/IP)
    -t, --getToken
        Generate a new, random authentication token (for use with the --auth
        flag).  This token can be used to remotely connect to the yed instance 
        over TCP/IP, so be careful with it.
    -S, --saveAll
        Save all modified buffers in the running instance.
    -L, --reloadAll
        Reload all buffers that have changed on disk in the running instance.
    -l, --ls
        List all of the buffers in the running instance.
    --readFile
        Read a remote file and write it to output.  (this was a test function, 
        and not very useful)
    -e <true|false>, --captureStderr=<true|false>
        By default, yed captures all standard error output that it generates 
        and writes it to a python command buffer (creating the buffer on 
        demand if it does not exist).  Setting this flag to false disables 
        this and lets standard error output go to standard error.  Values of 
        "yes" and "no" also work.
    -y <yed-dir>, --yedDir=<yed-dir>
        The directory where yed stores all of its states and instances - 
        "$HOME/.yed" by default.
    -h, --help
        Show this help text.
'''

def run(args):
    instanceName = 'Default'
    block = False
    serve = False
    authToken = None
    master = None
    captureStderr = True
    yedDir = None
    
    # check for environment variables
    envInstName = os.environ.get('YED_INST')
    if envInstName:
        instanceName = envInstName
    
    # parse the args
    commands = []
    opts, args = getopt.gnu_getopt(args[1:], 'bi:sm:a:tSLle:hy:', 
                                   ['block=', 'instance=', 'serve', 
                                    'master=',
                                    'auth=',
                                    'getToken',
                                    'saveAll',
                                    'reloadAll',
                                    'ls',
                                    'readFile',
                                    'captureStderr=',
                                    'help',
                                    'yedDir=',
                                    ])
    for opt, arg in opts:
        if opt in ('-i', '--instance'):
            instanceName = arg
        elif opt in ('-b', '--block'):
            block = True
        elif opt in ('-s', '--serve'):
            serve = True
        elif opt in ('-m', '--master'):
            master = arg
        elif opt in ('-a', '--auth'):
            authToken = arg
        elif opt in ('-t', '--getToken'):
            commands.append('getToken')
        elif opt in ('-S', '--saveAll'):
            commands.append('saveAll')
        elif opt in ('-L', '--reloadAll'):
            commands.append('reloadAll')
        elif opt in ('-l', '--ls'):
            commands.append('ls')
        elif opt == 'readFile':
            commands.append('readFile')
        elif opt in ('-e', '--captureStderr'):
            if arg.lower() not in ('yes', 'true'):
                captureStderr = False
        elif opt in ('-h', '--help'):
            print usage
            sys.exit(0)
        elif opt in ('-y', '--yedDir'):
            yedDir = arg

    YedApp(args, instanceName = instanceName, block = block,
           serve = serve,
           master = master,
           authToken = authToken,
           commands = commands,
           captureStderr = captureStderr,
           yedDir = yedDir
           ).run()
