#==============================================================================
#
#  $Id$
#
"""
   Buffer type definitions for C/C++/Java buffers.
"""
#
#  Copyright (C) 2008 Google Inc.
#
#   This file is part of SpugEdit.
#
#   SpugEdit is free software: you can redistribute it and/or modify it under
#   the terms of the GNU Lesser General Public License as published by the
#   Free Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   SpugEdit is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with SpugEdit.  If not, see <http://www.gnu.org/licenses/>.
#
#==============================================================================

import os, re
from yed.core import Command, RegionConfig, registerBufferType
from yed import CLikeParser, getApp, ParsingRules
from default import DefaultBufferType

class CxxParser(CLikeParser):
   tokens =  [
      ('preproc', re.compile(r'#.*'), None),
      ('comment', re.compile(r'//.*|/\*([^\*]|\*[^/])*\*/'), None),
      ('comment', re.compile(r'/\*'), re.compile(r'.*\*/')),
   ] + CLikeParser.tokens

class CxxRules(ParsingRules):
   """
      Rules for C++ buffers.
   """
   
   def getParserClass(self):
      return CxxParser

   __commentRx = re.compile(r'\s*//\s*')
   __commentContinuationRx = re.compile(r'(\s*)\*(/)?\s*')
   __multiLineCommentRx = re.compile(r'(\s*)/\*([^\*]|\*[^/])*')
   def getCommentPrefix(self, line, hardReturn):
      
      # continue single line comments if we are in "soft return" mode
      m = self.__commentRx.match(line)
      if m and not hardReturn:
         return m.group()
      
      # continue continuation comments always
      m = self.__commentContinuationRx.match(line)
      if m:
         # if this is the comment terminator ("*/"), we want to just give the 
         # whitespace
         if m.group(2):
            return m.group(1)[:-1]
         else:
            return m.group()
      
      # if we started a multi-line comment, the next line is prefixed by
      # an indented asterisk plus one space.
      m = self.__multiLineCommentRx.match(line)
      if m:
         return m.group(1) + ' * '
      
      return None
   
   class MkClassCmd(Command):
      def process(self, context):
         buf = context.buffer
         
         # absent a namespace, we just use the basename of the file
         basename = os.path.basename(buf.attrs['filename'])
         className = os.path.splitext(basename)[0]
         fullName = [className]
         
         # create a bunch of variables that can be expanded in the template
         templateName = os.path.join(getApp().getYedDir(), 'etc', 'header.c')
         underscoredName = '_'.join(fullName)
         parms = {
            'year': time.strftime('%Y', time.localtime(time.time())),
            'template': templateName,
            'basename': basename,
            'className': className,
            'upperCaseIdempotent': underscoredName.upper(),
            'mixedCaseIdempotent': underscoredName,
         }
         
         # if the template exists, expand it.  Otherwise just create something 
         # simple.
         if getApp().filer.isfile(templateName):
            template = open(templateName).read() % parms
         else:
            template = '''// create %(template)s to get a nicer comment here
#ifndef %(upperCaseIdempotent)s
#define %(upperCaseIdempotent)s

class %(className)s {

};

#endif
'''
         # expand the template into the beginning of the buffer
         buf.insert('1.0', template % parms)
         
   def lookupCommand(self, cmd, context):
      if cmd == 'mkclass':
         return self.MkClassCmd(context.top)

class CxxBufferType(DefaultBufferType):
    name = 'cxx'
    extensions = ['.cc', '.h', '.c', '.C', '.cpp', '.java', '.H']
    firstLinePatterns = [
        re.compile('.*-\*-.*c.*-\*\-', re.I),
        re.compile('.*-\*-.*java.*-\*\-', re.I),
    ]
    config = RegionConfig(DefaultBufferType.config, rules = CxxRules())

registerBufferType(CxxBufferType())
