#==============================================================================
#
#  $Id$
#
"""
   Buffer type definitions for exo buffers.
"""
#
#  Copyright (C) 2008 Google Inc.
#
#   This file is part of SpugEdit.
#
#   SpugEdit is free software: you can redistribute it and/or modify it under
#   the terms of the GNU Lesser General Public License as published by the
#   Free Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   SpugEdit is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with SpugEdit.  If not, see <http://www.gnu.org/licenses/>.
#
#==============================================================================

import re, os
from yed.core import Command, registerBufferType, RegionConfig, \
    registerInitializer
from yed import selectBufferFromFilename
from yed.buf.shell import ShellRules
from default import DefaultBufferType

import yed.shell
from yed import defaultInitializer, showBufferInfo, getApp, ReturnCmd

wsRx = re.compile(r'\s+')
commentRx = re.compile(r'#.*')
wordRx = re.compile(r'([^"\'\s]+) | (\' ([^\\\'] | \\.)* \') | '
                    '(" ([^\\"] | \\.)* ")', re.X)
annotationRx = re.compile(r'\[ (?P<stateChange>@)? (?P<bufId>\d+)? '
                           '(?P<noCache>\*)? \]\s*', 
                          re.X
                          )

# since this follows all other pattern matches, we don't have to check for any 
# escape sequences, we can just consume the rest of the string
incompleteWordRx = re.compile(r'(".*) | (\'.*)', re.X)

class ExoCommand:
    """
        Encapsulates a single exo command.
    
        Public-vars:
            begin: [str] the index of the beginning of the command line in the 
                buffer.
            end: [str] the index of the end of the command line in the buffer.
            ann: [@Annotation] the command annotation.
            words: [list<str>] the command words.
    """
    
    def __init__(self, buf, begin, endAnn, ann, words):
        self.__buf = buf
        self.begin = begin
        self.endAnn = endAnn
        self.ann = ann
        self.words = words

    @staticmethod
    def parseLine(line):
        """
            Parse the line into words.
           
            parms: 
                line: [str] a command line

            returns:
                [list<str>] 'line' tokenized into words.
        """
        words = []
        complete = True
        while line:
            m = wsRx.match(line)
            if m:
                line = line[m.end():]
                continue
            
            # check for a comment
            if commentRx.match(line):
                break
            
            m = wordRx.match(line)
            if m:
                # process quoted strings
                word = m.group(0)
                if word[0] in ('"', "'"):
                    word = eval(word)
                words.append(word)
                line = line[m.end():]
                continue
            
            m = incompleteWordRx.match(line)
            if m:
                # construct a complete quoted string and safely un-quote it.
                word = m.group(0)
                quote = word[0]
                word = eval(word + quote)
                words.append(word)
                complete = False
                break
        
        return complete, words
        
    @staticmethod
    def read(buffer, pos, emptyOk = False):
        """
            Reads the next command from the buffer.
            
            parms:
                buffer: [@yed.core.Buffer]
                pos: [str] the position in the buffer
                emptyOk: [bool] if true, it is acceptable to return without a 
                    command
            returns:
                [tuple<@ExoCommand or None, str>] the command to be read and 
                the index of the buffer position after the command.  Returns 
                None instead of a command if we've run out of buffer.
        """
        ann = None
        words = None
        
        # since pos can end up in the command, normalize it.
        pos = buffer.index(pos)
        
        # scan until we actually get a command
        while not words:

            # make sure we're not out of buffer
            if buffer.compare(pos, '==', 'end'):
                return None, pos
            
            # get the current line, forward pos to the beginning of the next 
            # one
            line = buffer.get(pos, '%s lineend' % pos)
            lastPos = pos
            pos = buffer.index('%s + 1 lines' % pos)
            
            # if we haven't got an annotation yet, check for one
            if not ann:
                begin = lastPos
                m = annotationRx.match(line)
                if m:
                    ann = Annotation.fromPatternMatch(m)
                    endAnn = buffer.index('%s + %d chars' % (begin, m.end()))
                    line = line[m.end():]
                else:
                    endAnn = begin

            # try to parse a line            
            complete, words = ExoCommand.parseLine(line)
            if not complete:
                raise Exception('inomplete command: %s' % repr(line))
            
            # leave if we don't need to get anything
            if emptyOk:
                break

        cmd = ExoCommand(buffer, begin, endAnn, ann or Annotation(), words)
        return cmd, pos
    
    def rewrite(self):
        """
            Rewrite the annotation in the buffer that it came from.
        """
        self.__buf.delete(self.begin, self.endAnn)
        self.__buf.insert(self.begin, self.ann)

def iterCommands(buf, endPos = 'end'):
    """
        Iterate over the commands in the buffer.
        
        parms:
            buffer: [@yed.core.Buffer]
            endPos: [str] index of the end point of the iteration
    """
    cur = '1.0'
    while buf.compare(cur, '<', endPos):
        cmd, cur = ExoCommand.read(buf, cur)
        if cmd is None:
            break
        yield cmd
        

class Annotation:
    
    def __init__(self, bufId = None, state = False, noCache = False):
        self.bufId = bufId
        self.state = state
        self.noCache = noCache

    @staticmethod
    def fromPatternMatch(match):
        return Annotation(match.group('bufId'),
                          match.group('stateChange'),
                          match.group('noCache')
                          )
                              
    def __str__(self):
        if self.state:
            return '[@] '
        elif self.bufId is not None:
            return '[%d%s] ' % (self.bufId, self.noCache and '*' or '')
        else:
            return '[%s] ' % (self.noCache and '*' or '')

def getNextBufId():
    app = getApp()
    bufId = app.get('exoBufId', 0)
    app['exoBufId'] = bufId + 1
    return bufId

def execStateCommand(cmd, context):
    """
        Try to execute a state command.
        
        returns:
            True if a state command was actually found, false if not.
    """
    if cmd.words[0] == 'cd':        
        cwd = context.buffer.getDirectory()
        context.buffer.attrs['cwd'] = \
            os.path.join(cwd, os.path.expanduser(cmd.words[1]))
        return True
    else:
        return False

def execBuiltinCommand(cmd, context):
    """
        Tries to execute a built-in command (if there is one to be executed).
        
        parms:
            cmd:  [@ExoCommand] the command
            context: [@yed.core.CommandContext] the command context.
        
        returns:
            [@Annotation] the command annotation.
    """
    if cmd.words[0] == 'yed':
        filename = os.path.join(context.buffer.getDirectory(), cmd.words[1])
        buffer = selectBufferFromFilename(context.top, filename)
        bufId = getNextBufId()
        buffer.nattrs['exoBufId'] = bufId
        buffer.attrs['parent' ] = context.buffer.name
        return Annotation(bufId = bufId)
    else:
        return None        
    
def traverseCommand(cmd, context, outBuf):
    """
        parms:
            cmd: [@ExoCommand] the command
            context: [@yed.core.CommandContext] the command context.
            outBuf: [@yed.core.Buffer] the buffer for the output of the 
                command.
        returns: [@Annotation] the command annotation
    """
    if execStateCommand(cmd, context):
        return Annotation(state = True)
    else:
        # reread all state command prior to the current command
        for prevCmd in iterCommands(context.buffer, 'insert linestart'):
            # XXX breaks if somebody changes the state command.
            if prevCmd.ann.state:
                execStateCommand(prevCmd, context)

        # see if this is a built-in command
        ann = execBuiltinCommand(cmd, context)
        if ann:
            return ann

        if not outBuf:
            # create the buffer
            outBuf = yed.shell.createTermBuffer(context.top, 
                                                'shell ' + 
                                                 ' '.join(context.args)
                                                )
            
            # get the next buffer id, store it.
            bufId = getNextBufId()
            outBuf.nattrs['exoBufId'] = bufId
            outBuf.attrs['parent'] = context.buffer.name
        else:
            # get the buffer id of the new buffer
            bufId = outBuf.nattrs['exoBufId']
        
        # insert the command itself at the top
        outBuf.insert('1.0', '$ ' + ' '.join(cmd.words), 'meta', '\n')
        
        # change to the appropriate working directory
        os.chdir(context.buffer.getDirectory())
            
        yed.shell.runProc(cmd.words, context, buffer = outBuf)        
        return Annotation(bufId = bufId)

class ExoReturnCmd(Command):
    
    def process(self, context):
        cmd, nextLine = ExoCommand.read(context.buffer, 'insert linestart', 
                                        True
                                        )
        
        # do a normal return if we didn't get anything
        if not cmd or not cmd.words:
            return ReturnCmd(context.top).process(context)
        
        # if we're at the end of the buffer, do a normal newline
        if context.buffer.compare(nextLine, '==', 'end'):
            ReturnCmd(context.top).process(context)

        # see if the buffer exists
        ann = cmd.ann
        buffer = None
        if ann.bufId:
            buffer = context.top.findBuffer(exoBufId = int(ann.bufId))
            if buffer:
                context.top.focus(buffer)
                
                # if the annotation allows caching, we're done.
                if not ann.noCache:
                    return 'break'
                
                # no caching - clear out the buffer
                buffer.delete('1.0', 'end')
        
        # we can't use a cached buffer - we have to execute the command
        ann = traverseCommand(cmd, context, buffer)
        ann.noCache = cmd.ann.noCache
        
        # change the annotation
        context.buffer.delete(cmd.begin, cmd.endAnn)
        context.buffer.insert(cmd.begin, ann)
        
        # go to the start of the next line
        context.buffer.mark_set('insert', nextLine)
        return 'break'

class ExoCreateCmd(Command):
    """
        Create an exo buffer homed in the current working directory of the 
        current buffer.
    """
    
    def process(self, context):
        cwd = context.buffer.getDirectory()
        filename = os.path.join(cwd, '.exo')
        selectBufferFromFilename(context.top, filename)

class ExoRules(ShellRules):
    def initialize(self, buffer):
        
        # remove all buffer numbers - these are not relevant any more
        for cmd in iterCommands(buffer):
            if cmd.ann.bufId:
                cmd.ann.bufId = None
                cmd.rewrite()
                
        # remove all memory of the modifications that we made.
        buffer.clearUndoStack()
        buffer.setModified(False)
        
        ShellRules.initialize(self, buffer)
            
_config = RegionConfig(
    DefaultBufferType.config,
    rules = ExoRules(),
)
_config.bindings['<Return>'] = ExoReturnCmd

class ExoBufferType(DefaultBufferType):
    
    name = 'exo'
    extensions = ['.exo']
    firstLinePatterns = [
        re.compile(r'#\s*yed\s*:\s*type\s+exo')
    ]
    config = _config

registerBufferType(ExoBufferType())

# register the function to register the commands
def initializer():
    getApp()['exo'] = ExoCreateCmd(None)
registerInitializer(initializer)