#==============================================================================
#
#  $Id$
#
"""
   Buffer type definitions for python buffers.
"""
#
#  Copyright (C) 2008 Google Inc.
#
#   This file is part of SpugEdit.
#
#   SpugEdit is free software: you can redistribute it and/or modify it under
#   the terms of the GNU Lesser General Public License as published by the
#   Free Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   SpugEdit is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with SpugEdit.  If not, see <http://www.gnu.org/licenses/>.
#
#==============================================================================

import re
from spug.util import StdError
from yed.core import Command, RegionConfig, registerBufferType, \
    createAttrConfig, Struct
from yed import applyBindings, BufferOutput, CLikeParser, defaultInitializer, \
    ParsingRules, ShellCommentRules, topBindings
from default import DefaultBufferType
import sys # XXX debug

def createPythonCommandBuffer(top):
   buffer = top.createBuffer('Python Command Line')
   PythonCommandBufferType().initialize(buffer, top)
   return buffer

class PythonBufferCmd(Command):
   """
      Go to the python buffer if it exists, create it if it does not.
   """
   
   def process(self, context):
      # look for the python buffer
      for buffer in context.top.buffers:
         if buffer.attrs.get('type') == 'python_cmd':
            context.top.focus(buffer)
            return 'break'
      else:
         # not found: create it.
         buffer = createPythonCommandBuffer(context.top)
         context.top.focus(buffer)
         
      return 'break'

class PythonParser(CLikeParser):
   tokens = [
      ('string', re.compile(r'"""'), re.compile(r'.*?"""')),
      ('string', re.compile(r"'''"), re.compile(r".*?'''")),
   ] + CLikeParser.tokens + [
      ('comment', re.compile(r'#.*'), None)
   ]

class PythonRules(ShellCommentRules):
   def getParserClass(self):
      return PythonParser

class PythonBufferType(DefaultBufferType):
    
    name = 'python'
    extensions = ['.py']
    firstLinePatterns = [
        re.compile(r'#!.*python|.*-\*-.*[Pp]ython.*-\*\-'),
    ]
    config = RegionConfig(DefaultBufferType.config, rules = PythonRules())

class ExecBufCmd(Command):

   def execrange(self, buffer, begin, end):
      try:
         exec buffer.get(begin, end) + '\n' in \
            sys.modules['__main__'].__dict__
      except Exception, ex:
         sys.stderr.write(StdError.errorText(ex) + '\n')         

   def process(self, context):
      range = self.computeRange(context.buffer)
      self.execrange(*(context.buffer,) + range)

      # show the prompt
      context.buffer.insert('end', 'ok', 'prompt', '\n')

      # move the insertion point and the exec start point
      context.buffer.mark_set('insert', 'end')
      context.buffer.mark_set('execstart', context.buffer.index('insert')) 

   def computeRange(self, buffer):
      range = buffer.tag_ranges('sel')
      if not range:
         insertPoint = buffer.index('insert')
         execPoint = buffer.index('execstart')
         
         # if we are after the final prompt, range is from the final promot
         # to the end of the buffer
         if float(insertPoint) > float(execPoint):
            range = (execPoint, 'end')
            buffer.tag_add('exec', range[0], range[1])
         
         # otherwise, the range is the last block of code before the cursor
         else:         
            range = buffer.tag_prevrange('exec', 'insert')
            
      return range   

c = createAttrConfig
class PythonCommandBufferType(PythonBufferType):
    name = 'python_cmd'
    extensions = []
    firstLinePatterns = []
    config = RegionConfig(PythonBufferType.config)
    config.bindings['<F9>'] = ExecBufCmd
    regions = Struct(
        PythonBufferType.regions,
        output = c(foreground = 'green'),
        error = c(foreground = 'red'),
        prompt = c(
            foreground = 'white',
            background = 'midnight blue',
            relief = 'raised',
        ),
    )

    def initialize(self, buffer, top):
        PythonBufferType.initialize(self, buffer, top)
        
        # set some special marks so that we can execute regions until we get 
        # all of the fuzz with region bindings worked out.
        buffer.mark_set('execstart', '1.0')
        buffer.mark_gravity('execstart', 'left')
        
        # mark this is "the python buffer"
        buffer.attrs['isPythonCommandBuffer'] = True
        
        # redirect standard output to it (standard error will already get 
        # redirected to it from App)
        sys.stdout = BufferOutput(buffer, 'output', 
                                    lambda text: text.mark_set('execstart', 'insert')
                                    )
    
    def restore(self, top, name, attrs):
        buf = top.createBuffer('Python Command Line')
        self.initialize(buf, top)
        self._restoreAttrs(buf, name, attrs)

topBindings['<F9>'] = PythonBufferCmd
registerBufferType(PythonBufferType())
registerBufferType(PythonCommandBufferType())
