#==============================================================================
#
#  $Id$
#
"""
    Communications functionality.
"""
#
#   Copyright (C) 2007 Michael A. Muller
#   Copyright (C) 2008 Google Inc.
#
#   This file is part of SpugEdit.
#
#   SpugEdit is free software: you can redistribute it and/or modify it under
#   the terms of the GNU Lesser General Public License as published by the
#   Free Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   SpugEdit is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with SpugEdit.  If not, see <http://www.gnu.org/licenses/>.
#
#==============================================================================

import base64, errno, glob, os, re, socket, time, traceback, urlparse
from threading import Condition, Event, Lock, Thread
from StringIO import StringIO
from Queue import Queue, Empty
from spug.web.cgi import parseQuery, encodeString, parseCookies
from spug.io.proactor import createProactor, UNIXAddress, INETAddress, \
    DataHandler
from spug.io.HTTPServer import HTTPConnectHandler, HandlerResolver, \
    RequestHandler, HTTPDataHandler
from spug.util.log import debug, error

from yed import getApp

class YedHandler(RequestHandler):
    """
        The yed handler takes a request, pushes it onto a queue to communicate
        back to the editor main thread, and then waits for a response.  When
        it gets a response, it serves it back over the connection.
    """
    def __init__(self, req, postHandler):
        """
            parms:
                req: [@spug.io.HTTPServer.Request] the HTTP request
                postHandler: [callable<string> or None] if specified, this 
                    will get called when all of the post data is read and 
                    passed the body of the post request.
        """
        self.__buffer = ''
        self.__closed = False
        self.__code = 200
        self.__msg = 'OK'
        self.__data = ''
        self.__contentLength = int(req.headers.get('content-length', 0))
        self.__postHandler = postHandler
    
    def getCode(self):
        return self.__code
    
    def getMessage(self):
        return self.__msg
    
    def readyToGet(self):
        if self.__buffer: return True

    def get(self, size):
        data = self.__buffer[:size]
        self.__buffer = self.__buffer[size:]
        return data

    def put(self, data):
        self.__data += data
        if len(self.__data) >= self.__contentLength and \
           self.__postHandler:
            self.__postHandler(self.__data)

    def readyToClose(self):
        return not self.__buffer and self.__closed
    
    def write(self, data):
        """
            Writes to the buffer that gets streamed back to the client.
        """
        assert not self.__closed
        self.__buffer += data
    
    def close(self):
        """
            Closes the output stream.  The connection will be closed when
            the remaining contents of the buffer is written back to the
            client.
        """
        self.__closed = True
    
    def writeReply(self, parcel, code = 200, message = 'OK', 
                   contentType = 'text/html'
                   ):
        """
            Writes a complete reply.
            
            parms:
                parcel::
                    [string] the message parcel
                code::
                    The HTTP result code
                message::
                    [string] Message to place in the result line.
                contentType::
                    [string] mime type of the message
        """
        self.__code = code
        self.__msg = message
        self.write('Content-Type: %s\n' % contentType)
        self.write('Content-Length: %d\n\n' % len(parcel))
        self.write(parcel)
        self.close()
            
    def writeError(self, code, message = 'Error', parcel = '', close = True):
        """
            Writes an error response.
            
            parms:
                code::
                    [int] the HTTP error code,
                message::
                    [string] message text.
                parcel::
                    [string] An optional page parcel.  By default, this will
                    be a message page constructed from the error code and
                    message.
                close::
                    [boolean] if true, close the connection after writing.
        """
        if not parcel:
            parcel = '<html><head><title>%d %s</title></head><body>' \
                '<h1>%s %s</h1>' % (code, message, code, message)
    
        self.writeReply(parcel, code, message)

class UIThreadRequest:
    """
        A request to be sent to the user interface thread.
        
        Public-vars:
            args::
                [dict<string, string>] arguments to the request (these are the 
                query parameters for the HTTP request)
            top::
                [@yed.Top] the toplevel window (this is populated when the 
                command is processed by the UI thread)
            body::
                [string or None] body of a post request.
    """
    
    def __init__(self, communicator, handler, command, args):
        """
            parms:
                communicator::
                    [@Communicator]
                handler::
                    [@YedHandler]
                command::
                    [callable<@UIThreadRequest>] a callback function
                    that will be invoked in the UI thread.  It is passed a
                    toplevel window and a dictionary of parameters.  It
                    should return a reply object.
                args::
                    [dict<string, string>] dictionary of query parameters.
        """
        self.__comm = communicator
        self.__command = command
        self.handler = handler
        self.reply = None
        self.args = args
        self.body = None
        self.top = None

    def __call__(self, top):
        """
            This gets called in the UI thread.  It passes control to the 
            "command" function passed in during construction.
        """
        self.top = top
        try:
            self.__command(self)
        except Exception, ex:
            # if no reply was sent, send one now.
            if not self.reply:
                self.sendReply('error(%s)' % repr(str(ex)))
            else:
                print 'error raised after reply was sent:'
                traceback.print_exc()
    
    def sendReply(self, reply):
        """
            Sends a reply message.
            
            parms:
                reply::
                    [any] A reply object.  This must either be a string or be 
                    convertible to a string.  It's string form will be written 
                    as the contents of the HTML page in the reply.
        """
        self.reply = reply
        self.__comm.sendReply(self)
        
class YedResolver(HandlerResolver):
    """
        Returns the appropriate request handler for a given path.
    """
    def __init__(self, comm, uiQueue, commands, auth):
        self.__comm = comm
        self.__uiQueue = uiQueue
        self.__commands = commands
        self.__auth = auth
    
    def __parseCookie(self, req, cookieName):
        
        # if there's a cookie header, parse it
        cookies = req.headers.get('cookie')
        if not cookies:
            return None
        cookies = parseCookies(cookies)
        
        return cookies.get(cookieName)
    
    class __PostHandler:
        """
            This class defers communication of a request to the UI thread 
            until the body of a POST request is received.
        """

        def __init__(self, uiQueue):
            self.request = None
            self.uiQueue = uiQueue

        def __call__(self, data):
            self.request.body = data
            self.uiQueue.put(self.request)

    def getRequestHandler(self, req):
        # for post methods, we need to create a callback that will be called 
        # when the complete data is sent.
        if req.method == 'POST':
            postHandler = self.__PostHandler(self.__uiQueue)
        else:
            postHandler = None
        
        # create the handler
        connection = YedHandler(req, postHandler)
        
        # make sure we have a comlete url
        path = req.path
        if path.startswith('/'):
            path = 'http://localhost' + path
            
        # parse the url
        scheme, location, path, parms, query, frag = \
            urlparse.urlparse(path)

        # parse the query
        try:
            parms = parseQuery(query)
        except ValueError:
            connection.writeError(400, 'Bad query parms')
            return connection
        
        # check for an authentication token in the cookie and the query parms
        authToken = parms.get('auth') or \
            self.__parseCookie(req, 'yed_auth')
        
        # check the path prefix
        if path.startswith('/auth/'):
            method = path[6:]
            
            if method == 'getToken':
                # get token method - authorized user wants a token
                if not self.__auth.authorized(authToken):
                    connection.writeError(306, 'Unauthorized')
                else:
                    token = getApp().auth.createToken()
                    connection.writeReply('ok %s' % token)
            else:
                # XXX implement login authentication
                connection.writeError(404, 'Not Found')
                    
        elif path.startswith('/yed/'):
            method = path[5:]
            
            # see if we have a matching command
            if self.__commands.has_key(method):
                
                # check cookie
                if not self.__auth.authorized(authToken):
                    connection.writeError(306, 'Unauthorized')
                    return connection
                
                # store the request in the UI queue
                request = UIThreadRequest(self.__comm, connection, 
                                          self.__commands[method], 
                                          parms
                                          )
                
                # if this was a post method, give the request to the post 
                # handler - he'll need this when the post is complete
                if req.method == 'POST':
                    postHandler.request = request
                else:
                    # otherwise we just put it on the queue.
                    self.__uiQueue.put(request)
            else:
                connection.writeError(404, 'Not Found')
        else:
            connection.writeError(404, 'Not Found')
        
        return connection

class Error(Exception):
    pass

class InstanceInUse(Error):
    """
        Raised when the instance already appears to be active on a different 
        host.
    """
    pass

class CommError(Error):
    """
        Base class for communication errors.
    """
    pass

class TimeoutError(CommError):
    """
        Raised when a remote call times out.
    """
    pass

class NotConnectedError(CommError):
    """
        Raised when a method is called when the connection is broken.
    """
    pass

class AuthError(CommError):
    """
        Raised when a call fails due to an authentication error.
    """
    pass

class RemoteError(Error):
    """
        Raised when a remote call raises an exception.  This is for 
        application errors.
    """
    pass

def getRemote(app):
    """
        Check for a unix socket in the instance directory and return a 
        Remote instance if one is discovered.  Raises InstanceInUse if there 
        is a socket specified for a different host.
    """
    # check for a socket created by this host
    sockName = _getSockPath()
    if os.path.exists(sockName):
        try:
            remote = app.communicator.createRemote(UNIXAddress(sockName))
            app.communicator.start()
            result = remote.ping()
            if result == 'ok':
                return remote
            return None
        except socket.error, ex:
            if ex[0] != errno.ECONNREFUSED:
                raise
        
        # did not connect or got an invalid response to a ping - remove 
        # the socket and move on
        os.remove(sockName)
    else:
        sockFiles = glob.glob(os.path.join(app.getInstDir(), 'sock*'))
    
        # give a stern warning if multiple socket files are discovered
        if len(sockFiles):
            raise InstanceInUse('Instance is in use: ' + ', '.join(sockFiles))
    
    return None

def getTCPRemote(app, host, port):
    return app.communicator.createRemote(INETAddress(host, port))

def _getSockPath():
    return os.path.join(getApp().getInstDir(), 'sock-' + socket.gethostname())

class Communicator:
    """
        The communication hub.
    """

    def __addConnectHandler(self, commands, auth):
        connectHandler = HTTPConnectHandler(YedResolver(self, self.__uiQueue,
                                                        commands,
                                                        auth
                                                        )
                                            )
        self.__communicators.append(connectHandler)
        return connectHandler
    
    def __init__(self, commands, authenticator):
        """
            parms:
                commands: [dict<string, callable>] command dictionary
                authenticator: [@Authenticator] authenticator used for 
                    insecure channels.
        """
        # create the proactor
        self.proactor = createProactor()
        
        # create the queue for the UI to send things to the comm thread
        self.__commQueue = self.proactor.makeControlQueue(self.__gotMessage)
        
        # create the queue for the comm thread to send things to the UI thread
        self.__uiQueue = Queue()
        
        # list of communicating objects for cleanup
        self.__communicators = []
        
        # background proactor thread
        self.__thread = None
        
        # path to the UNIX socket (filled in if we startServers())
        self.__sockPath = None
        
        self.__commands = commands
        self.__authenticator = authenticator
        
        # mapping from port number to TCP listener
        self.__servers = {}
    
    def createRemote(self, addr):
        """
            Create a new remote object.
            
            parms:
                addr: [spug.io.proactor.Address]
        """
        rem = Remote(self.__uiQueue, DurableSocket(self, addr))
        self.__communicators.append(rem)
        return rem

    def startServers(self):
        # if we've already started, do this from the control queue
        if self.__thread:
            self.__commQueue.add(self.__startServers)
        else:
            self.__startServers()
    
    def __startServers(self):
        # create a unix control socket
        self.__sockPath = _getSockPath()
        sockFile = self.__sockPath
        auth = SecureChannelAuthenticator()
        connectHandler = self.__addConnectHandler(self.__commands,
                                                  SecureChannelAuthenticator()
                                                  )
        oldUmask = os.umask(077)
        listener = \
            self.proactor.makeListener(UNIXAddress(sockFile), connectHandler)
        os.umask(oldUmask)
        
        # create a TCP socket
        connectHandler = self.__addConnectHandler(self.__commands, 
                                                  self.__authenticator
                                                  )
        tcpListener = \
            self.proactor.makeListener(INETAddress('', 0), connectHandler)
        print 'socket name = %s' % (tcpListener._sock.getsockname(),)
        port = tcpListener._sock.getsockname()[1]
        self.__servers[port] = tcpListener
        
        # XXX print out the command to connect remotely
        token = getApp().auth.createToken()
        print 'yed -i TestSlave -m localhost:%d -a %s' % (port, token)
    
    def addTCPServer(self, port):
        """
            Create a TCP server running on the specified port.  Returns the 
            actual port number (an integer).
            
            Parms:
                port: [int] port number.  If a value of zero is used, a port 
                    will be allocated from the ephemeral port range.
        """
        connectHandler = self.__addConnectHandler(self.__commands,
                                                  self.__authenticator
                                                  )
        listener = \
            self.proactor.makeListener(INETAddress('', port), connectHandler)
        if port == 0:
            port = listener._sock.getsockname()[1]
        self.__servers[port] = listener
        return port
    
    def removeTCPServer(self, port):
        """
            Takes down the TCP server running on the specified port.

            Parms:
                port: [int] port number.
        """
        self.proactor.remove(self.__servers[port])
        del self.__servers[port]
    
    def __gotMessage(self, elem):
        # run the message
        elem()
        
    def __run(self):
        debug('starting Communicator thread')
        while self.proactor.hasConnections():
            try:
                self.proactor.processOneEvent()
            except Exception, ex:
                error('Exception in proactor loop: %s', traceback.format_exc())
    
    def started(self):
        """
            Returns true if the communicator has been started.
        """
        return bool(self.__thread)

    def start(self):
        if not self.__thread:
            self.__thread = Thread(target = self.__run)
            self.__thread.start()
    
    def __stop(self):
        if self.__thread:
            # close all of the active communicators
            for handler in self.__communicators:
                handler.close()
            self.__commQueue.close()
            self.__thread = None
        
    def stop(self):
        self.__commQueue.add(self.__stop)

        # clean up our UNIX socket
        if self.__sockPath:
            os.remove(self.__sockPath)
    
    def __handleReply(self, reply):
        # we write the reply in the communication thread to save us from
        # having to deal with concurrency issues.  The main thing is,
        # the comm thread needs to be interrupted so that it can re-enter
        # poll while ready to write to the client.
        reply.handler.writeReply(str(reply.reply), contentType = 'text/plain')
    
    def nudge(self):
        """
            "Nudge" the communications thread so that any new outgoing data 
            can be sent.
        """
        self.__commQueue.add(lambda: None)
    
    def sendReply(self, reply):
        """
            Sends a reply from the UI thread to the communicator thread.
            
            parms:
                reply::
                    [@UIThreadRequest]
        """
        self.__commQueue.add(lambda reply = reply: self.__handleReply(reply))

    def __execQueueElem(self, top, elem):
        try:
            elem(top)
        except:
            traceback.print_exc()
    
    def processUIQueue(self, top):
        """
            Process the queue of actions to be executed from the UI thread.
            
            This should be called periodically from the UI thread.
            
            parms:
                top::
                    [@yed.core.Top] current toplevel
        """
        try:
            while True:
                # get but don't block
                elem = self.__uiQueue.get(False)
                self.__execQueueElem(top, elem)
        except Empty:
            return
    
    def runUIQueue(self):
        """
            Continuously process the UI queue, blocking on it.
        """
        while True:
            elem = self.__uiQueue.get()
            self.__execQueueElem(None, elem)
    
    def runInCommThread(self, func):
        """
            Run a function from the communication thread.
            
            parms:
                func: [callable<>] Function to run.
        """
        self.__commQueue.add(func)

    def runInUIThread(self, func):
        """
            Run a function in the UI thread.
            
            parms:
                func: [callable<@yed.core.Top>] a function that accepts the 
                toplevel window as an argument.
        """
        self.__uiQueue.put(func)

class ProtocolError(Exception):
    """
        Raised when we get something outside of the normal HTTP protocol.
    """
    pass

class _HTTPResponse:
    """
        Holds the response data.
    """
    
    def __init__(self, code, msg):
        self.code = code
        self.msg = msg
        self.headers = {}
        self.parcel = ''
    
    def __str__(self):
        headers = '\n'.join([
            '%s: %s' % (header, value) 
                for header, value in self.headers.items()
        ])
        return 'HTTP/1.1 %d %s\n%s\n\n%s' % \
            (self.code, self.msg, headers, self.parcel)

class _HTTPReplyParser:
    
    RESPONSE = 0
    HEADERS = 1
    BODY = 2
    
    __lineRx = re.compile(r'[^\n]*\n')
    __respRx = re.compile(r'HTTP/1\.1 (\d+) (.*)')
    __headerRx = re.compile(r'(\S+):\s*(.*)')
    def __init__(self):
        self.__buffer = ''
        self.reset()
    
    def reset(self):
        self.__state = self.RESPONSE
        self.__size = 0
        self.response = None
        self.gotResponse = False
    
    def __getLine(self):
        m = self.__lineRx.match(self.__buffer)
        if m:
            line = m.group()
            self.__buffer = self.__buffer[m.end():]
            return line
        else:
            return None
        
    def feed(self, data):
        self.__buffer += data

        # process as much stuff as we can        
        while True:

            if self.__state == self.BODY:
                
                # see if the buffer is of the requisite size (if no size was
                # defined, just keep on reading
                if self.__size and len(self.__buffer) >= self.__size:
                    self.response.parcel = self.__buffer[:self.__size]
                    
                    # leave the remainder of the data in the buffer for the 
                    # next request
                    self.__buffer = self.__buffer[self.__size:]
                    
                    self.gotResponse = True
                elif not self.__size:
                    self.response.parcel += self.__buffer
                    self.__buffer = ''
                    self.gotResponse = True
                
                # when processing the body, we always completely consume the 
                # data so there's no more processing.
                return
            
            else:
                # the rest of the states are line oriented, so try to get a 
                # complete line
                line = self.__getLine()
                if not line:
                    return
                    
                if self.__state == self.RESPONSE:
                    m = self.__respRx.match(line)
                    if not m:
                        raise ProtocolError('Malformed response line: %s' %
                                            repr(line)
                                            )
                    
                    # reset the response variables
                    self.gotResponse = False
                    self.response = _HTTPResponse(int(m.group(1)), m.group(2))
                    self.__state = self.HEADERS
                    self.__size == 0
                    
                elif self.__state == self.HEADERS:
                    
                    # empty line signals the start of the body
                    if line.strip() == '':
                        self.__state = self.BODY
                        continue
                        
                    # parse a header
                    m = self.__headerRx.match(line)
                    if not m:
                        raise ProtocolError('Malformed header line: %s' %
                                            repr(line)
                                            )
                    
                    header = m.group(1)
                    value = m.group(2)
                    
                    # check for the "content length" field, used to figure out
                    # how much body there is
                    if header.lower() == 'content-length':
                        try:
                            self.__size = int(value.strip())
                        except ValueError:
                            raise ProtocolError('illegal value for '
                                                 'content-length: %s' %
                                                 value
                                                )
                
                else:
                    assert 0, 'Invalid state: %d' % self.__state

class DurableSocket(DataHandler):
    """
        A socket that reconnects itself if it becomes disconnected and fits 
        nicely into the proactor framework.
    """
    def __init__(self, comm, addr):
        self.__comm = comm
        self.__addr = addr
        self.__readyToClose = False
        self.__reconnect = True
        self.__queue = Queue()
        self.__prefetch = ''
        self.__pendingRequests = []
        self.__lock = Lock()
        self.__stateChange = Condition(self.__lock)
        self.__connected = False
        
        self.__replyParser = _HTTPReplyParser()

        # and startthe connection        
        self.__comm.proactor.makeConnection(self.__addr, self)
    
    def send(self, data):
        self.__queue.put(data)
        self.__comm.nudge()

    def close(self):
        # signal ourselves ready to close
        self.__readyToClose = True
        self.__reconnect = False
        self.__comm.nudge()
    
    def reconnect(self):
        """
            Drop the connection and force a reconnect.
        """
        self.__readyToClose = True
        self.__comm.nudge()
    
    def addResponseHandler(self, handler):
        """
            Adds a response handler to the queue.  Response handlers are FIFO 
            - when a response comes in, it goes to the first available handler.
            
            parms:
                handler: [callable<@_HTTPResponse>]
        """
        self.__pendingRequests.append(handler)
    
    def __waitForCondition(self, timeout, condition):
        self.__lock.acquire()
        while not condition() and timeout is None or timeout > 0:
            startTime = time.time()
            self.__stateChange.wait(timeout)
            
            # subtract elapsed time from the timeout (if we're doing a timeout)
            elapsed = time.time() - startTime
            if timeout is not None:
                timeout -= elapsed
        self.__lock.release()
    
    def waitForConnect(self, timeout = None):
        """
            Block until we are connected.
        
            parms:
                timeout: [float or None] if present, the number of seconds to 
                wait for a connection
        """
        self.__waitForCondition(timeout, lambda: self.__connected)

    def waitForDisconnect(self, timeout = None):
        """
            Block until we are disconnected.
        
            parms:
                timeout: [float or None] if present, the number of seconds to 
                wait for a connection
        """
        self.__waitForCondition(timeout, lambda: not self.__connected)

    # implementation of DataHandler interface

    def readyToGet(self):
        return self.__prefetch or not self.__queue.empty()

    def readyToPut(self):
        return True

    def readyToClose(self):
        return self.__readyToClose

    def handleConnect(self):
        debug('reconnected!')
        self.__lock.acquire()
        self.__connected = True
        self.__stateChange.notifyAll()
        self.__lock.release()
        
    def peek(self, size):
        while not self.__queue.empty() and len(self.__prefetch) < size:
            self.__prefetch += self.__queue.get()
        
        return self.__prefetch[:size]

    def get(self, size):
        assert size <= len(self.__prefetch)
        self.__prefetch = self.__prefetch[size:]

    def put(self, data):
        self.__replyParser.feed(data)
        
        if self.__replyParser.gotResponse:

            # hold the response
            response = self.__replyParser.response

            # reset the reply parser so we're ready for the next request            
            self.__replyParser.reset()

            try:
                self.__lock.acquire()
                
                # make sure that there is a request to match this response to
                if not self.__pendingRequests:
                    error('Got apparently unsolicited response: %s' % response)
                    return
                    
                pending = self.__pendingRequests.pop(0)
                pending(response)
            finally:
                self.__lock.release()
        
    def handleError(self):
        #self.handleDisconnect()
        pass

    def handleDisconnect(self):
        debug('handling disconnect')
        
        # clean out any leftover crud in the handler
        self.__replyParser.reset()
        
        # fail all pending requests
        self.__lock.acquire()
        try:
            for responseHandler in self.__pendingRequests:
                responseHandler(_HTTPResponse(1001, 'Disconnected'))
            self.__pendingRequests = []
            
            # notify observers of the state change
            self.__connected = False
            self.__stateChange.notifyAll()
        finally:
            self.__lock.release()
        
        def reconnect():
            debug('reconnecting to %s:%s', self.__addr.host, self.__addr.port)
            self.__comm.proactor.makeConnection(self.__addr, self)
        
        if self.__reconnect:
            # create a new connection in one second
            self.__readyToClose = False
            self.__comm.proactor.schedule(1, reconnect)
    
class Remote:
    """
        Proxy to communicate with a remote instance of Yed.
        
        Public-vars:
            authToken: [string or None] set this to the authentication token 
            (if any)
    """
    def __init__(self, uiQueue, sock):
        self.__uiQueue = uiQueue
        self.__sock = sock
        self.__parser = _HTTPReplyParser()
        self.authToken = None
        self.__resultContext = {'error': self.__error}

    def waitForConnect(self):
        self.__sock.waitForConnect()
    
    def waitForDisconnect(self):
        self.__sock.waitForDisconnect()
    
    @staticmethod
    def __error(text):
        raise RemoteError(text)

    def __sendRequest(self, method, path, headers, data = '',
                      callback = None,
                      timeout = None
                      ):
        """
            Sends a request, returns a @_HTTPResponse
            
            parms:
                method::
                    [string] HTTP method - usually "GET" or "POST"
                path::
                    [string] request path
                headers::
                    [dict<string, string>] header dictionary
                data::
                    [string or None] if provided, this is parcel data to send 
                    (the body of a post request)
                callback::
                    [callable<@_HTTPResponse>] Function to call (from the UI 
                    thread) when the RPC completes or fails.
                timeout::
                    [float or None] if specified, this is the time to wait for 
                    a reply in seconds.  If a reply is not received within 
                    this interval, TimeoutError is raised.  This argument is 
                    ignored if "callback" is specified.
        """

        # if there's no callback, set up an event so we know when to stop 
        # waiting and someplace to put the response
        if not callback:
            responseHolder = []
            waiter = Event()

        def handleUserCallback(response):
            # if there is a callback, the response handler needs to schedule 
            # the callback on the UI thread.  Otherwise, it needs to directly 
            # send an event to unblock the wait.
            if callback:
                self.__uiQueue.put(lambda top: callback(response))
            else:
                responseHolder.append(response)
                waiter.set()
                
        self.__sock.addResponseHandler(handleUserCallback)
        
        request = StringIO()
        request.write('%s %s HTTP/1.1\n' % (method, path))
        for header, value in headers.items():
            request.write('%s: %s\n' % (header, value))
        if self.authToken:
            request.write('Cookie: yed_auth="%s"\n' % self.authToken)
        request.write('\n')
        self.__sock.send(request.getvalue())

        # send the post data
        while data:
            self.__sock.send(data[:4096])
            data = data[4096:]
        
        # if there's a callback, we're done
        if callback:
            return
        
        # Otherwise we have to wait for the response
        waiter.wait(timeout)
        if not waiter.isSet():
            self.__sock.reconnect()
            raise TimeoutError()
        return responseHolder[0]

    def __buildRequest(self, request, args):
        if args:
            query = \
                '&'.join(['%s=%s' % (key, encodeString(str(val))) 
                            for key, val in args.items()
                          ]
                         )
            request = '%s?%s' % (request, query)
        
        return request

    @staticmethod
    def __stealArg(args, key, default):
        """
            parms:
                args: [dict<string, any>] keyword arguments
                key: [string] key to look for
        """
        if args.has_key(key):
            result = args[key]
            del args[key]
            return result
        else:
            return default

    def __sendGetRequest(self, request, **args):
        request = self.__buildRequest(request, args)
        timeout = self.__stealArg(args, 'timeout', None)
        response = self.__sendRequest('GET', request, {}, timeout = timeout)
        return self.__makeResponse(response)
    
    def __sendPostRequest(self, request, data, **args):
        request = self.__buildRequest(request, args)
        timeout = self.__stealArg(args, 'timeout', None)
        response = self.__sendRequest('POST', request, 
                                      {'content-length': len(data)},
                                      data,
                                      timeout = timeout
                                      )
        return self.__makeResponse(response)     

    def __makeResponse(self, response):
        # return based on the HTTP result, then the parcel
        if response.code == 200:
            return response.parcel
        elif response.code == 1001:
            raise NotConnectedError(response.msg)
        elif response.code == 306:
            raise AuthError(response.msg)
        else:
            raise CommError('%d - %s' % (response.code, response.msg))
    
    def close(self):
        "Close the connection"
        self.__sock.close()

    def edit(self, filename, block):
        return self.__sendGetRequest('/yed/edit', filename = filename,
                                     block = block,
                                     timeout = not block and 30 or None
                                     )
    
    def ping(self):
        return self.__sendGetRequest('/yed/ping', timeout = 5)

    def ls(self, timeout = 10):
        return self.__sendGetRequest('/yed/ls', timeout = timeout)
    
    def saveAll(self):
        return self.__sendGetRequest('/yed/saveAll', timeout = 60)
    
    def reloadAll(self):
        return self.__sendGetRequest('/yed/reloadAll', timeout = 60)
    
    def readFile(self, filename):
        return self.__sendGetRequest('/yed/readFile', filename = filename,
                                     timeout = 30)
    
    def writeFile(self, filename, data, mode = None):
        if mode is not None:
            kwArgs = {'mode': mode}
        else:
            kwArgs = {}
        result = self.__sendPostRequest('/yed/writeFile', data, 
                                        filename = filename,
                                        timeout = 30,
                                        **kwArgs
                                        )
        words = result.split(None, 1)
        if words[0] == 'ok':
            return eval(words[1])
        elif words[0] == 'error:':
            raise Exception(words[1])
        else:
            # error: <error-code> - <text>
            raise Exception(words[0], words[1].split(None, 1))
    
    # naming conventions from here on in intentionally mirror the python API

    def listdir(self, dir):
        """
            Returns a dictionary mapping filenames to their "stat" tuples.
        """
        def error(text):
            raise RemoteError(text)
        return eval(self.__sendGetRequest('/yed/listDir', dir = dir,
                                          timeout = 10
                                          ),
                    self.__resultContext
                    )
    
    def stat(self, filename):
        result = self.__sendGetRequest('/yed/stat', filename = filename, 
                                       timeout = 5
                                       )
        
        # "error:" errors are protocol related
        assert not result.startswith('error:'), result
        
        words = result.split(None, 1)
        if words[0] == 'enoent':
            return None
        elif words[0] == 'ok':
            try:
                return eval(words[1])
            except:
                raise Exception('Unable to evaluate: %s' % repr(words[1]))
        else:
            # error: <error-code> - <text>
            raise Exception(words[0], words[1].split(None, 1))
    
    def glob(self, pattern):
        result = self.__sendGetRequest('/yed/glob', pattern = pattern, 
                                       timeout = 10)
        
        # XXX not entirely correct: files could start with "error:"
        assert not result.startswith('error:'), result
        
        # XXX ... and filenames can contain newlines
        if not result:
            return []
        else:
            return result.split('\n')
    
    def getToken(self):
        return self.__sendGetRequest('/auth/getToken', timeout = 5)

def getExistingInstance(self):
    pass

class Authenticator:
    """
        Authenticator interface.
    """
    
    def authorized(self, token):
        """
            Checks an authentication token, returns true if it is valid.

            parms:
                token: [string or None] authentication token.
        """
        raise NotImplementedError()

class BasicAuthenticator(Authenticator):
    """
        Simple authenticator implementation.
        
        Public-vars:
            tokens: [set<string>] set of currently valid tokens.
    """
    
    def __init__(self):
        self.tokens = set()

    def authorized(self, token):
        return token in self.tokens
    
    def createToken(self):
        # if we've got /dev/random, use that.
        if os.path.exists('/dev/random'):
            bytes = open('/dev/random').read(12)
        else:
            # nope.  Use a weak random number
            bytes = ''.join([chr(random.randrange(256)) for i in range(12)])
        token = base64.encodestring(bytes).strip()
        self.tokens.add(token)
        return token

class SecureChannelAuthenticator(Authenticator):
    """
        Authenticator for use by secure channels - always allow access.
    """
    
    def authorized(self, token):
        return True
