#==============================================================================
#
#  $Id$
#
"""
   Core editor functionality.
"""
#
#   Copyright (C) 2007 Michael A. Muller
#  Portions Copyright (C) 2008 Google Inc.
#
#   This file is part of SpugEdit.
#
#   SpugEdit is free software: you can redistribute it and/or modify it under
#   the terms of the GNU Lesser General Public License as published by the
#   Free Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   SpugEdit is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with SpugEdit.  If not, see <http://www.gnu.org/licenses/>.
#
#==============================================================================

from Tkinter import *
import tkFont
from spug.win.StdWin import *
import re
import os, glob, sys, traceback, stat
from pickle import Pickler, Unpickler
import log

normalChars = ''.join([chr(ch) for ch in range(32, 128)])
wsRx = re.compile(r'[ \t]+')

class Abort(Exception):
   """
      Raised by commands and prompts that are unable to complete (or require
      some user verification to complete).
   """
   
   def __init__(self, message, confirm = False):
      """
         parms:
            message: [str] exception text
            confirm: [bool] if true, give the user the option to confirm the 
               questionable action before retrying.
      """
      Exception.__init__(self, message)
      self.confirm = confirm

class PromptHandler:

   def handleData(self, data):
      """
         Called when the return key is pressed and data is passed back from
         the handler.
      """
      raise NotImplementedError()

   def handleCancel(self):
      """
         Called when "escape" is pressed and the prompt is aborted.
      """
      raise NotImplementedError()

class StatusWidget(Frame):
   """
      Base class for widgets installed in the status bar.
   """
   
   def handleDestroy(self):
      """
         Called when the widget is destroyed.
         
         May be overriden, default implementation does nothing
      """
      pass
      
class Prompt(StatusWidget):

   def __init__(self, parent, label, handler, initValue = '', entryWidth = 8):
      StatusWidget.__init__(self, parent)
      
      self.parent = parent
      self.handler = handler

      # create and position the entryfield (we pack from right to left because
      # it is important that the entryfield be visible)
      self.entry = Entry(self, width = entryWidth)
      self.entry.pack(side = RIGHT, expand = 1, fill = BOTH)
      self.entry.focus()

      # create and position the label
      self.label = Label(self, text = label)
      self.label.pack(side = RIGHT)

      # bind the keys
      self.entry.bind('<Return>', self.handleReturn)
      self.entry.bind('<Escape>', self.handleEscape)

      # set the initial value
      self.entry.insert(0, initValue)

   def getText(self):
      """
         Gets the text that will be passed to the callback through 
         handleData().  By default, this is simply the contents of the entry.  
         Derived classes may override.
      """
      return self.entry.get()
      
   def handleReturn(self, event):
      self.handler.handleData(self.getText())
      self.handler = None
      self.parent.removeStatusWidget(self)

   def handleEscape(self, event):
      # leave calling "handleCancel()" to the destroy handler
      self.parent.removeStatusWidget(self)
   
   def handleDestroy(self):
      if self.handler:
         self.handler.handleCancel()
         self.handler = None
      
      # clean up reference cycles
      del self.parent, self.label, self.entry

class HistoryPrompt(Prompt):
   """
      A prompt which preserves history - up and down arrows recall previous 
      contents.
   """
   
   def __init__(self, parent, label, handler, initValue = '', entryWidth = 8,
                history = None
                ):
      """
         Parms: See @Prompt.__init__()
            history::
               [list<string>] the current history list.  If we do a 
               "handleData", we'll add the new value to the end of this.
      """
      Prompt.__init__(self, parent, label, handler, initValue, entryWidth)
      if history is None: history = []
      
      # add an empty item to history for the current prompt contents
      history.append('')
      
      self.__history = history
      self.__cur = len(history) - 1
      self.entry.bind('<Up>', self.handlePrev)
      self.entry.bind('<Down>', self.handleNext)
   
   def __restoreFromHistory(self, increment):
      """
         Stores the current entryfield contents in history, adds 'increment' 
         to the history pointer and sets the entryfield from the new history 
         item.
         
         This is the inner method for accomplishing the "scroll through 
         history" behavior.
      """
      # store the current element
      self.__history[self.__cur] = self.getText()
   
      # restore the previous element    
      self.__cur += increment
      self.entry.delete(0, 'end')
      self.entry.insert(0, self.__history[self.__cur])

   def handlePrev(self, event):
      if self.__cur > 0:
         self.__restoreFromHistory(-1)
   
   def handleNext(self, event):
      if self.__cur < len(self.__history) - 1:
         self.__restoreFromHistory(1)
   
   def handleEscape(self, event):
      "Overrides @Prompt.handleEscape() to remove the last history item."
      del self.__history[-1]
      Prompt.handleEscape(self, event)

   def handleReturn(self, event):
      "Overrides @Prompt.handleReturn() to capture contents in history."
      self.__history[-1] = self.getText()
      Prompt.handleReturn(self, event)
         
class TypeAheadHandler:
   """
      The abstract interface for type-ahead handlers.
   """

   def getOptions(self, text):
      """
         Returns all options (a list<string>) that match the provided text.
         All options returned should start with the text.

         parms:
            text::
               [string] 
      """
      raise NotImplementedError()
   
   def getCommonText(self, options, text):
      """
         Returns the longest common text of the options starting with
         "text"
         
         parms:
            options::
               [list<string>] the list of options
            text::
               [string] the known common text
      """
      raise NotImplementedError()
   
   def handleSelect(self, text):
      """
         Called when the selection in the type-ahead's listbox changes.
         
         parms:
            text::
               [string] the selected item's text.
      """
      raise NotImplementedError()

class TypeAheadPrompt(Prompt):

   def __init__(self, parent, label, handler, initValue = '',
                typeAheadHandler = None
                ):
      Prompt.__init__(self, parent, label, handler, initValue)
      self.typeAheadHandler = typeAheadHandler
      self.entry.bind('<KeyPress>', self.handleKey)
      self.list = None
      
      # if the user explicitly selected an item from the list, this will 
      # hold the text of the item
      self.__selection = None

   def recalculateList(self, text = None, options = None):
      if options is None:
         assert text is not None
         options = self.typeAheadHandler.getOptions(text)

      self.list.deleteAll()
      for option in options:
         self.list.append(option)
      
      # reset the selection
      self.__selection = None
      
      # adjust the view so that the rightmost contents of the list are visible
      self.list.xview_moveto(1)

   def destroy(self):
      if self.list:
         self.list.destroy()
      Prompt.destroy(self)
   
   def setSelection(self, index):
      self.list.activate(index)
      self.list.selection_clear()
      self.list.selection_set(index)
      
      # we have to trigger this explicitly - not sure why selection_set 
      # doesn't do it for us.
      self.handleSelect(None)

   def handleSelect(self, event):
      """
         Called to handle a listbox select event.
      """
      # set the selection
      selections = self.list.curselection()
      if selections:
         self.__selection = self.list.get(selections[0])
         self.entry.delete(0, 'end')
         self.entry.insert(0, self.__selection)

   def getText(self):
      """
         Overrides @Prompt.getText() to offer the selection if it is defined.
      """
      if self.__selection is not None:
         return self.__selection
      else:
         return self.entry.get()

   def handleKey(self, event):
      if event.keysym == 'Tab':
         # user hit the tab key - recalculate the options.
         text = self.entry.get()
         options = self.typeAheadHandler.getOptions(text)
         if not options:
            self.bell()
         else:
            if not self.list:
               self.list = ScrollingList(self.parent)
               self.list.bind('<<ListboxSelect>>', self.handleSelect)
               self.parent.addPanelWidget(self.list)
               self.entry.focus()
            
            self.recalculateList(options = options)

            # replace the entry with the common text of all of the matches
            self.entry.delete(0, 'end')
            common = self.typeAheadHandler.getCommonText(options, text)
            self.entry.insert(0, common)
            self.entry.xview(len(common))
         return 'break'
      elif event.keysym == 'Return':
         return
      elif event.keysym == 'Escape':
         return
      elif event.keysym == 'BackSpace' and self.list:
         self.recalculateList(self.entry.get()[:-1])
      elif event.keysym == 'Up' and self.list:
         # cursor to the previous item in the list
         index = self.list.index('active')
         index -= 1
         if index >= 0:
            self.setSelection(index)
         
      elif event.keysym == 'Down' and self.list:
         # cursor up to the next item in the list
         index = self.list.index('active')
         index += 1
         if index < self.list.size():
            self.setSelection(index)
         
      elif event.char and event.char in normalChars:
         text = self.entry.get() + event.char
         if self.list:
            self.recalculateList(text)

class SearchPrompt(Prompt):
   """
      Another bottom-bar prompt combo, this one facilitates type-ahead
      searches by calling the handler at every change.
   """


   def __init__(self, top, text, handler):
      Prompt.__init__(self, top, text, handler)
      self.entry.bind('<KeyPress>', self.handleKey)

   def handleKey(self, evt):
      
      if evt.keysym == 'BackSpace':
         # just dispatch to the handler as long as there is something to
         # delete, ignore the result because the handler doesn't have the
         # object to abort
         data = self.entry.get()
         if data:
            self.handler.handleChange(data[:-1])
      
      elif evt.char and evt.char in normalChars:
         data = self.entry.get() + evt.char

         # ask the handler if the change is acceptable
         if not self.handler.handleChange(data):
            # nope.  Beep and abort.
            self.bell()
            return 'break'

class YesNoHandler:
   """
      Callback handler for @YesNoPrompt.  This is a different from a 
      PromptHandler, which handles a much more general set of events.
   """

   def handleYes(self):
      """
         Called when "yes" is selected from the YesNoPrompt.  Must be 
         implemented by derived class.
         
         If this returns true, the destruction of the YesNoPrompt will be 
         aborted.
      """
      raise NotImplementedError()

   def handleNo(self):
      """
         Called when "no" is selected from the YesNoPrompt.  Must be 
         implemented by derived class.
         
         If this returns true, the destruction of the YesNoPrompt will be 
         aborted.
      """
      raise NotImplementedError()

class YesNoPrompt(Prompt):
   """
      A prompt to provide the user with a yes/no query.
   """

   class Callback(PromptHandler):
      """
         The PromptHandler for the YseNoPrompt.  This houses and sends events
         to the YesNoHandler.
      """
      def __init__(self, handler):
         self.handler = handler
      def handleData(self, data):
         if data == 'y':
            self.handler.handleYes()
         else:
            self.handler.handleNo()
      def handleCancel(self):
         return self.handler.handleNo()
   
   _optionsLabel = ' (Y|N)'

   def __init__(self, top, text, handler):
      Prompt.__init__(self, top, text + self._optionsLabel, 
                      self.Callback(handler),
                      entryWidth = 1
                      )
      self.yesNoHandler = handler
      self.entry.bind('<KeyPress>', self.handleKey)

   def handleKey(self, event):
      if event.char and event.char == 'y':
         if not self.yesNoHandler.handleYes():
            self.handler = None
            self.parent.removeStatusWidget(self)
         return 'break'
      elif event.char and event.char == 'n':
         if not self.yesNoHandler.handleNo():
            self.handler = None
            self.parent.removeStatusWidget(self)
         return 'break'

   def handleReturn(self, event):
      """
         Override @Prompt.handleReturn() so we don't go terminating withut a 
         good answer.
      """
      self.bell()

class YesNoCancelHandler(YesNoHandler):
   """
      Callback handler for boxes that support yes and no with cancel as a
      separate action.
   """
   
   def handleCancel(self):
      """
         Called when cancel (Esc or "c") is selected.
         
         Must be implemented by derived class.
      """
      raise NotImplentedError()
      
class YesNoCancelPrompt(YesNoPrompt):
   """
      Prompt that offers the user the choice of yes/no or cancel (which aborts
      the action in its entirety).
   """
   
   class Callback(YesNoPrompt.Callback):
      def handleCancel(self):
         self.handler.handleCancel()
   
   _optionsLabel = ' (Y|N|C)'
   
   def handleKey(self, event):
      if event.char == 'c':
         if not self.yesNoHandler.handleCancel():
            self.handler = None
            self.parent.removeStatusWidget(self)
         return 'break'
      else:
         return YesNoPrompt.handleKey(self, event)

class Font:
   """
      A simplified font definition.  Presents various kinds of font 
      information as the things that we are actually intersted in.
   """
   
   def __init__(self, nameOrFontObj, spaceWidth = None):
      """
         parms:
            nameOrFontObj::
               [string or tkFont.Font] this can either be a font name or a 
               font object.
            spaceWidth::
               [int or None] if provided (and true), this is the width of a 
               space character (0x20) in pixels.  This is useful in at least 
               one case (the "fixed" font) where the tkFont.Font object 
               describes a font that is different from the one used by a 
               control when given that font name.
      """
      
      # decide whether we have a font name or a complete font object and set 
      # the appropriate instance variables
      self.__fontObj = self.__fontName = None
      if isinstance(nameOrFontObj, tkFont.Font):
         self.__fontObj = nameOrFontObj
      else:
         assert isinstance(nameOrFontObj, str)
         self.__fontName = nameOrFontObj
      
      self.__spaceWidth = spaceWidth

   @property
   def spaceWidth(self):
      if not self.__spaceWidth:
         self.__spaceWidth = self.__getFontObj().measure(' ')
      return self.__spaceWidth
   
   def __getFontObj(self):
      if not self.__fontObj:
         self.__fontObj = tkFont.Font(font = self.__fontName)
      return self.__fontObj
   
   def __str__(self):
      if not self.__fontName:
         self.__fontName = str(self.__fontObj)
      return self.__fontName
      
class Buffer(ScrollingText):
   """
      A text buffer.

      Public-vars:
         attrs: [dict<str, any>] persisted attributes (get saved as part of 
            the editor state).
         nattrs: [dict<str, any>] non-persisted attributes.
         rules: [@Rules] the buffer's rules object
         name: [str] the buffer name
   """

   def __init__(self, top, name):
      ScrollingText.__init__(self, top)
      self.attrs = {}
      self.nattrs = {}
      self.rules = None
      self.name = name
      
      self.__font = None
      self.__destroyHandlers = []

   def addDestroyHandler(self, func):
      """
         Adds a handler that will be called when the buffer is destroyed.
         
         parms:
            func::
               [callable<@Buffer>]
      """
      self.__destroyHandlers.append(func)
   
   def callDestroyHandlers(self):
      """
         Calls all of the destroy handlers.
      """
      for handler in self.__destroyHandlers:
         handler(self)
      
      # remove them
      self.__destroyHandlers = []
         
   def focus(self):
      """
         Override focus so that the focus is propagated to the text buffer,
         not the frame.
      """
      self.text.focus()

   def getModified(self):
      return bool(self.tk.call((self.text._w, 'edit', 'modified')))
   
   def setModified(self, val):
      self.tk.call((self.text._w, 'edit', 'modified', repr(val)))

   def getRow(self, index):
      """
         Returns the row number of the given index (an integer).
      """
      index = self.index(index)
      return int(index.split('.')[0])

   def getColumn(self, index):
      """
         Returns the column number of the given index (an integer).
      """
      index = self.index(index)
      return int(index.split('.')[1])
      
   def normalizeRange(self, a, b):
      """
         Given two buffer indexes, returns a tuple of the indexes in the 
         order that that they occur in the buffer.
      """
      if self.compare(a, '>', b):
         return b, a
      else:
         return a, b
   
   def moveSelection(self, startIndex, newIndex):
      """
         Moves the selection so that the insertion point is at the new index.
         
         parms:
            startIndex::
               [string] the start position of the motion - this will be the 
               beginning/end of the range if no text is currently selected
            newIndex::
               [string] the new index
      """
      # first get the current selection
      selRange = self.tag_ranges('sel')
      if selRange:
         # there is text selected - see if the insertion cursor corresponds to 
         # the beginning or the end
         insertIndex = self.index(startIndex)
         if selRange[0] == insertIndex:
            # insert is at the beginning of the selection
            if self.compare(newIndex, '>', selRange[1]):
               # we've crossed the end of the selection - remove till the end,
               # add from the end to the new end
               self.tag_remove('sel', insertIndex, selRange[1])
               self.tag_add('sel', selRange[1], newIndex)
            elif self.compare(newIndex, '>', selRange[0]) and \
                 self.compare(newIndex, '<', selRange[1]):
               # we've moved into the selection - just remove from the beginning 
               # to the old index
               self.tag_remove('sel', selRange[0], newIndex)
            else:
               # we've moved the selection to before the start of the
               # current selection - just add the new selection
               self.tag_add('sel', newIndex, selRange[0])
            self.mark_set('anchor', selRange[1])
         elif selRange[1] == insertIndex:
            # insert is at the end of the selection
            if self.compare(newIndex, '<', selRange[0]):
               # we've crossed the beginning of the selection - remove till the end,
               # add from the new beginning to the old beginning
               self.tag_remove('sel', selRange[0], insertIndex)
               self.tag_add('sel', newIndex, selRange[0])
            elif self.compare(newIndex, '>', selRange[0]) and \
                 self.compare(newIndex, '<', selRange[1]):
               # we've moved into the selection - just remove from the new index to 
               # the end
               self.tag_remove('sel', newIndex, selRange[1])
            else:
               # we've moved the selection to after the end of the
               # current selection - just add the new selection
               self.tag_add('sel', selRange[1], newIndex)
            self.mark_set('anchor', selRange[0])
         else:
            sys.stderr.write('index is neither the start nor the end of the '
                              'range.'
                             )
      else:
         # there is no current selection, just select from the start of the 
         # range to the end
         start, end = self.normalizeRange(startIndex, newIndex)
         self.tag_add('sel', start, end)
         self.mark_set('anchor', startIndex)
   
   def writeState(self, out):
      """
         Writes the buffer state to the given output stream for subsequent
         state restoration.
      """
      token = {'name': self.name, 'attrs': self.attrs}
      Pickler(out).dump(token)
   
   def readFrom(top, src):
      """
         Create and return a new buffer read from the state information 
         provided by the stream "src"
         
         parms:
            top::
               [@Top] container window
            src::
               [file.read] readable object
      """
      token = Unpickler(src).load()
      log.trace('loading %s...' % token['name'])
      bufferType = getBufferTypeByName(token.get('attrs', {}).get('type'))
      bufferType.restore(top, token['name'], token['attrs'])
         
   readFrom = staticmethod(readFrom)

   def __getTaggedRegion(self, start, end):
      """
         Returns the outer start and end points of all tags that are adjacent
         to the range specified by 'start' and 'end'.
         
         As a side-effect, this removes all of the syntax-sensitive tags that 
         it encounters
         
         parms:
            start::
               [string] an existing lower bound
            end::
               [string] an existing upper bound
      """
      outerStart, outerEnd = start, end
      
      # prepare a set of syntax sensitive tags
      tags = set(self.rules.getSyntaxSensitiveTags())
      
      # Find all tags adjacent to the lower bound
      for tag in self.text.tag_names('%s - 1 chars' % start):
         
         # ignore if appropriate
         if tag not in tags:
            continue
         
         # ignore the selection
         if tag == 'sel':
            continue

         try:
            tagStart, tagEnd = self.text.tag_prevrange(tag, start)
         except ValueError:
            continue
         
         # clear the range
         self.tag_remove(tag, tagStart, tagEnd)
         
         # adjust the limits
         if self.text.compare(tagStart, '<', outerStart):
            outerStart = tagStart
         if self.text.compare(tagEnd, '>', outerEnd):
            outerEnd = tagEnd
      
      # add any tags adjacent on the upper bound
      index = end
      for tag in self.text.tag_names(index):
         
         # ignore the selection
         if tag == 'sel':
            continue
            
         # get the full tag
         tagStart, tagEnd = self.text.tag_nextrange(tag, end)
         
         # clear the range
         self.tag_remove(tag, tagStart, tagEnd)
         
         # extend the limits
         if self.text.compare(tagEnd, '>', outerEnd):
            outerEnd = tagEnd
         if self.text.compare(tagStart, '<', outerStart):
            outerStart = tagStart
      
      return outerStart, outerEnd
   
   def insert(self, index, data, *extra, **kw):
      """
         Overrides insert to reevaluate tags that may be affected.
      """
      extra = list(extra)

      # store the current insertion point, mark the point after it with a mark
      # that will float to the right
      insertEnd = 'yed.Buffer.insert'
      self.text.mark_set(insertEnd, index)
      self.text.mark_gravity(insertEnd, 'right')
      
      # run through all of the extra arguments - data is implicitly the first
      while extra:
         
         # first thing here is always the tags, grab it and call the insert
         tags = extra.pop(0)
         self.__insert(insertEnd, data, tags, 
                       ignoreRules = kw.get('ignoreRules', False)
                       )
         
         # get the next chunk of data
         if extra:
            data = extra.pop(0)
         else:
            data = None
      
      # do the last (if there is one)
      if data:
         self.__insert(insertEnd, data)

      # cleanup the mark
      self.text.mark_unset(insertEnd)
      
   def __insert(self, index, data, tags = None, ignoreRules = False):
      global audit
      
      # deal with the empty string case (and non-string case), sanitize all 
      # null characters (Tk sees these as an empty string)
      data = str(data)
      if not data:
         return
      data = data.replace('\0', '\\0')

      # record the current position
      outerStart = insertStart = self.text.index(index)
      outerEnd = insertEnd = 'yed.Buffer.insert'
   
      # do the insert, use a tag to keep track of the beginning and the end
      # of the insert (I used to do make the start the index and the end
      # yed.Buffer.insert, but this doesn't work very well when the insertion
      # point is "end")
      self.text.insert(index, data, 'yed.lastInsert')
      ranges = self.text.tag_ranges('yed.lastInsert')
      insertStart, insertEnd = ranges[:2]
      outerStart, outerEnd = insertStart, insertEnd
      self.text.tag_remove('yed.lastInsert', insertStart, insertEnd)

      # reparse if necessary
      if not ignoreRules and self.rules:
  
         # get the outermost enclosing tags
         outerStart, outerEnd = self.__getTaggedRegion(outerStart, outerEnd)
      
         self.rules.retagRegion(self, outerStart, outerEnd)
      
      # apply all tags
      if isinstance(tags, list):
         for tag in tags:
            audit.append((tag, self.index(insertStart), self.index(insertEnd)))
            self.tag_add(tag, insertStart, insertEnd)
         audit.append('---')
      elif tags:
         audit.append((tags, self.index(insertStart), self.index(insertEnd)))
         self.tag_add(tags, insertStart, insertEnd)
         audit.append(self.tag_ranges(tags))
         audit.append('---')

   def delete(self, start, end = None):
      """
         Delete the specified range.k
         
         parms:
            start::
               [string] start of range
            end::
               [string or None] end of range
      """
      if end is None:
         end = '%s + 1 chars' % start
      elif self.text.compare(start, '>', end):
         # reverse values if they are backwards
         start, end = end, start
     
      # store the absolute start position (if it's based on a mark, it might
      # move on us after the delete)
      index = self.text.index(start)
      
      # delete the text
      self.text.delete(start, end)
      
      # reparse tags if necessary
      if self.rules:
         # get all of the tags adjacent to the deleted region
         start, end = self.__getTaggedRegion(index, index)
         
         # retag
         self.rules.retagRegion(self, start, end)
   
   def setAutoSeparators(self, enabled):
      """
         Turns "autoseparators" on or off - seperators are entries in the undo
         stack that mark off the last set of actions to undo.
      """
      self.text['autoseparators'] = enabled
   
   def clearUndoStack(self):
      """
         Deletes everything on the undo stack.
      """
      self.text.edit_reset()
   
   def setFont(self, font):
      """
         Sets the buffer font.
         
         parms:
            font::
               [@Font] a font object.
      """
      self.__font = font
      self.text['font'] = font
      
      # fix tab widths
      self.text['tabs'] = val = font.spaceWidth * self.attrs.get('hardTabWidth', 8)
   
   def setTabWidth(self, widthInChars):
      """
         Sets the width of a hard tab character in terms of the width of a 
         single blank space.
         
         parms::
            widthInChars::
               [int]
      """
      # if the font hsan't been assigned, get it from the widget
      if not self.__font:
         self.__font = Font(self.text['font'])

      self.text['tabs'] = self.__font.spaceWidth * widthInChars
      self.attrs['hardTabWidth'] = widthInChars
   
   def getTabWidth(self):
      """
         Returns the width of a hard tab character (an integer).
      """
      return self.attrs.get('hardTabWidth', 8)
   
   def getDirectory(self):
      """
         Returns the directory of the buffer (a string).  This can mean 
         different things depending on the buffer type, for file buffer types 
         it is the directory of the filename.
         
         Formally, the return value is defined by the following rules:
         -  if the buffer has a "cwd" attribute, it is returned
         -  if the buffer has a "filename" attriubute, the directory of that 
            filename is returned.
         -  the current working directory of the editor is returned.
      """
      # get the current working directory from the buffer, then from the 
      # buffer's filename, then from the editors current working directory.
      cwd = self.attrs.get('cwd')
      if cwd is None:
         filename = self.attrs.get('filename')
         if filename:
               cwd = os.path.dirname(filename)
         else:
               cwd = os.getcwd()
      return cwd


class Rules:
   """
      Rules for performing standard actions which vary based on the type of the 
      buffer contents.  the Rules derivatives encode special stuff for 
      different programming languages.
   """
   def initialize(self, buffer):
      """
         Called when the buffer is initialized.  This should determine the tab
         width, and create all tags.
      """
   
   def calcNewlinePrefix(self, buffer, index = 'insert', hardReturn = False):
      """
         Called when the return key is pressed.  Should return the expected
         prefix of the line to be inserted following the insertion point.
         
         parms:
            buffer::
               [@Buffer]
            index::
               [string] index referencing the line after which the new line is 
               to be inserted.
            hardReturn::
               [boolean] this is a hint that indicates whether the new prefix 
               was the result of a "hard return" - the user hitting the enter 
               key - or a soft return generated from a reflow or typing across 
               a margin
      """
      raise NotImplementedError()
   
   def isJoinable(self, buffer, index = 'insert'):
      """
         Returns a tuple<boolean, int> - the boolean indicates whether the 
         following line can be joined with the current line, the integer
         indicates how many characters of the following line can be deleted
         after the join.
      """
      # by default, lines are joinable if there is no next line
      joinable = \
         buffer.compare(buffer.index('%s + 1 lines' % index), '<', 'end')
      
      # if we're joinable, we can delete all leading whitespace
      deletable = 0
      if joinable:
         nextLine = buffer.get('%s + 1 lines linestart' % index, 
                               '%s + 1 lines lineend' % index
                               )
         m = wsRx.match(nextLine)
         if m: deletable = m.end()

      return (joinable, deletable)
   
   def getBreakpoint(self, buffer, index, after = False):
      """
         Return the index of the best point to break the line before or after 
         the index.  "after" indicates whether we look before or after.
      """
      if after:
         endPoint = '%s lineend'
         incr = '+'
         comparator = '>='
      else:
         endPoint = '%s linestart'
         incr = '-'
         comparator = '<='

      # store the point where we stop looking
      beginLine = buffer.index(endPoint % index)
      
      # continue until we find whitespace or the endpoint
      while True:
         nextIndex = buffer.index('%s %s 1 chars' % (index, incr))
         if buffer.get(nextIndex).isspace():
            return after and nextIndex or index
         
         # if we've gone back to the previous line, stop here
         if buffer.compare(nextIndex, comparator, beginLine):
            return None
         
         index = nextIndex

   
   def retagRegion(self, buffer, start, end):
      """
         Re-evaluate all tags in the specified region.
         
         parms:
            buffer::
               [@Buffer] the buffer
            start::
               [string] the point corresponding to the beginning of the region
               to scan.
            end::
               [string] the point corresponding to the end of the region
               to scan.
      """
   
   def getSyntaxSensitiveTags(self):
      """
         Returns all tags that are "syntax sensitive" - i.e. subject to 
         reevaluation when we insert text.
      """
      return []

   def lookupCommand(self, command, context):
      """
         Lookup the yed command in the rules, returns the command (a 
         @yed.core.Command instance) if it is defined, *None* if not. 
         
         May be overriden, base implementation always returns *None*.
         
         parms:
            command::
               [string] the command
            context::
               [@yed.core.CommandContext] the context
      """
      return False

   def reflow(self, buffer, start, end, textwidth):
      """
         Reflow the region of the buffer.  We define it here in rules to allow 
         a language to completely redefine how reflow should work.
         
         parms:
            buffer::
               [@core.Buffer]
            start::
               [string] the text start index
      """
      # mark the end so that it floats with us as we reformat
      endMark = 'yed.reflow.endRange'
      buffer.mark_set(endMark, end)
      buffer.mark_gravity(endMark, 'right')
      
      # keep track of the last place where we inserted a line break
      lastInsertionPoint = None
         
      cur = start
      while buffer.compare(cur, '<', endMark):
         
         # store some useful indeces
         lineStart = '%s linestart' % cur
         lineEnd = '%s lineend' % cur
         
         # remove all trailing whitespace - this is an unfortunate side effect,
         # ideally we should verify that this is not beyond the endMark first
         # but I really don't want to have to check that in _3_ places.
         line = buffer.get(lineStart, lineEnd)
         rstripped = line.rstrip()
         if rstripped != line:
            buffer.delete(lineStart, lineEnd)
            buffer.insert(lineStart, rstripped)
         
         # go to the margin
         cur = buffer.index('%s.%s' % \
                             (buffer.getRow(cur), textwidth)
                            )
         
         # if the margin is past the end mark, quit now
         if buffer.compare(cur, '>=', endMark):
            break
         
         col = buffer.getColumn(cur)
         if col >= textwidth:
            # if the line exceeds the margin, split it
            restOfLine = buffer.get(cur, '%s lineend' % cur)
            if restOfLine.strip():
               breakpoint = self.getBreakpoint(buffer, cur)
               
               # nassssty exception casssssse.  Since we don't communicate 
               # very well with calcNewlinePrefix(), we have to check to make 
               # sure that we're not trying to break at the same word that we 
               # last broke it.  If we are, then we need to move on to the 
               # next breakpoint.
               if lastInsertionPoint and \
                  buffer.compare(breakpoint, '==', lastInsertionPoint):
                  breakpoint = self.getBreakpoint(buffer, cur, True)
                  
                  # delete next whitespace
                  endline = '%s lineend' % breakpoint
                  if breakpoint:
                     remaining = buffer.get(breakpoint, endline)
                     m = wsRx.match(remaining)
                     if m:
                        buffer.delete(breakpoint, 
                                    '%s + %d chars' % (breakpoint, m.end())
                                    )
                        
                        # if the entire remaining line was whitespace, don't 
                        # insert a newline
                        if len(remaining) == m.end(): breakpoint = None
                     
               if breakpoint:
                  prefix = self.calcNewlinePrefix(buffer, breakpoint)
                  
                  # create a mark for the last insertion point and set it to 
                  # the breakpoint and floats after the text that we insert
                  if not lastInsertionPoint:
                     lastInsertionPoint = 'yed.reflow.lastInsertionPoint'
                  buffer.mark_set(lastInsertionPoint, breakpoint)
                  buffer.mark_gravity(lastInsertionPoint, 'right')
                  buffer.insert(breakpoint, '\n' + prefix)
 
         # if the line shorter than the margin, merge the next line
         else:
            # get the newline prefix of the next line
            joinable, indent = self.isJoinable(buffer, cur)
            if joinable:
               buffer.insert('%s lineend' % cur, ' ')
               buffer.delete('%s lineend' % cur, 
                             '%s.%s' % (buffer.getRow(cur) + 1, indent)
                             )
               
               # don't move to the next line
               continue
               
         # move to the next line
         cur = buffer.index('%s + 1 lines linestart' % cur)
      
      # remove the end mark
      buffer.mark_unset(endMark)
      if lastInsertionPoint: buffer.mark_unset(lastInsertionPoint)

class Struct:
   """
      Structs manage a set of attribute-value pairs with delegation to 
      another object.  Attribute names beginning with an underscore should be 
      avoided.  The dictionary __setitem__() and __getitem__() methods are 
      also supported, allowing non-standard attribute names to be used.  The 
      remainder of the dictionary interface is intentionally unsupported in 
      order to avoid namespace collisions between attributes and dictionar
      method names.
      
      Derived classes can manage the construction of attributes by defining
      methods of the form _get_<attr-name>() - if such a method exists, it 
      will be called in preference to delegation.

      Protected-vars:
         _next: [@Struct] object to delegate attribute requests to.
         _attrs: [dict<str, any>] the object's local attributes map
   """
   
   def __init__(self, _next = None, _dict = None, **kw):
      """
         parms:
            _next: [any] Object to delegate attribute requests to if an 
               attribute is not held locally.  Any object will do providing it
               supports iteration over attribute, value pairs.
            _dict: [dict<str, any>] dictionary that will be used as the 
               attribute map.  If this is provided, the keyword arguments are 
               ignored.
            kw: [dict<str, any>] initial mapping of attributes to values.
      """
      self.__dict__ = {
         '_next': _next,
         '_attrs': _dict or kw,
      }

   def __getattr__(self, attr):
      try:
         return self._attrs[attr]
      except KeyError, ex:
         # see of the class has a _get_<attr> method, let it do the 
         # construction if it does.
         getterName = '_get_' + attr
         if hasattr(self.__class__, getterName):
            return getattr(self, getterName)()
            
         if self._next:
            return getattr(self._next, attr)
         else:
            raise AttributeError(ex[0])

   def __setattr__(self, attr, val):
      self._attrs[attr] = val

   # omitting implementing a __delattr__() because delegation requires that we 
   # either cache the fact that an attribute was deleted or make deletion mean 
   # "fall back to the value of the delegate" which seems broken.  And I don't 
   # need to do attribute deletion.
   
   def __iter__(self):
      
      # first deliver all of the attributes provided by our delegate, as long 
      # as we don't offer them ourselves.
      if self._next:
         for key, val in self._next:
            if not self._attrs.has_key(val):
               yield key, val
      
      # yield all of our attribute value pairs
      for key, val in self._attrs.iteritems():
         yield key, val
   
   def __getitem__(self, attr):
      return self.__getattr__(attr)
   
   def __setitem__(self, attr, val):
      return self.__setattr__(attr, val)

class RegionConfig(Struct):
   
   def _get_rules(self):
      val = self._next and self._next.rules
      self._attrs['rules'] = val
      return val
   
   def _get_inheritBindings(self):
      self._attrs['inheritBindings'] = val = True
      return val
   
   def _get_font(self):
      val = self._next and self._next.font
      self._attrs['font'] = val
      return val
   
   def _get_attrs(self):
      next = self._next and self._next.attrs
      self._attrs['attrs'] = val = Struct(next)
      return val
   
   def _get_bindings(self):
      next = self._next and self._next.bindings
      self._attrs['bindings'] = val = Struct(next)
      return val

def createAttrConfig(**kw):
   """
      Returns a RegionConfig consisting only of attributes.
      
      parms:
         kw: [dict<str, any>] dictionary of attributes and values.
   """
   config = RegionConfig()
   config.attrs = Struct(**kw)
   return config

class BufferTypeClass(type):
   
   def __init__(cls, name, bases, dict):
      if not dict.has_key('config'):
         # if config is not defined, create a config with attributes that 
         # delegate to the first base class
         cfg = dict['config'] = RegionConfig()
         cfg.attrs = Struct(self.__lookup(bases, 'config').attrs)
      
class BufferType:
   """
      Aggregates all information on the buffer type.
      
      Public-vars:
         name: [str] name of the type.  Type names must be unique.
         extensions: [list<str>] list of filename extensions that indicate the 
            buffer type.
         firstLinePatterns: [list<_sre.SRE_Pattern>] list of regular 
            expressions matching a first line that indicates the buffer type.
         config: [@RegionConfig] the configuration for the buffer itself.
         regions: [@Struct] Configuration for regions of the file.
   """
   
   name = 'unknown'
   extensions = []
   firstLinePatterns = []
   config = RegionConfig()
   regions = Struct()

   def initialize(self, buffer, top):
      """
         Initialize the buffer.
         
         parms:
            buffer: [@Buffer]
            top: [@Top]
      """
      # set the tab width
      tabWidth = top.app.get('hardTabWidth', 8)
      buffer.setTabWidth(tabWidth)

      # enable undo   
      buffer.text['undo'] = True

      # do event bindings
      for event, binding in self.config.bindings:
         buffer.text.bind(event, binding(top))
      
      # apply attributes
      buffer.text.configure(**dict(self.config.attrs))
      if self.config.font:
         buffer.setFont(self.config.font)
      
      # do the same for all of the regions
      for tagName, config in self.regions:
         for event, binding in config.bindings:
            buffer.text.tag_bind(tagName, event, binding(top))
         
         buffer.text.tag_config(tagName, **dict(config.attrs))
         if config.font:
            buffer.text.tag_config(tagName, font = config.font)
   
      # do the standard key bindings
      buffer.text.bind('<KeyPress>', AnyKey(top))
      buffer.text.bind('<FocusIn>', FocusHandler(top, buffer))
      
      # store the buffer type
      buffer.attrs['type'] = self.name
      
      # bind rules and do rules initialization
      buffer.rules = self.config.rules
      buffer.rules.initialize(buffer)
   
   def restore(self, top, name, attrs):
      """
         Restore a buffer from the attributes.
         
         Must be implemented by derived classes.
         
         parms:
            top: [@Top] the toplevel window.
            name: [str] the buffer name.
            attrs: [dict<str, any>] attributes dictionary loaded from the 
               statefile.
      """
      raise NotImplementedError('BufferType.restore() not implemented for ' +
                                 self.__class__.__name__
                                )

_bufferTypes = {}
_bufferTypesByExtension = {}
_defaultBufferType = None
def registerBufferType(bufferType):
   """
      Registers a new buffer type.
      
      Raises DuplicateBufferType if the buffer type name has already been 
      registered.  Prints a warning message to standard output if one of the 
      extensions is already registered.
   """
   if _bufferTypes.has_key(bufferType.name):
      raise DuplicateBufferType('Type %s was already registered' % 
                                bufferType.name
                                )
   
   _bufferTypes[bufferType.name] = bufferType
   for ext in bufferType.extensions:
      if _bufferTypesByExtension.has_key(ext):
         print 'Warning: extension %s already registered while registering ' \
            'type %s' % (ext, bufferType.name)
      else:
         _bufferTypesByExtension[ext] = bufferType

initializers = []
def registerInitializer(initializer):
   """
      Registers a function to be called during the initialization sequence 
      (after the App object has been instantiated).
   """
   initializers.append(initializer)

def getBufferType(filename, firstline):
   """
      Returns the @BufferType matching the buffer with the specified 
      characteristics.
   """
   for type in _bufferTypesByExtension.itervalues():
      for pattern in type.firstLinePatterns:
         if pattern.match(firstline):
            return type
   
   return _bufferTypesByExtension.get(os.path.splitext(filename)[1],
                                      _defaultBufferType
                                      )

def getBufferTypeByName(typeName):
   """
      Returns the @BufferType matching the buffer with the given name.
      Returns the defaulf buffer type (set with @setDefaultBufferType())
      if there is no matching name.

      parms:
         typeName: [str] a buffer type name
   """
   return _bufferTypes.get(typeName, _defaultBufferType)

def setDefaultBufferType(bufferType):
   """
      Sets the default buffer type - this is used to restore buffers that 
      don't have a specified buffer type.
      
      parms:
         bufferType: [@BufferType]
   """
   global _defaultBufferType
   _defaultBufferType = bufferType

audit = []

class FocusHandler:
   """
      Called when a buffer gets the focus.
   """
   def __init__(self, top, buffer):
      self.top = top
      self.buffer = buffer
   
   def __call__(self, event):
      if not self.buffer.attrs.get('keepStatusWidgets'):
         self.top.removeAllStatusWidgets()

class Top(Frame):
   """The toplevel window."""

   def applyBindings(self, bindings):
      for key, factory in bindings.items():
         self.top.bind(key, factory(self))

   def __init__(self, top, topBindings, app):
      Frame.__init__(self, top)
      self.top = top
      self.app = app
      self.buffers = []
      self.currentBuffer = None
      self.status = Label(self, text = 'Yed, baby')
      self.status.grid(row = 1, column = 0, sticky = NW)
      self.grid(row = 0, column = 0, sticky = NSEW)
      top.rowconfigure(0, weight = 1)
      top.columnconfigure(0, weight = 1)
      self.__statusWidgets = []
      self.__geom = None

      self.applyBindings(topBindings)

   def focus(self, buffer, forceFocus = True):
      """
         Give keyboard focus to the specified buffer and raise it to the top
         of the stack.

         parms:
            buffer:
               [Buffer]
            forceFocus::
               [boolean] if true, grab buffer focus even if the application 
               doesn't currently have the focus.
      """

      # raise and focus the toplevel (since we can be given focus remotely)
      if forceFocus:
         toplevel = self.winfo_toplevel()
         toplevel.tkraise()
      
      # raise the buffer, grab the focus if requested.
      buffer.tkraise()
      if forceFocus:
         buffer.text.focus_force()
      
      # record it as the current buffer
      self.currentBuffer = buffer

   def createBuffer(self, name):
      
      # create a new buffer
      text = Buffer(self, name)
      
      # insert it after the current buffer
      if self.currentBuffer:
         index = self.buffers.index(self.currentBuffer)
         self.buffers.insert(index + 1, text)
      else:
         self.buffers.append(text)
         
      # pack it into the frame
      text.grid(row = 0, column = 0, sticky = NSEW)
      self.rowconfigure(0, weight = 1)
      self.columnconfigure(0, weight = 1)

      if not self.currentBuffer:
         self.currentBuffer = text

      return text
   
   def nextBuffer(self, forward = True):
      """
         Cycle focus to the next (or previous) buffer in the ring.
      """
      buf = self.currentBuffer
      i = self.buffers.index(buf)
      i = (i + (forward and 1 or -1)) % len(self.buffers)
      buf = self.buffers[i]
      self.focus(buf)
      return buf

   def selectBuffer(self, bufferName, subtext = False, forceFocus = True):
      """
         Selects the buffer with the specified name.  Returns the buffer, 
         returns *None* if no buffer of that name exists.
         
         parms:
            bufferName::
               [string] the buffer name to find
            subtext::
               [boolean] if true, select the buffer even if the name contains
               the text
            forceFocus::
               [boolean] if true, grab buffer focus even if the application 
               doesn't currently have the focus.
      """
      for buffer in self.buffers:
         if buffer.name == bufferName or \
            subtext and bufferName in buffer.name:
            self.focus(buffer, forceFocus)
            return buffer
      else:
         return None
   
   def insertBuffer(self, anchor, buffer):
      """
         Change the position of 'buffer' in the buffer ring so that it is 
         located after 'anchor'.
         
         parms:
            anchor: [Buffer] buffer to insert it after.
            buffer: [Buffer] buffer to be moved
      """
      self.buffers.remove(buffer)
      i = self.buffers.index(anchor)
      self.buffers.insert(i + 1, buffer)
   
   def findBuffer(self, **query):
      """
         Finds the first buffer that matches the query criteria.
         
         Returns *None* if no matching buffer is found
         
         parms:
            query::
               [dict<string, object>] a dictionary of buffer attributes to 
               query.  Al of them must succeed.
      """
      for buffer in self.buffers:
         for attr, val in query.items():
            if buffer.attrs.get(attr) != val and \
               buffer.nattrs.get(attr) != val:
               break
         else:
            return buffer
      else:
         return None

   def destroyBuffer(self, buffer):
      """
         Destroys the buffer - if this was the last buffer in the toplevel, 
         also destroys the toplevel and returns true
      """
      # if we're going to delete the current buffer, cycle to the previous 
      # buffer
      if buffer is self.currentBuffer:
         self.nextBuffer(False)
         
      # call the destroy handlers
      buffer.callDestroyHandlers()

      # if we're going to delete the last buffer, quit the editor
      if len(self.buffers) == 1:
         
         # before we destroy there's a couple of things that we need to do to 
         # go down clean.
         
         # notify the app
         self.app.handleShutdown()
         
         # store the geometry so that a subsequent state save doesn't try to 
         # query a destroyed window
         self.__geom = self.winfo_toplevel().wm_geometry()
         
         self.top.destroy()
         return True
      else:
         # just remove the buffer and delete it
         self.buffers.remove(buffer)
         buffer.destroy()
         return False

   def addStatusWidget(self, widget):
      self.__statusWidgets.append(widget)
      widget.grid(row = 1, column = 0, sticky = NSEW)
   
   def __hasFocus(self, widget):
      """
         Returns true if the widget or any of its children has the focus.
      """
      focus = self.focus_get()
      while str(focus) != '.':
         if widget == focus:
            return True
         focus = focus.nametowidget(focus.winfo_parent())
      else:
         return False
         
   def removeStatusWidget(self, widget):
      
      # if this widget currently has the focus, and there's no status window
      # lined up to get it next, give the focus back to the buffer
      if self.__hasFocus(widget) and len(self.__statusWidgets) == 1 \
         and self.currentBuffer:
         self.currentBuffer.focus()
         
      self.__statusWidgets.remove(widget)
      widget.handleDestroy()
      widget.destroy()
      
   def removeAllStatusWidgets(self):
      for widget in self.__statusWidgets:
         if self.__hasFocus(widget) and self.currentBuffer:
            self.currentBuffer.focus()

         widget.handleDestroy()            
         widget.destroy()
      self.__statusWidgets = []

   def setStatusText(self, text):
      """
         Sets the text of the status bar at the bottom of the window.

         parms:
            text::
               [string]
      """
      self.status['text'] = text
   
   def showError(self, text):
      """
         Shows the error text in the status bar.
      """
      self.setStatusText(text)

   def addPanelWidget(self, widget):
      widget.grid(row = 0, column = 0, sticky = NSEW)
   
   def writeState(self, out):
      """
         Writes top-level state information to the specified output file.
      """
      token = {
         'bufferCount': len(self.buffers), 
         'currentBuffer': self.currentBuffer.name,
         'geometry': self.__geom or self.winfo_toplevel().wm_geometry(),
      }
      Pickler(out).dump(token)
      for buffer in self.buffers:
         buffer.writeState(out)
   
   def readFrom(self, src):
      """
         Returns a new buffer read from the provided source file.
         
         parms:
            src::
               [file.read] source of state info
      """
      token = Unpickler(src).load()
      #print 'loaded %s' % token
      for i in range(token['bufferCount']):
         Buffer.readFrom(self, src)
      
      self.selectBuffer(token['currentBuffer'], forceFocus = False)
      geo = token.get('geometry')
      if geo:
         self.winfo_toplevel().wm_geometry(geo)

class CommandContext:
   """
      This gets passed to a command whenever it is invoked and contains the 
      context of the command.
   """

   def __init__(self, top, buffer, event, args):
      """
         parms:
            top: [@Top] the toplevel
            buffer: [@Buffer] the current buffer
            event: ??? the event that triggered this
            args: [list<string>] list of additional arguments (which may be 
               passed in when the command is invoked from the command line).
            result: [str or None] this field allows yedlets to set the return 
               code for the process function - it is mainly useful for yedlets
               bound to key events to allow them to return 'break' to abort 
               subsequent key processing.
      """
      self.top = top
      self.buffer = buffer
      self.event = event
      self.args = args
      self.result = None

class Command:
   """
      Base class for all yed commands.
   """
   def __init__(self, top):
      self.__top = top

   def __call__(self, event, *args):
      if self.process(CommandContext(self.__top, self.__top.currentBuffer, 
                                     event,
                                     args
                                     )
                      ):
         return 'break'

   def process(self, context):
      """
         This method executes the command.
         
         Must be implemented by derived classes.
         
         parms:
            context: [@CommandContext]
         
         returns:
            True if all subsequent event processing should be aborted, false 
            if not.
      """
      raise NotImplementedError()

class CommandFactory:
   """
      Formal definition of a "command factory" - a class whose instances can 
      be used to create commands.
   """

   def __call__(self, top):
      raise NotImplementError()   

class RangeCmd(Command):
   """
      Abstract base class for commands that operate on a range of lines.
   """
   
   # override this variable to leave the selection after processing the command
   unselectAfterProcessing = True
   
   # set this to true if your command performs multiple edits and you want 
   # them to appear to be a single edit for purposes of undo/redo
   atomicEdits = False
   
   def process(self, context):
      ranges = context.buffer.tag_ranges('sel')
      if len(ranges) == 2:
         start, end = ranges
         
         # turn off auto-separators if we are in atomic edits mode
         if self.atomicEdits:
            context.buffer.edit_separator()
            context.buffer.setAutoSeparators(False)
            
         try:
            result = self.processRange(context, start, end)
         finally:
            if self.atomicEdits:
               context.buffer.setAutoSeparators(True)
               context.buffer.edit_separator()
         
         # unselect if appropriate
         if self.unselectAfterProcessing:
            context.buffer.tag_remove('sel', start, end)
         
         return result
         
      else:
         # give it to the non-range processing function
         return self.processNonRange(context)
   
   def processRange(self, context, start, end):
      """
         Called with a range.  Must be implemented by derived class.
         
         If this returns true, the range will remain selected (the selection
         will not be removed)
         
         parms:
            start::
               [string] start index
            end::
               [string] end index
      """
      raise NotImplementedError()
      
   def processNonRange(self, context):
      """
         Called when a range function is called when no range is selected.
         By default, this just gives an alarm bell and an error message.
         Derived classes may override.
      """
      context.top.setStatusText('You must select a range')
      context.buffer.bell()
      return True

class AnyKey(RangeCmd):
   """
      Command invoked when any key is pressed.  This handles basic character
      insertion.
   """
   
   # leave the selection intact
   unselectAfterProcessing = False
   
   def __isPlainKey(self, event):
      return event.char and len(event.char) == 1 and \
         32 <= ord(event.char) < 127
      
   def processRange(self, context, start, end):
      if self.__isPlainKey(context.event):
         context.buffer.setAutoSeparators(False)
         context.buffer.delete(start, end)
         context.buffer.insert('insert', context.event.char)
         context.buffer.see('insert')
         showBufferInfo(context.top, context.buffer)
         context.buffer.setAutoSeparators(True)
         return True
      else:
         # XXX this sets the buffer state information _before_ the action,
         # which is normally not what we want.
         showBufferInfo(context.top, context.buffer)
         
   def processNonRange(self, context):
      if self.__isPlainKey(context.event):
         
         buffer = context.buffer
         
         # see if we've gone behond the textwidth
         if buffer.getColumn('insert') > buffer.attrs.get('textwidth', 78) and \
            buffer.rules:
            
            # find the best line break point (if there is no breakpoint, 
            # there's not a lot we can do)
            breakpoint = context.buffer.rules.getBreakpoint(buffer, 'insert')
            if breakpoint:
               prefix = buffer.rules.calcNewlinePrefix(buffer, breakpoint)
               buffer.insert(breakpoint, '\n' + prefix)
            
         buffer.insert('insert', context.event.char)
         buffer.see('insert')
         showBufferInfo(context.top, buffer)
         return True
      else:
         # XXX this sets the buffer state information _before_ the action,
         # which is normally not what we want.
         showBufferInfo(context.top, context.buffer)

def showBufferInfo(top, buffer):
   """
      Puts the buffer info into the toplevel's status bar.
   """
   top.setStatusText(buffer.name + 
                     (buffer.getModified() and ' [Modified]' or '') +
                     ' ' + buffer.index('insert')
                     )

