# GPL3 me
"""
    Abstract filesystem interfaces.
"""

import os, stat, glob, fnmatch, copy
import sys # XXX for debug

class FileObj:
    """
        File info object.
        
        Public vars:
            dir: [string] directory
            name: [string] filename
            mode: [int] filetype and permissions
            inode: [long] inode number
            dev: [long] device number
            nlink: [int] number of hard links
            uid: [int] user ID of owner
            gid: [int] group ID of owner
            size [long] size of file, in bytes
            atime: [int] time of most recent access
            mtime: [int] time of most recent content modification
            ctime: [int] time of meta-data change, or time of creation in 
                   windows.
    """
    
    def __init__(self, dir, name, mode, inode, dev, nlink, uid, gid, size,
                 atime, mtime, ctime):
        self.dir = dir
        self.name = name
        self.mode = mode
        self.inode = inode
        self.dev = dev
        self.nlink = nlink
        self.uid = uid
        self.gid = gid
        self.size = size
        self.atime = atime
        self.mtime = mtime
        self.ctime = ctime

    @property    
    def fullname(self):
        return os.path.join(self.dir, self.name)
    
    def __str__(self):
        return self.fullname
    
    def translate(self, prefix):
        """
            Returns a new FileObj that is a copy of self with "prefix" 
            prepended to the path.
            
            parms:
                prefix: [string]
        """
        dup = copy.copy(self)
        dup.dir = prefix + self.dir
        return dup
    
    def asStat(self):
        """
            Returns the stat information as a standard stat tuple.
        """
        return (self.mode, self.inode, self.dev, self.nlink, self.uid, 
                self.gid, self.size, self.atime, self.mtime, self.ctime)

class Filer:
    """
        Base class for an interface to a filesystem.
    """
    
    def listdir(self, dir):
        """
            Returns the contents of the specified directory (list<string>).
            
            Must be implemented by derived classes.
            
            parms:
                dir: [string] the directory name
        """
        raise NotImplementedError()
    
    def read(self, filename):
        """
            Reads the specified file.  Returns the entire file contents.
            
            Must be implemented by derived classes.
            
            parms:
                filename: [string] the file name
        """
        raise NotImplementedError()
    
    def write(self, filename, contents, mode = None):
        """
            Writes the specified file.  Returns the stat info of the new file.
            
            Must be implemented by derived classes.
            
            parms:
                filename: [string] the file name
                contents: [string] the entire contents of the file.
                mode: [int] permissions of the file
        """
        raise NotImplementedError()
    
    def isfile(self, filename, cacheOk = False):
        """
            Returns true if 'filename' is a file object.
            
            Must be implemented by derived classes.

            parms:
                filename: [string] the file name
                cacheOk: [boolean] if true, then it's ok to retyrn a stat 
                    information stored in a cache.
        """
        raise NotImplementedError()
    
    def isdir(self, filename):
        """
            Returns true if 'filename' is a directory.
            Must be implemented by derived classes.

            parms:
                filename: [string] the file name
        """        
        raise NotImplementedError()
    
    def glob(self, pattern):
        """
            Returns the list of all files matching the specified wildcard 
            pattern.
            
            Must be implemented by derived classes.
            
            parms:
                pattern: [string] the filename pattern.  All of the standard 
                wildcards are supported.
        """
        raise NotImplementedError()
    
    def stat(self, filename, cacheOk = False):
        """
            Returns a tuple of status information for a file.  Returns *None* 
            if the file does not exist or is otherwise inaccessible.
            
            Must be implemented by derived classes.
            
            parms:
                filename: [string]
                cacheOk: [boolean] if true, then it's ok to retyrn a stat 
                    information stored in a cache.
        """
        raise NotImplementedError()
    
    def exists(self, filename):
        """
            Returns true if the specified file exists.
            
            parms:
                filename: [string]
        """
        return bool(self.stat(filename))
    
    def getPrefixed(self, stem):
        """
            Returns a list<FileObj> of filenames beginning with the specified 
            stem.
            
            parms:
                stem: [string] the beginning of a path.  This will match all 
                entries starting with the name that are in the same directory 
                as the name.  For example, "/foo/bar" will match "/foo/bart" 
                and "/foo/barman", but not "/foo/bar/baz"
        """
        raise NotImplementedError()

class LocalFiler(Filer):
    """
        Basic local filesystem interface.
    """
    
    def listdir(self, dir):
        return os.listdir(dir)
    
    def read(self, filename):
        return open(filename).read()
    
    def write(self, filename, contents, mode = None):
        result = open(filename, 'w').write(contents)
        if mode is not None:
            os.chmod(filename, mode)
        return os.stat(filename)
    
    def isfile(self, filename, cacheOk = False):
        return os.path.isfile(filename)
    
    def isdir(self, filename):
        return os.path.isdir(filename)
    
    def glob(self, pattern):
        return glob.glob(pattern)
    
    def stat(self, filename, cacheOk = False):
        try:
            return os.stat(filename)
        except OSError:
            return None
    
    def getPrefixed(self, stem):
        dirname, basename = os.path.split(stem)
        results = []
        if not os.path.isdir(dirname):
            return results
        for file in self.listdir(dirname):
            if file.startswith(basename):
                fullName = os.path.join(dirname, file)
                
                # statInfo can be None if the file has been removed since 
                # listing the dir or if it is a dangling symlink.
                statInfo = self.stat(fullName)
                if statInfo is not None:
                    results.append(FileObj(dirname, file, *statInfo))

        return results

class RemoteFiler(Filer):
    
    def __init__(self, remote):
        self.__remote = remote
        self.__cachedPath = None
        self.__cached = None
    
    def __del__(self):
        self.__remote.close()
    
    def __getDir(self, dir):
        if self.__cachedPath != dir:
            result = self.__remote.listdir(dir)
            self.__cached = {}
            for key, val in result.items():
                self.__cached[key] = FileObj(dir, key, *val)
            self.__cachedPath = dir
        return self.__cached
        
    def listdir(self, dir):
        filenames = [f.name for f in self.__getDir(dir).iteritems()]
        filenames.sort()
        return filenames
    
    def read(self, filename):
        return self.__remote.readFile(filename)
    
    def write(self, filename, contents, mode):
        return self.__remote.writeFile(filename, contents, mode)

    def stat(self, filename, cacheOk = False):
        dir, base = os.path.split(filename)
        if self.__cachedPath == dir and cacheOk:
            result = self.__cached.get(base)
            if result: result = result.asStat()
        else:
            result = self.__remote.stat(filename)
        assert result is None or isinstance(result, tuple)
        return result
    
    def isfile(self, filename, cacheOk = False):
        st = self.stat(filename)
        if not st:
            return False
        else:
            return stat.S_ISREG(st[stat.ST_MODE])
    
    def isdir(self, filename):
        st = self.stat(filename)
        if not st:
            return False
        else:
            return stat.S_ISDIR(st[stat.ST_MODE])
    
    def glob(self, pattern):
        """XXX deprecated."""
        dirname, basename = os.path.split(pattern)
        results = []
        for file in self.__getDir(dirname).iteritems():
            if fnmatch.fnmatch(file.name, basename):
                results.append(file.fullname)
        return results

    def getPrefixed(self, stem):
        results = []
        dirname, basename = os.path.split(stem)
        for file in self.__getDir(dirname).itervalues():
            if file.name.startswith(basename):
                results.append(file)
        return results
    
class CompoundFiler(Filer):
    """
        A combination filer that allows paths to be designated as local or 
        global (effectively, a crude form of mounting).
        
        Paths are stored sequentially, and inserted from the beginning, so the 
        last matching path overrides any previously inserted matching path.
    """
    
    def __init__(self):
        self.__filers = []
    
    def mount(self, path, filer):
        """
            Adds a new filer for providing access to the specified path.

            parms:
                path: [string] the path to the filer.
                filer: [@Filer] the filer used for access to names in that 
                    path.
        """
        # normalize the path to the form "/.../"
        if not path.endswith('/'):
            path += '/'
        
        # Make sure that the path is not already mounted.
        for filerPath, existingFiler in self.__filers:
            if path == filerPath:
                raise KeyError(filerPath)

        self.__filers.insert(0, (path, filer))
    
    def unmount(self, path):
        """
            Remove a mounted path.  Raises KeyError if the path is not mounted.
            
            parms:
                path: [string] the path to be unmounted.  This should have 
                previously been mounted with @mount().
        """
        # normalize the path to the form "/.../"
        if not path.endswith('/'):
            path += '/'

        i = 0
        for mountPoint, filer in self.__filers:
            if mountPoint == path:
                del self.__filers[i]
                break
            i += 1
        else:
            raise KeyError(path)
    
    def __findMatch(self, path):
        for mountPoint, filer in self.__filers:
            if path.startswith(mountPoint):
                return filer, mountPoint, '/' + path[len(mountPoint):]
        else:
            assert False, "No path matching %s" % repr(path)

    def listdir(self, dir):
        filer, prefix, dir = self.__findMatch(dir)
        return filer.listdir(dir)
    
    def read(self, filename):
        filer, prefix, filename = self.__findMatch(filename)
        return filer.read(filename)
    
    def write(self, filename, contents, mode = None):
        filer, prefix, filename = self.__findMatch(filename)
        return filer.write(filename, contents, mode)
    
    def isfile(self, filename, cacheOk = False):
        filer, prefix, filename = self.__findMatch(filename)
        return filer.isfile(filename, cacheOk)
    
    def isdir(self, filename):
        filer, prefix, filename = self.__findMatch(filename)
        return filer.isdir(filename)
    
    def glob(self, pattern):
        # XXX won't work in the general case
        #print >>sys.stderr, 'globbing %s' % pattern
        filer, prefix, pattern = self.__findMatch(pattern)
        return [prefix[:-1] + x for x in filer.glob(pattern)]

    def getPrefixed(self, stem):
        filer, prefix, filename = self.__findMatch(stem)
        return [f.translate(prefix[:-1]) for f in filer.getPrefixed(filename)]

    def stat(self, filename, cacheOk = False):
        filer, prefix, filename = self.__findMatch(filename)
        return filer.stat(filename, cacheOk)