#==============================================================================
#
"""
    Shell commands.
"""
#
#  Copyright (C) 2008 Google Inc.
#
#   This file is part of SpugEdit.
#
#   SpugEdit is free software: you can redistribute it and/or modify it under
#   the terms of the GNU Lesser General Public License as published by the
#   Free Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   SpugEdit is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with SpugEdit.  If not, see <http://www.gnu.org/licenses/>.
#
#==============================================================================

import os, stat
from threading import Lock
from spug.io.proactor import DataHandler
from core import Command, registerInitializer
from yed import defaultInitializer, getApp, showBufferInfo
from yed.buf.term import TermBufferType

class ShellController(object):
    """
        Keeps track of an open subprocess.  When all of the output streams 
        close, waits for the process to complete.
    """
    
    def __init__(self, communicator, buffer):
        self.__process = None
        self.__opened = set([1, 2])
        self.__lock = Lock()
        self.__comm = communicator
        self.__buffer = buffer

    def __writeTerminator(self, top):
        # write a nice terminator to the buffer
        self.__buffer.insert('end', '<END>', 'meta')

    def __waitForTermination(self):
        self.__process.wait()
        self.__process = None
        self.__comm.runInUIThread(self.__writeTerminator)

    def __setProcess(self, process):
        self.__lock.acquire()
        try:
            self.__process = process
            # if all of the handles have already been closed, we need to do a 
            # wait now.
            if not self.__opened:
                self.__waitForTermination()
        finally:
            self.__lock.release()
    
    def __getProcess(self):
        return self.__process

    process = property(__getProcess, __setProcess)
    
    def handleClose(self, fd):
        self.__lock.acquire()
        try:
            self.__opened.remove(fd)
            if not self.__opened and self.__process:
                self.__waitForTermination()
        finally:
            self.__lock.release()
    
    def kill(self):
        """Kill the process if it has not already been killed."""
        self.__lock.acquire()
        try:
            if self.__process: self.__process.kill()
        finally:
            self.__lock.release()
    
class OutHandler(DataHandler):
    """
        Receives output, sends it to the buffer.
    """
    
    def __init__(self, fd, communicator, buffer, controller):
        self.__fd = fd
        self.__comm = communicator
        self.__buf = buffer
        self.__controller = controller
    
    def __writeToBuffer(self, data):
        if self.__buf:
            self.__buf.insert('end', data)

    def readyToGet(self):
        return False
    
    def readyToPut(self):
        return True
    
    def readyToClose(self):
        return False
    
    def put(self, data):
        self.__comm.runInUIThread(lambda top: self.__writeToBuffer(data))
    
    def handleDisconnect(self):
        self.__controller.handleClose(self.__fd)
    
    def handleBufferDestroyed(self):
        self.__buf = None

class NullInputHandler(DataHandler):
    """
        Handler for standard input that immediately closes the stream.
    """
    
    def readyToGet(self):
        return False
    
    def readyToPut(self):
        return False
    
    def readyToClose(self):
        return True

def createTermBuffer(top, name):
    "Create a new buffer for command output and return it."
    # create a new buffer, bring it to the top
    buffer = top.createBuffer(name)
    TermBufferType().initialize(buffer, top)
    top.focus(buffer)
    showBufferInfo(top, buffer)
    return buffer
    
def isExecutable(path):
    """Returns true if the path is an executable file."""
    try:
        statBuf = os.stat(path)
    except OSError, ex:
        return False

    def hasPerm(perm):
        return statBuf.st_mode & perm or \
            statBuf.st_mode & (perm << 3) and statBuf.st_gid in groups or \
            statBuf.st_mode & (perm << 6) and statBuf.st_uid == uid
    
    # make sure that it is a file and is readable and writable
    uid = os.getuid()
    groups = os.getgroups()
    return stat.S_ISREG(statBuf.st_mode) and hasPerm(1) and \
        hasPerm(4)

def runProc(args, context, buffer = None, bufferName = None, 
            pathLookup = True):
    """Run a process, collect output in a buffer.
    
    Args:
        args: [list<str>] the full command argv
        context: [yed.core.CommandContext]
        buffer: [@yed.Buffer or None] If specified, this is the buffer to use 
            for command standard output and standard error.  If None, a new 
            buffer is created.
        bufferName: [str or None] If defined, and a buffer is to be created, 
            this will be the name of the new buffer.
        pathLookup: [bool] if true (the default), resolve command names that 
            aren't absolute paths by searching the directories in the 
            environment's PATH variable.
    """            
    if not buffer:
        buffer = createTermBuffer(context.top, 'shell ' + ' '.join(args))
    
    # store commonly used variables and create process control objects.
    app = getApp()
    comm = app.communicator
    controller = ShellController(comm, buffer)
    stdoutHandler = OutHandler(1, comm, buffer, controller)
    stderrHandler = OutHandler(2, comm, buffer, controller)
    
    # do a path lookup if requested and the command isn't an absolute path.
    if pathLookup and not os.path.isabs(args[0]):
        cmd = args[0]
        paths = os.environ.get('PATH', '').split(os.pathsep)
        for path in paths:
            fullCmd = os.path.join(path, cmd)
            if isExecutable(fullCmd):
                args[0] = fullCmd
                break
        else:
            context.top.showError('%s not found' % repr(cmd))
            return

    # function to start the subprocess, runs in a subshell.
    def startSubprocess():
        
        # the buffer's destroy handler - unhooks the output handlers from 
        # the buffers and kills the process if it is still running.
        def killProcess(buffer):
            stdoutHandler.handleBufferDestroyed()
            stderrHandler.handleBufferDestroyed()
            comm.runInCommThread(controller.kill)

        # function to be run in the UI thread that adds the destroy 
        # handler.
        def addDestroyHandler(top):
            buffer.addDestroyHandler(killProcess)

        # function to be run in the UI thread that displays an error.
        def showError():
            context.top.showError(error)

        try:
            (process, stdin, stdout, stderr) = \
                app.communicator.proactor.makeSubprocess(args,
                                                         NullInputHandler(),
                                                         stdoutHandler,
                                                         stderrHandler)
            controller.process = process
            comm.runInUIThread(addDestroyHandler)
        except OSError, ex:
            error = ex[0]
            comm.runInUIThread(showError)
    
    app.communicator.runInCommThread(startSubprocess)

class ShellCmd(Command):
    """
        shell <cmd> ...
        
        Executes a shell command, creates a new buffer for the command's 
        output.
    """
    
    def process(self, context):
        runProc(context.args, context)

def registerAll():
    getApp()['sh'] = ShellCmd(getApp().top)

registerInitializer(registerAll)
