#==============================================================================
#
"""
    Communications unit tests.
"""
#
#  Copyright (C) 2008 Google Inc.
#
#   This file is part of SpugEdit.
#
#   SpugEdit is free software: you can redistribute it and/or modify it under
#   the terms of the GNU Lesser General Public License as published by the
#   Free Software Foundation, either version 3 of the License, or (at your
#   option) any later version.
#
#   SpugEdit is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with SpugEdit.  If not, see <http://www.gnu.org/licenses/>.
#
#==============================================================================

from threading import Lock, Thread
from spug.io.proactor import INETAddress
from yed.comm import Authenticator, _HTTPResponse, AuthError, Communicator, \
    NotConnectedError, TimeoutError
from unittest import main, TestCase

class MockAuthenticator(Authenticator):
    """
        Lets us test authentication and keep it out of the way most of the 
        time.
    """
    
    def authorized(self, token):
        return token == 'pass'

class NormalCommTests(TestCase):
    
    def uiThreadFunc(self):
        while not self.finished:
            self.comm.processUIQueue(None)
    
    def lsCommand(self, request):
        self.commandLock.acquire()
        self.commandLock.release()
        request.sendReply('ok')

    def setUp(self):
        self.comm = Communicator({'ls': self.lsCommand}, 
                                 MockAuthenticator()
                                 )
        self.comm.start()
        self.finished = False
        self.commandLock = Lock()
        self.__uiThread = Thread(target = self.uiThreadFunc)
        self.__uiThread.start()
        self.port = self.comm.addTCPServer(0)
        self.rem = self.comm.createRemote(INETAddress('localhost', self.port))
        self.rem.authToken = 'pass'

    def tearDown(self):
        if self.port:
            self.comm.removeTCPServer(self.port)
        self.finished = True
        self.__uiThread.join()
        del self.comm

    def testRemoteCommands(self):
        self.assertEquals('ok', self.rem.ls())
    
    def testTimeoutRecovery(self):
        self.commandLock.acquire()
        try:
            self.rem.ls(0)
        except TimeoutError:
            pass
        
        self.commandLock.release()
        self.rem.waitForConnect()
        self.testRemoteCommands()
    
    def testExceptions(self):
        # XXX I've observed this test to fail sporadically, but I'm not sure 
        # why - it's as though waitForDisconnect doesn't always work.
        self.comm.removeTCPServer(self.port)
        self.port = 0
        self.rem.waitForDisconnect()
        try:
            self.rem.ls()
            self.fail()
        except NotConnectedError, ex:
            self.assertEquals('Disconnected', ex[0])
    
    def testAuthExceptions(self):
        self.rem.authToken = 'fail'
        self.assertRaises(AuthError, self.rem.ls)

main()